;;;; File: dm-services.lsp
;;; SCCS Version: %W%
;;; Contains: Contains the functions which provide the interfaces to the DM
;;; Author: Richard Kirby (rbk) and Jeff Dalton
;;; Created: Wed Oct 31 13:40:54 1990
;;; Updated: Sat Jul  8 22:45:33 1995 by Jeff Dalton
;;; Release Version: %Y%
;;; Copyright: (c) 1992, 1993, 1994 AIAI, University of Edinburgh
;;; This material may be reproduced by or for the U.S. Government pursuant
;;; to the copyright license under the clause at DFARS 252.227-7032
;;; (June 1975) -- Rights in Technical Data and Computer Software (Foreign).

(in-package :oplan)

;;;; Interrupt handler.

;;; Present a menu to the user for interrogating the Plan State and
;;; ADS, etc.

(defun dm-interrupt-handler ()
  (loop
    (ecase (dm-developers-menu-choice)
      (:nodes (do-display-all-nodes *debug-io*)
	      (terpri *debug-io*))
      (:tome  (do-display-tome *debug-io*)
	      (terpri *debug-io*))
      (:gost  (do-display-gost *debug-io*)
	      (terpri *debug-io*))
      (:psvs  (psv-print-psv-table *debug-io*)
	      (terpri *debug-io*))
      (:break
       (break "Return to get DM Developer's menu."))
      (:quit
       (force-output *debug-io*)
       (return)))))

(defun dm-developers-menu-choice ()
  (menu-request
    `("-heading" "DM Developer's Menu"
      "Display nodes=:nodes"
      "Display TOME=:tome"
      "Display GOST=:gost"
      "Display PSVs=:psvs"
      "BREAK IN=:break"
      "-line"
      "QUIT=:quit")))

(defun do-display-all-nodes (stream)
  (walk-nodes #'(lambda (n) (do-display-a-node n stream))))

(defun do-display-a-node (n stream)
  (let ((start-tp (n-begin-tpoint n))
	(end-tp (n-end-tpoint n))
	(*print-level* nil)
	(*print-length* nil))
    (declare (ignore start-tp end-tp))
    (format stream "~S~10T : Begin PRENODES = ~S~%"
	    (n-tag n)
	    (mapcar #'ne-tag (ne-pre-ends (n-begin n))))
    (format stream "(~S)~10T : Begin POSTNODES = ~S~%"
	    (n-type n)
	    (mapcar #'ne-tag (ne-post-ends (n-begin n))))
    (format stream "~S~10T : End PRENODES = ~S~%"
	    (n-tag n)
	    (mapcar #'ne-tag (ne-pre-ends (n-end n))))
    (format stream "(~S)~10T : End POSTNODES = ~S~%"
	    (n-type n)
	    (mapcar #'ne-tag (ne-post-ends (n-end n))))
    (format stream "~10T : ~A~%"
	    (psv-replace-psvs-in-pattern (n-pattern n)))))

(defun do-display-tome (stream &optional (filter-out-undefs t))
  (let ((act::*bindings* nil))
    (let ((g (generator '(TGM::TOME ?? ??) '??))
	  (*print-level* nil)
	  (*print-length* nil))
      (format stream "~%The TOME:~%")
      (do ((item (try-next g) (try-next g)))
	  ((null item))
	(if (or (null filter-out-undefs)
		(null (eql (item-inst (value item)) :undef)))
	    (format stream "~A = ~A~%"
		    (item-inst item) (item-inst (value item))))))))

(defun do-display-gost (stream &optional (filter-out-undefs t))
  (let ((act::*bindings* nil))
    (let ((g (generator '(TGM::GOST ?? ?? ?? ??) '??))
	  (*print-level* nil)
	  (*print-length* nil))
      (format stream "~%The GOST:~%")
      (do ((item (try-next g) (try-next g)))
	  ((null item))
	(if (or (null filter-out-undefs)
		(null (eql (item-inst (value item)) :undef)))
	    (format stream "~A = ~A~%"
		    (item-inst item) (item-inst (value item))))))))


;;;; Schema processing

;; Index of schemas by expansion field.
(defvar *schema-expansion-table* (make-hash-table :size 10 :test #'equal))

;; Index of schemas by only-use-for-effects field.
(defvar *schema-effects-table* (make-hash-table :size 10 :test #'equal))

;; Index of schemas by name field.
(defvar *schema-name-table* (make-hash-table :size 10 :test #'equal))

; /\/: Why was this ever here?  [jwd 01 Apr 93]
; (defvar schema-defs nil)

(defun clear-schema-tables ()
  (clrhash *schema-expansion-table*)
  (clrhash *schema-effects-table*)
  (clrhash *schema-name-table*)
  (clear-schema-possible-effects-tables))

;; Loads in some more schema definitions into the schema tables.
;
; See database_manager/delta/schemacode.lsp.
;
; (defun init-schemas (schema-defs)
;   ...)

;; Takes a schema pattern, and a requirement for matching, and the possible
;; schema vars with their restrictions, and returns the required bindings of
;; variables to things in the object, if a match (taking the restrictions
;; into account), else nil.
(defun match-up (pattern object binding-list)
  (let ((act::*bindings* binding-list))
    (obmatch pattern object)))

;; Returns the variable name for a ACT:GIVEN actor.
(defun given-var (x)
  (second (third x)))

;; /\/: "db-" in db-instantiate-schema because kp_supportlib also has
;; an instantiate-schema (which does a db-request).

(defun db-instantiate-schema (schema bindings)
  ;; This function goes through some of the fields of the schema, instantiating
  ;; any where there are actors, to the specified bindings. Before that though,
  ;; if any schema vars are left unbound presently, then replace them with a
  ;; Plan State variable.
  (let (act::*bindings*)
    (when (consp bindings)
      (mapc #'(lambda (x)
		(if (and (not (psv-p (var-name x)))
			 (eql (var-value x) :undef))
		    (setf (var-value x)
			  (psv-create-psv schema (var-name x) bindings))))
	    bindings)
      ;; Now for each new psv, if the bindings tuple specifies some not sames,
      ;; add the not-sames as restrictions.
      (dolist (tuple bindings)
	(let ((tuple-value (var-value tuple)))
	  (when (psv-p tuple-value)
	    (let ((not-sames (var-not-sames tuple)))
	      (when not-sames
		(dolist (not-same not-sames)
		  (psv-add-restriction tuple-value
		      :NOT (var-value (get-tuple not-same bindings))))))))))
    (setq schema (copy-schema schema))
    (setf (schema-expands schema) (instance (schema-expands schema) bindings))
    (setf (schema-only-use-for-effects schema)
	  (instance (schema-only-use-for-effects schema) bindings))
    (setf (schema-nodes schema) (instance (schema-nodes schema) bindings))
    (setf (schema-conditions schema)
	  (instance (schema-conditions schema) bindings))
    (setf (schema-effects schema)
	  (instance (schema-effects schema) bindings))
    (setf (schema-resources schema)
	  (instance-if-bound (schema-resources schema) bindings))
    (setf (schema-time-windows schema)
	  (instance-if-bound (schema-time-windows schema) bindings))
    (cons schema bindings)))

(defun instance-if-bound (x bindings)
  (let ((inst-x (instance x bindings)))
    (when (contains-psvs-p inst-x)
      (cerror "continue anyway" "Put a PSV in ~S,~%giving ~S" x inst-x))
    inst-x))

(defun contains-psvs-p (x)
  (or (psv:psv-p x)
      (and (consp x)
	   (or (contains-psvs-p (car x))
	       (contains-psvs-p (cdr x))))))

(defun create-binding-tuples-from-schema (schema)
  (let ((result (mapcar
		 #'(lambda (x)
		     (if (actorp (cadr x))
			 (make-tuple (car x) :undef nil (cadr x))
			 (if (eql (cadr x) 'undef)
			     (make-tuple (car x) :undef nil actorsym)
			     (make-tuple (car x) (cadr x) nil actorsym))))
		 (schema-vars schema))))
    ;; Now for each entry in the vars-relations field, set the not sames field
    ;; if appropriate.
    (let ((act::*bindings* result))
      (dolist (not-sames (schema-relationships schema))
	(let ((var1 (given-var (car not-sames)))
	      (var2 (given-var (caddr not-sames))))
	  (add-to-not-sames var1 var2)))
      (setq result act::*bindings*))
    result))


;;;; node and link addition.

(defun add-a-link (from to)
  ;; RBK - 8/1/92 Just check that the to node-end is not already before the
  ;; from node-end, in which case return nil, so higher up can decide what
  ;; to do.
  (cond ((gop:gop-precedes to from)
	 (whats-going-on "~A precedes ~A so cannot link the other way"
			 to from)
	 nil)
	(t
	 (link-etags from to))))

;; This is called from KS-ACHIEVE for adding a new node to the network to
;; satisfy a particular condition.
(defun add-a-node (type pattern pre-list post-list)
  (let ((node (ads-add-node :type type :pattern pattern)))
    (dolist (post post-list)
      (link-etags (list node :end) post))
    (if pre-list
	(dolist (pre pre-list)
	  (link-etags
	    pre (list node :begin)))
      ;; Add a link from the end of NODE-1 (the start node) to this node, so
      ;; that all the initial effects are necessairly before.  But if a
      ;; global after-point is defined, use it instead of the end_of NODE-1.
      (link-etags
         (or (global-after-point)
	     (list 'node-1 :end))
	 (list node :begin)))
    node))

(defun add-an-initial-node (schema-node)
  (ads-add-node
    :type (node-type schema-node)
    :pattern (node-pattern schema-node)))

(defun add-time-window (from to min max)
  (let ((from-tp (or (and (eql from :ABST0)
			  (intern "PT-0"))
		     (ne-time-point (get-node-end from))))
	(to-tp (ne-time-point (get-node-end to))))
    (tpn-add-time-constraint from-tp to-tp min max)))


;;;; Alternative handling.

;;; Process-alt-db-changes is Called just after changing context for a
;;; new alternative to make changes to the database state.  It Returns
;;; nil if can't make any of the changes required, and this will then
;;; force another alternative to be chosen.

#+:undef
(defun process-alt-db-changes (alt-rec)
  (dolist (change (alt-db-changes alt-rec) t)
    (ecase (car change)
      (:BIND
       (psv-begin-var-transaction)
       (if (psv-set-value (second change) (third change))
	   (psv-commit-var-transaction)
	   (progn
	     (psv-abort-var-transaction)
	     (whats-going-on "Cannot make binding ~S = ~S"
			     (second change) (third change))
	     (return-from process-alt-db-changes nil))))
      (:RESTRICT
       (psv-begin-var-transaction)
       (if (psv-add-restriction (second change) :NOT (third change))
	   (psv-commit-var-transaction)
	   (progn
	     (psv-abort-var-transaction)
	     (whats-going-on "Cannot make restriction ~S /= ~S"
			     (second change) (third change))
	     (return-from process-alt-db-changes nil))))
      (:NOT-LINK
       #|
       (unless (qa::qa-add-to-not-link-table (second change) (third change))
	 (return-from process-alt-db-changes nil))|#))))


;;;; New domain / TF compilation

;; Holds the domain.
(defvar *domain* nil)

(defvar *loaded-TFs* nil)

(defvar *redefined-functions* nil)

(defun clear-domain-information ()
  (setq *domain* nil)
  (setq *loaded-TFs* nil)
  (dolist (f *redefined-functions*)
    (advice- f 'domain-specific))
  (setq *redefined-functions* nil))

;;; Function redefinition macro for use in included Lisp code.

(defmacro defun-for-domain (name lambda-list &body body)
  (let ((global-name (concat-name name "-DOMAIN-SPECIFIC")))
  `(progn
     (defun ,global-name ,lambda-list ,@body)
     (pushnew ',name *redefined-functions*)
     (advice+ ',name 'domain-specific
       #'(lambda (previous)
	   (declare (ignore previous))
	   #',global-name)))))

;;; Load in a domain TF descripter file.

;;; The argument will normally be a pathname.  However, we also allow it
;;; to be a namestring.  The _name_ (as opposed to the directory, type,
;;; etc) field of the pathname is significant.  It gives us the domain
;;; name and is used to determine whether the domain has already been
;;; loaded.

;;; N.B. If you compile and load a .tf file successfully, then you can't
;;; reload it after a change unless you go through an :init first.

;;; The result of init-new-domain is either nil (if something went wrong)
;;; or a list containing the domain name and the names of all available
;;; task schemas.

;;; /\/: Actually, it's the only the names of all tasks in the .tf file
;;; just processed.  See below.

(defun init-new-domain (filename) ; -> nil or (domain-name task-name...)
  (let* ((path (pathname filename))
	 (ident (pathname-name path))
	 (schema-list nil))

    ;; See if we've already loaded this domain.  Return nil if so.
    (when (member ident *loaded-TFs* :test #'string-equal)
      (dev-debug :non-fatal-error "TF for domain ~A already loaded.~%" ident)
      (return-from init-new-domain nil))

    ;; Call the TF compiler.
    (let ((defaulted-path (defaulted-tf-pathname path))
	  (*package* (find-package :oplan))
	  (tf-errors 0))
      (multiple-value-setq (*domain* tf-errors)
	(compile-domain defaulted-path))

      ;; See if any errors.  Return nil if there were.
      (when (> tf-errors 0)
	(dev-debug :fatal-error "Error in loading TF file ~A." filename)
	(return-from init-new-domain nil))

      ;; So we've just read a valid TF file.
      (dev-debug :information "Found TF file ~A~%" filename)

      ;; Evaluate any "language lisp" code.
      (when (domain-included-code *domain*)
	(mapc #'eval (domain-included-code *domain*)))

      ;; Process always facts.
      (dolist (effect (domain-always *domain*))
	(tgm-add-effect (car effect) (cadr effect) :always)
	(tgm-commit-to-adding-effect
	   (car effect) (cadr effect) :always))

      ;; Process types
      (dolist (type-spec (domain-object-types *domain*))
	(psv-add-type (car type-spec) (cadr type-spec)))

      ;; Init schemas, if any.
      (setq schema-list (domain-schemas *domain*))
      (when (consp schema-list)
	(init-schemas schema-list))

      ;; And we're done.
      (push ident *loaded-TFs*)
      (cons ident
	    (list-available-tasks)))))

(defun defaulted-tf-pathname (path)
  (let ((path-with-type (merge-pathnames path (make-pathname :type "tf"))))
    (if (find #\/ (namestring path))
	path-with-type
      (merge-pathnames
        path-with-type
	(concat-string (get-parameter :oplan-tf-dir) "/")))))

;;; "Memoized" interface to the TF compiler.

(defvar *previous-truename* nil)
(defvar *previous-write-date* nil)
(defvar *previous-domain* nil)

(defun compile-domain (pathname)
  (let* ((truename (truename pathname))
	 (write-date (file-write-date truename)))
    (cond ((and (equal *previous-truename* truename)
		(equal *previous-write-date* write-date))
	   ;; Previous result is still valid.
	   ;; We can assume 0 errors because of what we do below.
	   (MONITOR :TF
	     (:DETAILED "Reusing ~S domain." (pathname-name pathname)))
	   (values *previous-domain* 0))
	  (t
	   ;; Need to compile a new domain, so call the compiler.
	   (multiple-value-bind (domain errors)
	       (tf-compiler:compile-from-tf-file pathname)
	     (mon:monitor :TF (:DETAILED (:NO-MODULE-PREFIX) ""))
	     ;; If no errors, remember the result.
	     (when (= errors 0)
	       (setq *previous-truename* truename
		     *previous-write-date* write-date
		     *previous-domain* domain))
	     (values domain errors))))))

;;; /\/: We want the list of tasks in the order of their definition
;;; in the TF file.  However, the total schema list is not built up
;;; in *domain*.  Instead, we get a new domain struct each time we
;;; compile a TF file; and so *domain* will hold only the schemas
;;; from the most recent TF compilation.  Now, we could look in the
;;; schema-name-table, but then the order wouldn't be preserved.
;;; What we do now is let the TA keep track of the total list.

;;; Maybe this can all be fixed once we get rid of the convert-new-tf-
;;; compiler-output-to-my-old-style funciton.  /\/

(defun list-available-tasks () ; -> list of task names
  (mapcar #'schema-name
	  (remove-if-not #'(lambda (name)
			     (eql (search "task_" name) 0))
			 (domain-schemas *domain*)
			 :key #'schema-name)))

;;;; Node descriptions

;; address could be a filename, or a channel like :LEFTOUT to send the plan
;; fragment out on.
(defun print-plan-nodes (address)
  (if (stringp address)
      (progn
	(cerror "Does nothing" "Writing to a file not implemented.")
	nil)
      (ecase address
	(:DESCRIPTIONS
	 ;; For sending to the PW viewer.
	 (make-a-list-of-node-descriptions)))))

(defun make-a-list-of-node-descriptions ()
  (mapcar #'node->node-description (sort (list-nodes) #'node-lessp)))

(defun node->node-description (node)
  ;; N.B. The result must match the viewer's node-description structure.
  (let ((node-name  (n-tag node))
	(begin-pre  (mapcar #'ne-tag (ne-pre-ends (n-begin node))))
	(begin-post (mapcar #'ne-tag (ne-post-ends (n-begin node))))
	(end-pre    (mapcar #'ne-tag (ne-pre-ends (n-end node))))
	(end-post   (mapcar #'ne-tag (ne-post-ends (n-end node))))
	(node-type  (n-type node))
	(pattern    (psv:psv-actorise-pattern
		     (n-pattern node))))
    ;; Remove the links between the two ends of the node.
    (setq begin-post (remove (etag node-name :end) begin-post :test #'equal))
    (setq end-pre (remove (etag node-name :begin) end-pre :test #'equal))
    ;; Construct a description.
    (list node-name
	  begin-pre
	  begin-post
	  end-pre
	  end-post
	  (node-time-bounds node)
	  node-type
	  pattern)))

(defun node-time-bounds (node)
  (let ((begin-tp (n-begin-tpoint node))
	(end-tp (n-end-tpoint node)))
    (list
      (tpoint-min begin-tp)
      (tpoint-max begin-tp)
      (tpoint-min end-tp)
      (tpoint-max end-tp))))


;;;; GOST and TOME descriptions

(defun print-gost (address)
  (if (stringp (car address))
      (progn
	(cerror "Does nothing" "Writing to a file not implemented.")
	nil)
      (tgm-make-a-list-of-the-gost)))

(defun print-tome (address)
  (if (stringp (car address))
      (progn
	(cerror "Does nothing" "Writing to a file not implemented.")
	nil)
      (tgm-make-a-list-of-the-tome)))


;;;; QA-ALL

(defun do-qa-all (node-id)
  (let ((node-tag (intern (concatenate 'string "NODE-" node-id)
			  (find-package "OPLAN"))))
    (when (get-node node-tag)
      ;; OK -- the node exists.
      (qa::qa-all-at-node `(,node-tag :end)))))


;;;; Whats-going-on output.

;; Uses whats-going-on to stuff out the Plan State in the current context.
(defun print-database ()
  (unless *print-whats-going-on-p*
    (return-from print-database))
  (whats-going-on "Context Num = ~S" (contextnum *cucontext*))
  (whats-going-on "The Triggered Agenda Table:~%~{~S~%~}" 
		  (ctxt-symbol-value '*triggered-agenda-table*))
  (whats-going-on "The Untriggered Agenda Table:~%~{~S~%~}"
		  (ctxt-symbol-value '*untriggered-agenda-table*))
  (whats-going-on "The Alternative Agenda Table:~%~{~S~%~}"
		  *alternatives-table*)
  (whats-going-on "The TOME:~%~{~S~%~}" (tgm-make-a-list-of-the-tome t nil))
  (whats-going-on "The GOST:~%~{~S~%~}" (tgm-make-a-list-of-the-gost t nil))
  (whats-going-on "The PSVs:~%~{~A~%~}" (psv-make-a-list-of-the-psvs))
  (whats-going-on "Nodes:~%~{~S~%~}" (make-a-list-of-node-descriptions)))

;;; ---------------------------- Change History ----------------------------
;;; (Who)   (When) 		     (What)
;;;
;;;  JWD   March 94	Reorganization and cleanup.  Much rewriting.
;;;
