;;;; File: nodes.lsp
;;; SCCS Version: %W%
;;; Contains: Definition of nodes and node-ends
;;; Author: Jeff Dalton
;;; Created: 02 March 1994
;;; Updated: Thu Jul  6 17:58:05 1995 by Jeff Dalton
;;; Release Version: %Y%
;;; Copyright: (c) 1994, AIAI, University of Edinburgh
;;; This material may be reproduced by or for the U.S. Government pursuant
;;; to the copyright license under the clause at DFARS 252.227-7032
;;; (June 1975) -- Rights in Technical Data and Computer Software (Foreign).

;;; Both nodes and node-ends are represented explicitly by structs.
;;; In addition, there is a time-point for every node-end (though not
;;; necessarily a node-end for every time-point).

;;; Some parts of O-Plan work only on node ends and don't need to
;;; know about nodes at all.

(in-package :datastructures :nicknames '(:ads))

(use-package :util)
(use-package :ctxt)
(use-package :schema-defs)

;; Nodes
(export '(*estimated-max-nodes*))
(export '(ads-node n-tag n-parent-tag n-child-tags n-numbers
	  n-type n-pattern n-begin n-end))
(export '(get-node))
(export '(n-begin-tpoint n-end-tpoint))

;; Node-end tags
(export '(etag make-etag etag-node etag-end))

;; Node-ends
(export '(ads-node-end ne-tag ne-pre-ends ne-post-ends
	  ne-time-point ne-index))
(export '(ne-node-tag ne-node ne-end))
(export '(get-node-end get-named-end))
(export '(node-end-count))

;; Basic operations on nodes
(export '(walk-nodes list-nodes node-lessp))
(export '(walk-node-ends list-node-ends))

;; Global after-point
(export '(global-after-point set-global-after-point))

;; Node construction and linking
(export '(ads-add-node make-node-and-ends))
(export '(link-etags link-node-ends))
(export '(link-etags-else-error link-node-ends-else-error))

;; Extracting information
(export '(walk-node-ends-depth-first node-ends-between))

;; Descriptions
(export '(make-a-list-of-node-descriptions))

;; Support for KS requests
(export '(ads-expand-node))

;; /\/: Backwards compatibility
(export '(ds-add-plan-manager-network-node
	  ds-make-a-list-of-nodes
	  ds-init

	  ;; Constructor and field accessor fns for the node-end structure.
	  make-ds-node-end
	  ds-node-tag
	  ds-node-end))
	  

;;; Globals

(defparameter *estimated-max-nodes* 500
  "A plausible upper bound on the number of nodes, used when
   constructing arrays and hash-tables.")

(defvar *node-table* (make-hash-table :size *estimated-max-nodes*
				      :test #'eq)
  "A mapping from node-tags to node structs [context-layered].")

(defvar *node-end-count* 0
  "The number of existing node-ends [context-layered].")

(defvar *global-after-point* nil
  "A node-end tag that all new nodes must be liked after [context-layered].")

(defun init-nodes-db ()
  (clrhash *node-table*)
  (setq *node-end-count* 0)
  (setq *global-after-point* nil)
  t)


;;;; Structures

;;; /\/: The "ads-" prefix is chiefly to distinguish names defined here
;;; from names defined elsewhere.  For instance, the TF compiler defines
;;; a "node" struct, and schemadef.lsp defines some "node-" accessors.
;;; Perhaps eventually this file should own "node", but at present that's
;;; not the case.  Similar problems arise for "node-end", which sometimes
;;; means a node-end tag such as (node-1 :end).

;;; /\/: Maybe use "plan-" instead of "ads-".

;;; Nodes

(defstruct (ads-node (:conc-name n-))
  tag				;symbol, e.g. node-3-2
  numbers			;numbers from tag, e.g. (3 2) 
  type				;node type
  pattern			;description, typically an action pattern
  parent-tag			;symbol, e.g. node-3
  child-tags			;list of symbols, e.g. (node-3-2-1 ...)
  begin				;node-end that represents the begin_of the node
  end)				;node-end that represents the end_of the node

(defun get-node (node-tag)
  (or (derefincontext (gethash node-tag *node-table*))
      (error "There is no node with tag ~S." node-tag)))

(defsetf get-node (node-tag) (new-node-struct)
  `(updateincontext
     #'(lambda (k v) (setf (gethash k *node-table*) v))
     #'(lambda (k) (gethash k *node-table*))
     ,node-tag
     ,new-node-struct))


;;; Node-end tags

;;; /\/: Again we have to pick a name that won't conflict.
;;; /\/: Instead of e.g. (node-3 :begin) should we have (:begin_of node-3)?

;;; N.B. Must be comparable by EQUAL.  (/\/ really?)

(defstruct (end-tag (:conc-name etag-)
		    (:constructor make-etag)
		    (:constructor etag (node end))
		    (:type list))
  node				;node tag, e.g. node-1
  end)				;end, either :begin or :end


;;; Node-ends

(defstruct (ads-node-end (:conc-name ne-)
			 (:print-function print-ads-node-end))
  tag				;tag + end of the node it's an end of
  i-pre-ends			;node ends linked directly before this one
  i-post-ends			;node ends linked directly after this one
  time-point			;time point that corresponds to this node-end
  (index (next-ne-index)	;index in any associated arrays
	 :type fixnum))

(define-context-accessor ne-pre-ends ne-i-pre-ends)
(define-context-accessor ne-post-ends ne-i-post-ends)

(defun print-ads-node-end (ne stream depth)
  ;; This :print-function exists so that we don't have to worry about
  ;; circularity when printing node-ends.
  (declare (ignore depth))
  (format stream "#<~A ~S ~D>" (type-of ne) (ne-tag ne) (ne-index ne)))

;;; Pseudo-accessors

(defun ne-node-tag (ne) (etag-node (ne-tag ne)))
(defun ne-node (ne)     (get-node (etag-node (ne-tag ne))))
(defun ne-end (ne)      (etag-end (ne-tag ne)))

;;; Node pseudo-accessors (here to be after the node-end defstruct
;;; and hence able to have ne- accessors compiled inline).

(defun n-begin-tpoint (n) (ne-time-point (n-begin n)))
(defun n-end-tpoint (n)   (ne-time-point (n-end n)))

;;; Node-end lookup

(defun get-node-end (etag)
  (let ((n (get-node (etag-node etag))))
    (ecase (etag-end etag)
      (:begin (n-begin n))
      (:end (n-end n)))))

(defun get-named-end (node end)
  (ecase end
    (:begin (n-begin node))
    (:end (n-end node))))

;;; Index / count operations

(defun next-ne-index ()
  ;; N.B. Call this whenever a node-end is constructed.
  (let ((i (ctxt-symbol-value '*node-end-count*)))
    (assert (<= i most-positive-fixnum))
    (incf (ctxt-symbol-value '*node-end-count*))
    i))

(defun node-end-count ()
  (ctxt-symbol-value '*node-end-count*))


;;;; Basic operations on nodes

(defun walk-nodes (fn)
  (maphash #'(lambda (k v)
	       (declare (ignore k))
	       (let ((node (derefincontext v)))
		 (unless (null node)
		   (funcall fn node))))
	   *node-table*))

(defun list-nodes ()
  (let ((nodes '()))
    (walk-nodes #'(lambda (n) (push n nodes)))
    nodes))

(defun node-lessp (node-1 node-2)	;for sorting lists of nodes
  (list-lessp #'< (n-numbers node-1)
	          (n-numbers node-2)))

(defun walk-node-ends (fn)
  (walk-nodes
    #'(lambda (n)
	(funcall fn (n-begin n))
	(funcall fn (n-end n)))))

(defun list-node-ends ()
  (let ((node-ends '()))
    (walk-node-ends #'(lambda (ne) (push ne node-ends)))
    node-ends))


;;;; Global after-point

(defun global-after-point ()
  (ctxt-symbol-value '*global-after-point*))

(defun set-global-after-point (etag)
  (setf (ctxt-symbol-value '*global-after-point*)
	etag))


;;;; /\/: Backwards-compatibility

(defparameter *hash-table-size* 500
  "Initial size of the various hash tables.")

(defstruct (ds-node-end (:conc-name ds-) (:type list))
  "Data structure for describing the end of a plan manager network node.
     node-tag - The node tag.
     node-end - Either :begin or :end."
  node-tag
  node-end)

(defun ds-init ()
  "Resets all the hash tables."
  (tpn-init-tpn)
  (init-nodes-db))


;;;; Node construction

;;; Make-node-and-ends constructs a node, its two node-ends, and the
;;; two time-points associated with the node-ends.  The references back
;;; and forth between these structures are put in place; and the begin
;;; node-end is linked before the end node-end, both at the node-end
;;; (ADS) level and by a time-constraint in the TPN.

;;; /\/: The :PARENT argument is the parent's tag (or nil if no parent),
;;; rather than the parent node struct itself.  Maybe this should change.

(defun ads-add-node (&rest args)
  (n-tag (apply #'make-node-and-ends args)))

(defun make-node-and-ends
    (&key type
	  pattern
	  parent
	  delay
	  ) ; -> node
  (assert (null delay))			;/\/ for now
  ;; First get a tag
  (multiple-value-bind (node-tag num)
      (pop-gensym (if parent (concat-string parent "-") "NODE-"))
    ;; Now build structures, bottom-up
    (let* ((begin-tpoint-tag (tpn-add-time-point 0 *infinity*))
	   (end-tpoint-tag (tpn-add-time-point 0 *infinity*))
	   (begin-end
	    (make-ads-node-end
	      :tag        (etag node-tag :begin)
	      :time-point (tpn-get-tpoint begin-tpoint-tag)
	      :index      (next-ne-index)))
	   (end-end
	    (make-ads-node-end
	      :tag        (etag node-tag :end)
	      :time-point (tpn-get-tpoint end-tpoint-tag)
	      :index      (next-ne-index)))
	   (numbers
	    (append (and parent (n-numbers (get-node parent)))
		    (list num)))
	   (node
	    (make-ads-node
	      :tag        node-tag
	      :numbers    numbers
	      :type       type
	      :pattern    pattern
	      :parent-tag parent
	      :begin      begin-end
	      :end        end-end
	      )))
      ;; Record the node in the node-table.  Note that until we do this
      ;; we can't reach the node from its node-ends, because node-ends don't
      ;; contain a pointer to their node.  They each have a tag that contains
      ;; the node's tag instead.
      (setf (get-node node-tag) node)
      ;; The begin-end must be before the end-end.  Moreover, linking
      ;; them cannot fail.
      (link-node-ends-else-error begin-end end-end delay)
      ;; And we're done.
      node)))


;;;; Adding links

;;; Link-node-ends constrains one node-end to be before another in time.
;;; If the TPN says this is impossible, it returns false without changing
;;; the plan state.  Otherwise, it adds a constraint to the TPN, links
;;; the node-end structs via their pre- and post-ends slots, and returns
;;; true.

(defun link-etags (from-tag to-tag &optional delay-between)
  (link-node-ends (get-node-end from-tag) (get-node-end to-tag) delay-between))

(defun link-node-ends (from to &optional delay-between) ; -> true / false
  (assert (null delay-between))			;/\/ for now

  ;; Constrain the associated time-points.
  (unless
      (tpn-add-time-constraint
	  (ne-time-point from)
	  (ne-time-point to)
	  0
	  *infinity*)
    (return-from link-node-ends nil))	;indicate failure

  ;; Link the node-ends.
  (push to (ne-post-ends from))		;from -> to
  (push from (ne-pre-ends to))		;from <- to

  ;; Success
  t)

(defun link-etags-else-error (from-tag to-tag &optional delay-between)
  (or (link-etags from-tag to-tag delay-between)
      (cerror "Continue anyway."
	      "Can't link ~S --> ~S, delay ~S."
	      from-tag to-tag delay-between)))

(defun link-node-ends-else-error (from-ne to-ne &optional delay-between)
  (or (link-node-ends from-ne to-ne delay-between)
      (cerror "Continue anyway."
	      "Can't link ~S --> ~S, delay ~S."
	      from-ne to-ne delay-between)))


;;;; Following links

;;; (walk-node-ends-depth-first fn &key from)
;;;
;;; Calls the function fn on node-end structs that are linked after
;;; (the ne struct) from, proceeding depth-first along ne-post-ends
;;; links.  The descent continues from a node-end only if fn returns
;;; true.  Node-ends are visited at most once, and an error is signalled
;;; if a cycle is found.
;;;
;;; Walk-node-ends-depth-first will visit node-ends in a less random
;;; order than (the present version of) walk-node-ends and may even
;;; be faster.

(defun walk-node-ends-depth-first (fn &key (from (earliest-node-end)))
  (let ((marks (make-node-end-mark-vector)))
    (declare (type simple-vector marks))
    (macrolet ((mark (ne) `(svref marks (ne-index ,ne))))
      (label walk ((at from))
	(ecase (mark at)
	  ((:start)
	   (error "Node end cycle at ~S." at))
	  ((:finish)
	   ;; Already been here
	   )
	  ((nil)
	   ;; A node-end we haven't yet visited.
	   (setf (mark at) :start)
	   (let ((descend-p (funcall fn at)))
	     (when descend-p
	       (dolist (successor (ne-post-ends at))
		 (walk successor))))
	   (setf (mark at) :finish)))))))

(defun make-node-end-mark-vector ()
  (make-array (node-end-count) :initial-element nil))

(defun earliest-node-end ()
  (get-node-end
    (etag (intern (string '#:node-1) :oplan)
	  :begin)))


;;; (node-ends-between before after)
;;;
;;; Returns a list of the node-end structs that are linked after "after"
;;; and before "before" (and are hence between them).
;;;
;;; "Before" and "after" are ne structs.  Neither will be part of the
;;; result.  "Before" will normally be linked (directly or indirectly)
;;; before "after", but that's not necessary (you'll just get a null
;;; result).  The strategy is to walk the successors of "before",
;;; collecting those that are on paths that lead to "after".

(defun node-ends-between (before after) ; -> list of ne-structs
  (let ((marks (make-hash-table :test #'eq))
	(result '()))
    (macrolet ((mark (node) `(gethash ,node marks)))
      (setf (mark before) :looking)
      (setf (mark after) :on-path)
      (label walk ((ends (ne-post-ends before)))
	(dolist (at ends)
	  (ecase (mark at)
	    ((:looking)
	     (error "Node end cycle at ~S." at))
	    ((:on-path :off-path)
	     ;; Already processed
	     )
	    ((nil)
	     (let ((post-ends (ne-post-ends at)))
	       (walk post-ends)
	       (cond ((loop for p in post-ends
			    thereis (eq (mark p) :on-path))
		      (setf (mark at) :on-path)
		      (push at result))
		     (t
		      (setf (mark at) :off-path)))))))))
    result))


;;;; Expansion

(defvar *expand-subnodes*)

;;; Returns a list of (tag type pattern) descriptions of all nodes
;;; created.

(defun ads-expand-node (node-tag schema-subnodes schema-orderings)

  ;; First create all the subnodes
  (let* ((*expand-subnodes*
	  '())
	 (number->subtag-alist
	  (mapcar #'(lambda (sub)
		      (cons (node-number sub)
			    (n-tag (make-schema-subnode node-tag sub))))
		  schema-subnodes))
	 (node
	  (get-node node-tag)))

    ;; Create links between the various subnodes, if given.
    (flet ((number->subtag (number)
	     (or (lookup number number->subtag-alist)
		 (error "Node ~S has no subnode numbered ~S."
			node-tag number))))
      (dolist (ord schema-orderings)
	(link-etags-else-error
	  (etag (number->subtag (ordering-node-number1 ord))
		(ordering-node-end1 ord))
	  (etag (number->subtag (ordering-node-number2 ord))
		(ordering-node-end2 ord)))))

    ;; /\/: Reverse alist so links to/from the parent will occur
    ;; in the same order they used to.
    (setq number->subtag-alist (nreverse number->subtag-alist))

    ;; Make sure all the subnodes are linked in somewhere.
    (dolist (number+subtag number->subtag-alist)
      (let* ((subtag (cdr number+subtag))
	     (subnode (get-node subtag)))

	;; If pre field of a subnode is not set yet, then create a
	;; link from the start of -node- to the start of -subnode-.
	(unless (ne-pre-ends (n-begin subnode))
	  (link-node-ends-else-error
	    (n-begin node) (n-begin subnode)))

	;; If post field of a subnode is not set yet, then create a link
	;; from the end of -subnode- to the end of -node-.
	(unless (ne-post-ends (n-end subnode))
	  (link-node-ends-else-error
	    (n-end subnode) (n-end node)))))

    ;; Finished!
    (setq *expand-subnodes* (nreverse *expand-subnodes*)) ; /\/
    (mapcar #'(lambda (n)
		(list (n-tag n) (n-type n) (n-pattern n)))
	    *expand-subnodes*)))

(defun make-schema-subnode (parent-tag sub)
  (if (node-iterators sub)
      (iterate-schema-subnode parent-tag sub)
    (let ((n (make-node-and-ends
	       :type (node-type sub)
	       :pattern (node-pattern sub)
	       :parent parent-tag)))
      (push n *expand-subnodes*)
      n)))

;;; ---------------------------- Change History ----------------------------
;;; (Who)   (When) 		     (What)
;;;
