;;;; File: schemacode.lsp
;;;; SCCS Version: %W%
;;;; Contains: Schema handling code.
;;;; Author: Richard Kirby (rbk)
;;;; Created: Mon Aug 31 14:12:21 1992
;;;; Updated: Thu Aug 11 17:05:17 1994 by Jeff Dalton
;;;; Release Version: %Y%
;;;; Copyright: (c) 1992, AIAI, University of Edinburgh
;;;; This material may be reproduced by or for the U.S. Government pursuant
;;;; to the copyright license under the clause at DFARS 252.227-7032
;;;; (June 1975) -- Rights in Technical Data and Computer Software (Foreign).

(in-package :oplan)

;; Loads in some more schema definitions into the schema tables.
;; Use the fixed first word of a pattern as the key.
(defun init-schemas (schema-defs)
  (let (s)
    (dolist (schema schema-defs)
      (if (schema-expands schema)
	  (let ((fixed-word (first (schema-expands schema))))
	    (if (setq s (gethash fixed-word *schema-expansion-table*))
		(if (listp s)
		    (setf (gethash fixed-word *schema-expansion-table*)
			  (nconc s (list schema)))
		    (setf (gethash fixed-word *schema-expansion-table*)
			  (list s schema)))
		(setf (gethash fixed-word *schema-expansion-table*)
		      (cons schema nil)))))
      (if (schema-only-use-for-effects schema)
	  (dolist (effect (schema-only-use-for-effects schema))
	    (let ((fixed-word (first (car effect))))
	      (if (setq s (gethash fixed-word *schema-effects-table*))
		  (if (listp s)
		      (setf (gethash fixed-word *schema-effects-table*)
			    (nconc s (list (cons effect schema))))
		      (setf (gethash fixed-word *schema-effects-table*)
			    (list s (cons effect schema))))
		  (setf (gethash fixed-word *schema-effects-table*)
			(list (cons effect schema)))))))
      ;; Should only have one schema per name, so should we replace, or give
      ;; a message to indicate that we have a schema with the same name as
      ;; another.
      (if (setq s (gethash (schema-name schema) *schema-name-table*))
	  (break "SCHEMA ~S already present." (schema-name schema))
	  (setf (gethash (schema-name schema) *schema-name-table*) schema))))
  (build-possible-effects-table)
  (build-expansions-and-achieves-tables))

(defun find-schemas-matching-expansion-pattern (pattern)
  ;; /\/: Try this:
  (setq pattern (psv-actorise-pattern pattern))
  ;; /\/
  (let ((possibles (gethash (first pattern) *schema-expansion-table*))
	result matchup)
    (dolist (v possibles)
      (when (setq matchup
		  (match-up pattern (schema-expands v)
			    (create-binding-tuples-from-schema v)))
	;; Because t is not a valid binding list.
	(if (eql matchup t) (setq matchup nil))
	(setq result (cons (cons v matchup) result))))
    (unless result
      (warn #|break|# "NO SCHEMA TO EXPAND ~S" pattern))
    (nreverse result)))

(defun find-schemas-for-using-for-effect (effect)
  (let* ((pattern (first effect))
	 (value (second effect))
	 (possibles (gethash (first pattern) *schema-effects-table*))
	 result matchup)
    ;; /\/: Try this:
    (setq pattern (psv-actorise-pattern pattern))
    (setq value (psv-actorise-pattern value))
    ;; /\/
    (dolist (v possibles)
      (let ((oufe (car v))
	    (schema (cdr v)))
	(if (setq matchup
		  (match-up pattern (effect-pattern oufe)
			    (create-binding-tuples-from-schema schema)))
	    (when (setq matchup
			(match-up value (effect-value oufe) matchup))
	      ;; Because t is not a valid binding list.
	      (if (eql matchup t) (setq matchup nil))
	      (setq result (cons (cons schema matchup) result))))))
    (unless result
      (warn #|break|# "NO SCHEMA FOR EFFECT ~S" effect))
    (nreverse result)))

(defun find-schema-with-name (name)
  (gethash name *schema-name-table*))

(defvar *schema-possible-effects-table* (make-hash-table :test #'equal)
  "Table of schema name as key and a list of the first fixed word of all the
   possible effects that could be added from using the schema and all its 
   implied expansions.")

(defun build-possible-effects-table ()
  (let (schemas reconsider-list)
    ;; Construct a list of all the schema names.
    (maphash #'(lambda (k v) (declare (ignore v)) (push k schemas))
	     *schema-name-table*)
    (labels ((construct-effects-list (schema-name)
	       (let ((schema (gethash schema-name *schema-name-table*))
		     effects)
		 (unless (gethash schema-name *schema-possible-effects-table*)
		   ;; Mark this schema as being considered.
		   (setf (gethash schema-name *schema-possible-effects-table*)
			 t)
		   ;; Get all the first fixed words of all the oufe and effects
		   ;; of this schema.
		   (dolist (effect (schema-only-use-for-effects schema))
		     (push (first (effect-pattern effect)) effects))
		   (dolist (effect (schema-effects schema))
		     (push (first (effect-pattern effect)) effects))
		   ;; Now consider each action node of this schema, and look
		   ;; at the schemas that could be used to expand the action.
		   (dolist (node (schema-nodes schema))
		     (dolist (expansion-schema
			       (gethash (first (node-pattern node))
					*schema-expansion-table*))
		       (let ((expansion-schema-effects
			      (gethash (schema-name expansion-schema)
				       *schema-possible-effects-table*)))
			 (if (consp expansion-schema-effects)
			     (setq effects
				   (append expansion-schema-effects effects))
			     (if (eql expansion-schema-effects t)
				 (pushnew schema-name reconsider-list
					  :test #'equal)
				 (construct-effects-list
				  (schema-name expansion-schema)))))))
		   ;; Now consider each achievable condition of this schame,
		   ;; and look at the schemas that could be used to achieve the
		   ;; condition.
		   (dolist (cond (schema-conditions schema))
		     (if (eql (con-type cond) 'achievable)
			 (dolist (oufe-schema
				   (gethash (first (con-pattern cond))
					    *schema-effects-table*))
			   (let ((oufe-schema-effects
				  (gethash (schema-name (cdr oufe-schema))
					   *schema-possible-effects-table*)))
			     (if (consp oufe-schema-effects)
				 (setq effects
				       (append oufe-schema-effects effects))
				 (if (eql oufe-schema-effects t)
				     (pushnew schema-name reconsider-list
					      :test #'equal)
				     (construct-effects-list
				      (schema-name (cdr oufe-schema)))))))))
		   ;; We should now have atleast a partial list of effects for
		   ;; this schema, so set the entry.
		   (setf (gethash schema-name *schema-possible-effects-table*)
			 (remove-duplicates effects))))))
      (loop
       (if (null schemas)
	   (if (null reconsider-list)
	       (return)
	       (setq schemas reconsider-list reconsider-list nil)))
       (construct-effects-list (pop schemas))))))

(defvar *possible-effects-from-expansions-table* (make-hash-table)
  "This table maps the first fixed-word of a pattern for expansion to a list
   of all the possible effects that could be added from this expansion.")

(defvar *possible-effects-from-achieves-table* (make-hash-table)
  "This table maps the first fixed-word of a pattern for satisfying an
   achieve, to a list of all the possible effects that could be added
   from adding to the plan network.")

(defun build-expansions-and-achieves-tables ()
  (maphash #'(lambda (k v)
	       (let (effects)
		 (dolist (schema v)
		   (setq effects
			 (append (gethash (schema-name schema)
					  *schema-possible-effects-table*)
				 effects)))
		 (setf (gethash k *possible-effects-from-expansions-table*)
		       (remove-duplicates effects))))
	   *schema-expansion-table*)
  (maphash #'(lambda (k v)
	       (let (effects)
		 (dolist (schema v)
		   (setq effects
			 (append (gethash (schema-name (cdr schema))
					  *schema-possible-effects-table*)
				 effects)))
		 (setf (gethash k *possible-effects-from-achieves-table*)
		       (remove-duplicates effects))))
	   *schema-effects-table*))

(defun get-possible-effects-from-expanding (pattern)
  (gethash (car pattern) *possible-effects-from-expansions-table*))

(defun get-possible-effects-from-achieving (pattern)
  (gethash (car pattern) *possible-effects-from-achieves-table*))

(defun print-possible-effects-table (&optional (stream t))
  (format stream "~&Possible effects table:~%")
  (let ((*print-level* nil) (*print-length* nil))
    (maphash #'(lambda (k v)
		 (format stream "~S: ~S~%" k v))
	     *schema-possible-effects-table*))
  (terpri stream))

(defun print-schema-tables (&optional (stream t))
  (let ((*print-level* nil) (*print-length* nil))
    (format stream "~&Expands schemas:~%")
    (maphash #'(lambda (k v)
		 (format stream "~S: ~S~%" k v))
	     *schema-expansion-table*)
    (format stream "~%OUFE schemas:~%")
    (maphash #'(lambda (k v)
		 (format stream "~S: ~S~%" k v))
	     *schema-effects-table*)))

(defun clear-schema-possible-effects-tables ()
  (clrhash *schema-possible-effects-table*)
  (clrhash *possible-effects-from-expansions-table*)
  (clrhash *possible-effects-from-achieves-table*))

;;;; ---------------------------- Change History ----------------------------
;;;; (Initials) (Date of change)           (Comment)
