;;;; File: embedded-tfc.lsp
;;; SCCS Version: %W%
;;; Contains: Top-level for the embedded TF syntax checker (tfc)
;;; Author: Jeff Dalton <J.Dalton@ed.ac.uk>
;;; Created: 03 March 1994
;;; Updated: Tue Aug 30 17:35:17 1994 by Jeff Dalton
;;; Release Version: %Y%
;;; Copyright: (c) 1994, AIAI, University of Edinburgh
;;; This material may be reproduced by or for the U.S. Government pursuant
;;; to the copyright license under the clause at DFARS 252.227-7032
;;; (June 1975) -- Rights in Technical Data and Computer Software (Foreign).

(in-package :tf-compiler)

;;; Tfc is a TF syntax-checker ("c" for "check").  Actually, it calls
;;; the full TF compiler so that you get "semantic" checks as well.
;;; Tfc does not normally have an executable image of it's own.
;;; Instead, it's built into the O-Plan image and is invoked by the
;;; "-tfc" command-line argument.  All arguments after "-tfc" are
;;; interpreted by Tfc, not by  O-Plan.
;;;
;;; Tfc can therefore be called from the Unix shell by one of the following:
;;;
;;;   oplan -tfc [-q] [-l] filename
;;;   oplan -tfc [-q] [-l] -d directory
;;;
;;; The "-q" is for "quiet" and suppresses progress reports.
;;; The "-l" is for "levels" and produces a level analysis.
;;;
;;; If the filename is the single character "-", TF forms will be
;;; read from standard-input.  Progress reports will be turned off
;;; automatically.
;;;
;;; If no additional arguments are supplied after "-tfc", Tfc
;;; enters a loop in which you can type Tfc commands.  E.g
;;;
;;;    shell% oplan -tfc
;;;    tfc> -q -l snark.tf
;;;    tfc>
;;;
;;; To exit, type "q" (not "-q") as a command.
;;;
;;; There's also a tfc function that can be called from the O-Plan
;;; Lisp interaction window.  /\/: Does this work?
;;;

;;; Entry point from the O-Plan top-level

(defun run-as-tfc (arg-offset)
  ;; First some required setup
  (monitors:mon-init-db)
  ;; Now we can actually run
  (let ((args (argv->arglist arg-offset)))
    (if args
	(apply #'tfc args)
	(tfc-top-level))
    (exit-lisp 0)))

(defun argv->arglist (offset)
  ;; /\/: Can't use argc, because Lucid doesn't have one.
  (loop for i from offset while (argv i)
	collect (argv i)))


;;; A top-level

(defun tfc-top-level ()
  (loop
    (princ "tfc> ")
    (let ((args (break-args (read-line))))
      (when (string= (first args) "q")
	(return))
      (apply #'tfc args))
    (terpri)))


;;; The tfc function

(defun tfc (&rest args)
  (let ((*running-as-tfc* t)		;if anyone cares
	(*print-case* :downcase)
	(*print-pretty* t)
	(*package* (find-package :tf-compiler))
	(*progress-reports* t)
	(*level-report* nil)
	(directory nil)
	(filename nil)
	(stdin nil)
	(i 0))

    ;; Process args
    (labels ((arg () (nth i args))
	     (pop-arg () (prog1 (arg) (incf i))))
      (when (string-equal (arg) "-q")
	(setq *progress-reports* nil)
	(pop-arg))
      (when (string-equal (arg) "-l")
	(setq *level-report* t)
	(pop-arg))
      (cond ((string-equal (arg) "-d")
	     (pop-arg)
	     (setq directory (pop-arg)))
	    ((string-equal (arg) "-")
	     (pop-arg)
	     (setq *progress-reports* nil)
	     (setq stdin t))
	    (t
	     (setq filename (pop-arg))))
      (unless (null (arg))
	(usage-error)
	(return-from tfc)))

    ;; Do as requested.
    (cond (directory (tfc-directory directory))
	  (filename (tfc-file filename))
	  (stdin (tfc-stdin))
	  (t (usage-error)))
    
    (values)))

(defun usage-error ()
  (tfc-error
    "Usage: tfc [-q] [-l] filename~%~
        ~6t tfc [-q] [-l] -d directory~%"))

(defun tfc-error (message &rest format-args)
  (format *error-output* "~&~?~%" message format-args))

(defun tfc-file (filename)
  (let ((path (merge-pathnames filename (make-pathname :type "tf"))))
    (if (probe-file path)
	(multiple-value-bind (result n-errors)
	    (compile-from-tf-file path)
	  (declare (ignore result))
	  (fresh-line)
	  (when (> n-errors 0)
	    (tfc-error "~D Errors." n-errors)))
      (tfc-error "Can't find file ~S" (namestring path)))))

(defun tfc-stdin ()
  (multiple-value-bind (result n-errors)
      (compile-tf-stream *standard-input*)
    (declare (ignore result))
    (fresh-line)
    (when (> n-errors 0)
      (tfc-error "~D Errors." n-errors))))

(defun tfc-directory (directory)
  (let ((path (pathname (directory-name directory))))
    (if (probe-file path)
	(compile-test-files path :messages t)
      (tfc-error "Can't find directory ~S" (namestring path)))))

(defun directory-name (name)
  (if (eql (char name (1- (length name))) '#\/)
      name
    (concatenate 'string name "/")))

;;; End
