;;;; File: levels.lsp
;;; SCCS Version: %W%
;;; Contains: Action / effect level algorithms and analysis
;;; Author: Jeff Dalton <J.Dalton@ed.ac.uk>
;;; Created: Thu Sep 30 15:49:01 1993 by Jeff Dalton
;;; Updated: Tue Jul  4 00:06:42 1995 by Jeff Dalton
;;; Release Version: %Y%
;;; Copyright: (c) 1993, 1994, AIAI, University of Edinburgh
;;; This material may be reproduced by or for the U.S. Government pursuant
;;; to the copyright license under the clause at DFARS 252.227-7032
;;; (June 1975) -- Rights in Technical Data and Computer Software (Foreign).

(in-package :tf-compiler)

;;; /\/: Always effects and task effects both get level 0.
;;; /\/: What if we add a new task after we've got a plan?  Presumably
;;;      it can introduce effects at levels we thought were finished.
;;; /\/: We need a better way to collect lists of unique values.
;;;      "Dynamic sets"?  How much time are we spending in duplicate
;;;      elimination anyway?
;;; /\/: Schema-names should be symbols.

;;;; Utilities

;;; We want to define a mapping from action names to successor action
;;; names.  This turns out to be a pain.

(defun get-schema-action-name (schema)
  (if (schema-expands schema)
      (car (schema-expands schema))
      (intern				;/\/ which package?
       (string-upcase (schema-name schema)))))

(defun schema-node-action-names (schema)
  ;; Collect action-names from the action nodes of the schema.
  (delete-duplicates
   (loop for n in (schema-nodes schema)
	 when (eq (node-type n) 'action)
	 collect (car (node-pattern n)))))

;;; From here we assume *domain* is the domain to use.

;;; /\/: But since it's only a couple o' functions, maybe we should
;;; pass the extra parameter instead.

(defun action-schemas (action-name) ; -> list of schemas
  (loop for s in (domain-schemas *domain*)
	when (eq (get-schema-action-name s) action-name)
	collect s))

;;; /\/: It will be better to build a table directly, looping through
;;; the schemas only once (at least for the schema-node actions).
;;; Having a o_u_f_e to schema mapping would also be useful.

(defun build-action-successor-table (domain)
  (macrolet ((next-actions (name) `(gethash ,name next-action-table)))
    (let ((next-action-table (make-hash-table :test #'eq))
	  (oufe-to-schema-table
	   (transpose-graph
	    (domain-schemas domain)
	    #'(lambda (schema)
		;; Just the oufe effect names
		(mapcar #'(lambda (effect) (car (effect-pattern effect)))
			(schema-only-use-for-effects schema))))))
      (dolist (s (domain-schemas domain))
	(let ((s-a (get-schema-action-name s)))
	  ;; Make sure all action names are in the table,
	  ;; even if they have no successors.
	  (setf (next-actions s-a) (next-actions s-a))
	  ;; Each schema provides links from its action-name to the
	  ;; names that begin the patterns of its action nodes.
	  (dolist (n-a (schema-node-action-names s))
	    (pushnew n-a (next-actions s-a)))
	  ;; Now for links via achieve / oufe.
	  (dolist (c (schema-conditions s))
	    (when (or (eq (con-type c) 'achieve)
		      (eq (con-type c) 'achievable)) ;for now /\/
	      (dolist (oufe-s (gethash (car (con-pattern c))
				       oufe-to-schema-table))
		(pushnew (get-schema-action-name oufe-s)
			 (next-actions s-a)))))))
      ;; nreverse the values
      (maphash #'(lambda (k v)
		   (setf (next-actions k) (nreverse v)))
	       next-action-table)
      next-action-table)))


;;; More utilities...

(defun schema-effect-names (schema)
  (delete-duplicates
   (mapcar #'(lambda (effect) (car (effect-pattern effect)))
	   (append (schema-effects schema)
		   (schema-only-use-for-effects schema)))))

(defun action-effect-names (action-name)
  (delete-duplicates
   (loop for s in (action-schemas action-name)
	 append (schema-effect-names s))))

(defun domain-action-names (domain)
  (let ((actions '()))
    (dolist (s (domain-schemas domain))
      (pushnew (get-schema-action-name s) actions))
    ;; Check that all action nodes are represented
    (let ((unexpandable-actions '()))
      (dolist (s (domain-schemas domain))
	(dolist (a (schema-node-action-names s))
	  (unless (member a actions)
	    (pushnew a unexpandable-actions))))
      (values
       (nreverse actions)
       (nreverse unexpandable-actions)))))

(defun domain-effect-names (domain)
  (let ((effects '()))
    (dolist (p-v (domain-always domain))
      (pushnew (car (pv-pattern p-v)) effects))
    (dolist (s (domain-schemas domain))
      (dolist (e (schema-effect-names s))
	(pushnew e effects)))
    (nreverse effects)))


;;; Still more utilities...

(defun task-schema-p (schema)
  (eql (search "task_" (schema-name schema))
       0))

(defun domain-task-schemas (domain)
  (remove-if-not #'task-schema-p (domain-schemas domain)))

(defun domain-root-actions (domain)
  (mapcar #'get-schema-action-name (domain-task-schemas domain)))


;;; And finally: levels!

(defun build-level-tables (domain)
  (let* ((*domain* domain)
	 (root-actions (domain-root-actions domain))
	 (action-successor-table (build-action-successor-table domain))
	 (all-actions (hash-table-keys action-successor-table)))
    (multiple-value-bind
	  (components
	   component-successor-table
	   action-to-component-table)
	(build-component-graph root-actions
			       (hash-table->function action-successor-table))
      ;; All-actions will include any actions that are not reachable from
      ;; the root actions.
      (let* ((action-levels
	      (build-action-level-table
	       domain
	       root-actions
	       all-actions
	       action-to-component-table
	       component-successor-table))
	     (effect-levels
	      (build-effect-level-table
	       domain
	       action-levels)))
	(values
	 action-levels
	 effect-levels
	 components)))))		;n.b. action names, not effect names

(defun build-action-level-table (domain
				 root-actions
				 all-actions
				 action-to-component-table
				 component-successor-table)
  (let* ((component-level-table
	  (find-longest-path-lengths*
	   (mapcar (hash-table->function action-to-component-table)
		   root-actions)
	   (hash-table->function component-successor-table)))
	 (action-level-table
	  (make-hash-table :test #'eq)))
    ;; We now have levels for the components, but not for the actions
    ;; in the components.  So we have to get the action levels.
    (let ((unreachable-actions '()))
      (dolist (action all-actions)
	(let ((component (gethash action action-to-component-table)))
	  (when (null component)
	    (push action unreachable-actions))
	  ;; unreachable actions get level 0,
	  ;; others get their component's level.
	  (setf (gethash action action-level-table)
		(gethash component component-level-table 0))))
      (when unreachable-actions
	(tf-warning "Domain \"~A\" has unreachable actions ~S."
		    (domain-name domain)
		    unreachable-actions)))
    action-level-table))

;;; The level of an effect E is max of the levels of all actions A
;;; such that E is in (action-effect-names A).  If there is no such
;;; action, the level of E is 0.

(defun build-effect-level-table (domain action-level-table)
  (let ((effects (domain-effect-names domain))
	(actions (hash-table-keys action-level-table))
	(table (make-hash-table :test #'eq)))
    (macrolet ((effect-level (e) `(gethash ,e table))
	       (action-level (a) `(gethash ,a action-level-table)))
      (dolist (e effects)
	(setf (effect-level e) 0))
      (dolist (a actions)
	(let ((level (action-level a)))
	  (dolist (e (action-effect-names a))
	    (when (< (effect-level e) level)
	      (setf (effect-level e) level)))))
      table)))

;;; We'll also want to check condition levels.

(defun check-condition-levels (domain action-levels effect-levels components)
  (let ((level-errors 0))
    (dolist (schema (domain-schemas domain)) ;includes tasks
      (let ((schema-level
	     (gethash (get-schema-action-name schema) action-levels)))
	;; Schema-level is the level of an action.  Certain condition types
	;; restrict how the effects that could satisfy them can differ in
	;; level from the action.
	(assert (integerp schema-level))
	(dolist (c (schema-conditions schema))
	  (unless (eq (con-type c) 'compute)
	    ;; Look at the effect that corresponds to the condition
	    (let* ((effect-name (car (con-pattern c)))
		   (effect-level (gethash effect-name effect-levels)))
	      (cond ((null effect-level)
		     (tf-warning "No level for effect \"~A\"." effect-name))
		    ((not (effect-level-ok-for-condition-p
			   (con-type c) effect-level schema-level))
		     ;; Level mismatch, a.k.a. hierarchical promiscuity.
		     (incf level-errors)
		     (condition-level-error
		      schema c effect-level schema-level))))))))
    ;; If there are level errors, or if the user wants to know anyway,
    ;; produce a level description.
    (when (> level-errors 0)
      (tf-message "~D level error~:P.  Level assignments follow."
		  level-errors))
    (when (or (> level-errors 0)
	      *level-report*)
      (report-domain-levels domain action-levels effect-levels components))))

(defun condition-level-error (schema condition effect-level action-level)
  (tf-warning				;was tf-error
   "Level mismatch in schema \"~A\",~%~
      ~6T for condition \"~S {~S ...}\".~%~
      ~6T The level of effect \"~S\" is ~D,~%~
      ~6T while that of action \"~S\" is ~D,~%~
      ~6T but the effect should not be at ~A level."
     (schema-name schema)
     (con-type condition)
     (car (con-pattern condition))
     (car (con-pattern condition))
     effect-level
     (get-schema-action-name schema)
     action-level
     (ecase (signum (- effect-level action-level))
       (-1 "a higher (lower numbered)" )
       ( 0 "the same"                  )
       ( 1 "a lower (higher numbered)" ))))

(defun effect-level-ok-for-condition-p (cond-type effect-level action-level)
  (ecase cond-type
    ((only_use_if only_use_for_query)
     (<= effect-level action-level))
    ((supervised)
     (>= effect-level action-level))
    ((unsupervised)
     ;; /\/: AAAI 94 paper says: same or higher if cond is sat externally
     ;; to the sub-actions of the schema or same or lower if sat by the
     ;; sub-actions.  [I suppose we could see whether only the subactions
     ;; _could_ satisfy it.  If not, I don't know what check we could make.]
     ; (>= effect-level action-level)
     t)
    ((achieve
      achievable)			;/\/ for now
     ;; Achieve for action A will force the actions for schemas with
     ;; corresponding oufe conditions to be >= level(A).  Therefore, the
     ;; effects that go with the schemas will also have level >= level(A).
     ;; /\/: However, when there's no oufe, there may still be an ordinary
     ;; effect at a higher (lower-numbered) level.  Since we allow such
     ;; effects to satisfy achieves, we should probably treate them as
     ;; oufe effects when calculating levels, but we don't.  Instead,
     ;; we treat this (rare) case as a level error.
     (>= effect-level action-level))))

(defun report-domain-levels (domain action-levels effect-levels components)
  (tf-message "")
  (tf-message "Levels for domain \"~A\"" (domain-name domain))
  (tf-message "")
  (report-domain-components components)
  (tf-message "")
  (report-level-assignments "action" action-levels)
  (tf-message "")
  (report-level-assignments "effect" effect-levels)
  (tf-message "")
  (values))

(defun report-domain-components (components)
  (let ((interesting
	 (remove-if-not #'(lambda (c) (not (null (cdr c)))) ; len > 1
			components)))
    (tf-message "There are ~D strongly connected components in the~%~
                 action graph that have length greater than one."
		(length interesting))
    (when interesting
      (tf-message "They are:")
      (dolist (component interesting)
	(tf-message "~3T~S" component)))))

(defun report-level-assignments (type level-table)
  (tf-message "~@(~A~) levels:" type)
  (let* ((items
	  (hash-table-keys level-table))
	 (levels
	  (group-by-numeric-key
	    items
	    (hash-table->function level-table))))
    (loop for level-members in levels
	  for level from 0
	  when level-members
       do (tf-message "~4D ~S" level (car level-members))
          (loop for item in (cdr level-members)
	     do (tf-message "     ~S" item)))))
  

;;; Entry point from the main compiler.

;;; *domain* should be the domain to check.
;;; All new schemas, etc, should already be in place.

(defvar *action-level-table* nil)
(defvar *effect-level-table* nil)

(defun process-domain-levels (domain)
  (assert (eq *domain* domain))
  (tf-progress "Level analysis~%")
  (tf-progress "Building level tables~%")
  (multiple-value-bind (action-levels effect-levels components)
      (build-level-tables domain)
    (setq *action-level-table* action-levels
	  *effect-level-table* effect-levels)
    (tf-progress "Checking condition levels~%")
    (check-condition-levels domain action-levels effect-levels components)))

;;; Procedures for looking up levels

(defun action-level (action-name)
  (or (gethash action-name *action-level-table*)
      (error "No level for action ~S." action-name)))

(defun effect-level (effect-name)
  (or (gethash effect-name *effect-level-table*)
      (error "no level for effect ~S." effect-name)))

;;; ---------------------------- Change History ----------------------------
;;; (Who)   (When) 		     (What)
;;;
