;;;; File: di-toplevel.lsp
;;; SCCS Version: %W%
;;; Contains: Toplevel control for the Diary.
;;; Author: Jeff Dalton and Richard Kirby
;;; Created: Mon Jun 10 16:32:08 1991
;;; Updated: Tue Nov 22 19:07:09 1994 by Jeff Dalton
;;; Copyright: (c) 1992, AIAI, University of Edinburgh
;;; This material may be reproduced by or for the U.S. Government pursuant
;;; to the copyright license under the clause at DFARS 252.227-7032
;;; (June 1975) -- Rights in Technical Data and Computer Software (Foreign).

(in-package :diary)

;;; Needs XP, termcaplib, monitors, mcc stuff, ...

(use-package '(:util :time-util :sim-clock :dev :pseudo-process :ipc :mcc
	       :components :xwindowio :initsystem))

(import '(oplan::make-event))

(import 'di-startup :oplan)		;/\/ for startup

(export '*simulated-seconds-per-second*)

;;;; Diary messages
;;;
;;; Messages to the Diary are sent by
;;;
;;;   (ipc-write :MCC (mcc-make-message :MCC :DIARY <diary-message>))
;;;
;;; or by the equivalent
;;;
;;;   (ipc-write-diary <diary-message>)
;;;
;;; So a <diary-message> is the mcc-message-args of an mcc-message.
;;; It can take the following forms:
;;;
;;; :EVENT-AT simulated-time agenda-body
;;;
;;;    Arranges for an event to be sent to :LEFTIN at a certain time.
;;;
;;;    Agenda-body must be something that can be used as the body of
;;;    an agenda entry (ie, of an agenda-rec struct).  When the due-time
;;;    arrives, the Diary sends the agenda-body to :LEFTIN as the body
;;;    of an event struct, and the IM converts that to an agenda entry
;;;    before sending it to the AM.
;;;
;;; :START-CLOCK universal-time simulated-time scale-factor
;;;
;;;    Establishes a relation between universal (real) and simulated
;;;    time and allows events to be dispatched when their due-times
;;;    arrive.
;;;
;;; :STOP-CLOCK
;;;
;;;    Prevents events from being dispatched until the clock is
;;;    restarted.
;;;
;;; :CLEAR-DIARY
;;;
;;;    Removes all entries from the Diary.
;;;

;;;; The Diary

;;; The *diary* is a list of diary-entry structs sorted by due-time,
;;; where the due-time is a CL "universal time".  An event is sent off
;;; via :LEFTIN to become an agenda entry when its time comes up.

(defvar *diary* nil)

(defstruct (diary-entry (:conc-name di-))
  due-time			;sim time when event should occur
  event)			;an agenda-entry body

(defun add-diary-entry (entry)
  (setq *diary* (merge 'list *diary* (list entry) #'< :key #'di-due-time))
  entry)

(defun delete-diary-entry (entry &rest delete-args)
  (setq *diary* (apply #'delete entry *diary* delete-args))
  entry)


;;;; Time control

;;; See suppirt/sim-clock.lsp for an explanation.

(defvar *simulated-seconds-per-second* 1) ;used in the REA

(defun set-diary-alarm-clock ()
  (when (and *diary* (the-clock-is-running))
    (let ((next-due-time (di-due-time (first *diary*))))
      (set-to-sleep-until
        (simulated->universal-time next-due-time)))))


;;;; Initialization and event handler

(defun di-startup ()
  "This is the top level function for the Diary (DI) process."
  (unless (is-init *whoami*)
    (dev-error *whoami* "is-init failed."))
  (setf (component-init-fn *pprocess*) #'di-init)
  (x-redirect-errors :diout)
  (clear-screen *debug-io*)
  (mcc-register-with-im)
  (define-a-clock)
  (ipc-set-event-handler 'di-event-prehandler))

(defun di-init ()
  (clear-screen *debug-io*)
  (setq *diary* nil)
  (clear-pop-gensymtable)		;/\/ really?
  (dev-debug-msg "Initialised" *whoami* :information))

(defun di-event-prehandler (self)
  ;; Process all messages now, rather than one each time we run,
  ;; to keep the diary maximally up-to-date.
  (while (next-event-p self)
    (let ((e (next-event self)))
      (di-event-handler (message-route e) (message-contents e))))
  ;; Activate any diary events whose time has come.
  (check-due-times)
  ;; Tell the user if we have anything scheduled.
  (cond ((and *diary* (the-clock-is-running))
	 (let ((wakeup-time (di-due-time (car *diary*))))
	   (display-a-msg "Next event at: ~A."
	      (plan-time-string wakeup-time))))
	((null *diary*)
	 (display-a-msg "No diary entries, so waiting.")))
  ;; Arrange to wake up, if we have to.
  (set-diary-alarm-clock))

(defun di-event-handler (stream event)
  (ecase stream
    (:mcc (mcc-handler event))
    (:diaryin
      (ecase (mcc-msg-ident event)
	(:diary
	  ;; Handle a diary message
	  (handle-diary-message (mcc-msg-args event)))))))


;;; Check for events whose time has come...

(defun check-due-times ()
  (when (and (the-clock-is-running) *diary*)
    (let ((now (get-simulated-time)))
      (while (and *diary* (>= now (di-due-time (car *diary*))))
	(send-event (pop *diary*))))))

(defun send-event (event)
  (display-a-msg "At ~A: ~A"
    (plan-time-string (di-due-time event))
    (di-event event))
  (ipc-send :LEFTIN :EVENT (make-event :BODY (di-event event))))


;;;; Diary-message handlers

(defun handle-diary-message (message)
  (let ((operation (car message))
	(args (cdr message)))
    (apply (ipc-get-handler :di operation)
	   args)))

(defmessage (:di :event-at) (sim-time agenda-body)
  (add-diary-entry
    (make-diary-entry
      :due-time sim-time
      :event agenda-body)))

(defmessage (:di :start-clock) (universal-time simulated-time scale-factor)
  (setq *simulated-seconds-per-second* scale-factor)
  (start-clock universal-time simulated-time scale-factor))

(defmessage (:di :stop-clock) ()
  (stop-clock))

(defmessage (:di clear-diary) ()
  (setq *diary* nil))


;;;; Routines for limited-width messages

(defvar *max-length* nil)

(defun get-max-length ()
  (or *max-length* (setq *max-length* (1- (is-get-window-width :diout)))))

(defun display-a-msg (format-string &rest format-args)
  ; (scroll-up)
  (fresh-line *debug-io*)
  (format-limited-width
     *debug-io* (get-max-length) "~?" format-string format-args)
  (force-output *debug-io*))

;;; Here is the primitive, a version of format:

(defvar *buffer* nil)

(defun ensure-buffer (size)
  (declare (ignore size))
  (or *buffer*
      (setq *buffer* (make-string-output-stream))))

(defun format-limited-width (stream limit format-string &rest format-args)
  (let ((*print-pretty* nil)
	(buf (ensure-buffer limit))
	(str nil))
    (unwind-protect (apply #'format buf format-string format-args)
      (setq str (get-output-stream-string buf)))
    (write-string str stream :end (min limit (length str)))))

;;; ---------------------------- Change History ----------------------------
;;; (Who)   (When) 		     (What)
