;;;; File: inter-agent.lsp
;;; SCCS Version: %W%
;;; Contains: Agent-spawning and inter-agent communications code
;;; Author: Jeff Dalton <J.Dalton@ed.ac.uk>
;;; Created: September 1994
;;; Updated: Tue Feb 14 00:47:52 1995 by Jeff Dalton
;;; Release Version: %Y%
;;; Copyright: (c) 1994, AIAI, University of Edinburgh
;;; This material may be reproduced by or for the U.S. Government pursuant
;;; to the copyright license under the clause at DFARS 252.227-7032
;;; (June 1975) -- Rights in Technical Data and Computer Software (Foreign).

(in-package :mcc)

;;; Send-to-agent-on-right

;;; /\/: In principle, sending to the right should always work the
;;; same way, but in fact it depends on whether we're sending to
;;; the Exec or to the World.  This is because the Exec is a completely
;;; separate system we have to tell the OS to run, while the World is
;;; (in this sense) part of the Exec.

;;; N.B. Knowledge sources (and other MCC clients) send to the right
;;; by calling mcc-send-to-rightout, *not* send-to-agent-on-right.
;;; They have to send a :PRINT :RIGHTOUT message to the IM via MCC,
;;; which is what mcc-send-to-rightout does, and then the IM will
;;; call send-to-agent-on-right while handling that message.

(defun send-to-agent-on-right (msg)
  (if (get-parameter :asexec)		;so which agent is that, anyway?
      (send-exec-to-world msg)		;if we're the Exec, it's the World
      (send-planner-to-exec msg)))	;else we're the Planner & it's the Exec


;;; World connection

;;; /\/: Perhaps we shouldn't start the World until we know we need it.
;;; After all, the Exec might use some other, external world simulator.

(defun send-exec-to-world (msg)
  (ipc-write :rightout msg))		;that was easy!


;;; Exec connection

;;; /\/: We should have a way to start a new Exec if an existing Exec
;;; exits.  So we need a way to tell that it _has_ exited.

(defvar *exec-host* nil)
(defvar *exec-display* nil)
(defvar *exec-io* nil)

(defvar *exec-shell-command* nil)

(defun send-planner-to-exec (msg)
  (if (socket-server (find-socket :rightout))	;anyone there?
      (ipc-write :rightout msg)			;yes -- send to them
    (send-planner-to-external-exec msg)))	;no -- start external exec

(defun send-planner-to-external-exec (msg)
  ;; Record the plan in a file in case we want to run the Exec later.
  (with-open-file (out (temp-filename "planner-to-exec")
		       :direction :output)
    (write msg :stream out :escape t :pretty t :length nil :level nil)
    (terpri out)
    (force-output out))
  ;; Run the exec now if desired and not already running.
  (when (and (null *exec-io*)
	     (run-exec-agent-p))
    (run-exec-agent))
  (when *exec-io*
    (ipc-write :rightout msg)))

(defun run-exec-agent-p ()
  (ensure-exec-parameters)
  (loop
    (ecase (exec-option-menu)
      (:yes (return t))
      (:no  (return nil))
      (:new-host
       (setq *exec-host* (ask-user "Host-name: ")))
      (:new-display
       (setq *exec-display* (ask-user "Display: "))))))

(defun ensure-exec-parameters ()
  (unless *exec-host*
    (setq *exec-host*
	  (or (getenv "OPLANEXECHOST")
	      "localhost")))
  (unless *exec-display*
    (setq *exec-display*
	  (or (getenv "OPLANEXECDISPLAY")
	      (getenv "DISPLAY")
	      (concat-string *exec-host* ":0.0"))))
  (unless *exec-shell-command*
    (setq *exec-shell-command*
	  (concat-string (get-parameter :oplan-dir) "/bin/oplan"))))

(defun exec-option-menu ()
  (menu-request
    `("-heading" ,(format nil "Exec on ~S display: ~S?"
		    *exec-host* *exec-display*)
      "Yes=:yes"
      "No=:no"
      "Change host=:new-host"
      "Change display=:new-display")))

(defun ask-user (question &rest format-args)
  ;; /\/ There's a similar routine with the same name in the TA.
  (format *terminal-io* "~&~?" question format-args)
  (read-line *terminal-io*))

(defun run-exec-agent ()
  (setq *exec-io*
	(ipc-run-agent :exec
	  "rsh" *exec-host*
	  "cd" (namestring (working-directory-pathname))
	  ";"
	  "env" (concat-string "DISPLAY=" *exec-display*)
		(concat-string "OPLANTMPDIR=" (get-parameter :oplan-tmp-dir))
	        ;; Add env vars above this line
	  *exec-shell-command*
	  (util:get-parameter :image-name)
	  "-connect"
	  "-asexec"))
  (new-connector :right-io
    :io-stream *exec-io*
    :socket-to-read :rightout
    :socket-to-write :rightin
    :input-guard 'rightin-guard)
)

(defun rightin-guard (self input)
  (declare (ignore self))
  ;; Check input before it's sent to :rightin
  (assert (consp input))
  (values t
	  ;; /\/: The :read is ignored.
	  ;; /\/: Should probably do the same sort of thing we do
	  ;;      for :leftin (ie, make an event struct).
	  (list :read input)))

;;; ---------------------------- Change History ----------------------------
;;; (Who)   (When) 		     (What)
;;;
