;;;; File: KS-ADD-TO-TASK.lsp
;;; SCCS Version: %W%
;;; Contains: KS-ADD-TO-TASK
;;; Author: Jeff Dalton <J.Dalton@ed.ac.uk>
;;; Created: 01 June 1995
;;; Updated: Tue Jun 27 17:44:35 1995 by Jeff Dalton
;;; Release Version: %Y%
;;; Copyright: (c) 1995, AIAI, University of Edinburgh
;;; This material may be reproduced by or for the U.S. Government pursuant
;;; to the copyright license under the clause at DFARS 252.227-7032
;;; (June 1975) -- Rights in Technical Data and Computer Software (Foreign).

(in-package :knowledge-source)

;;; KS-ADD-TO-TASK can get as (ag-body event):
;;;
;;;   (:ADD-TO-TASK subop arg...)
;;;
;;; Suboperations:
;;;
;;; :ACTION pattern
;;;
;;;    Adds a node and expands.
;;;
;;;    /\/: Might want to be able to specify the node-end to link the
;;;    action before.
;;;
;;;    /\/: If the action's added during plan execution, we need to
;;;    pay attention to what's already been executed and make sure
;;;    the new action's after that.
;;;
;;; No other suboperations have yet been defined.
;;;

(defun KS-ADD-TO-TASK (event)
  (ensure-loaded-ks :execute)		;for various routines we'll call /\/
  (destructuring-bind (op subop &rest args) (ag-body event)
    (assert (eq op :ADD-TO-TASK))
    (apply
       (ecase subop
	 (:ACTION 'add-action-to-task))
       args)))

(defun add-action-to-task (pattern)

  ;; See if any schemas look like they'll expand the action represented
  ;; by the pattern.
  (let ((schemas (get-expand-schemas pattern)))
    (when (null schemas)
      (notify-user "No schemas will expand ~S" pattern)
      (return-from add-action-to-task nil)))

  ;; If executing, make a global after-point after everything that
  ;; has already been executed or sent to the Exec so that any changes
  ;; are to later parts of the plan.  Note that we can't link after
  ;; the :ready node-ends, because that might violate time-windows.
  (when (member :exec (global-activities))
    (db-call 'pinch-plan-at
      (db-call 'ads-add-node :type 'dummy)
      :sent))

  ;; Block backtracking.
  (block-backtracking)

  ;; Add a node to represent the new action.
  (let ((node-tag
	 (db-call 'ads-add-node
		  :type 'action
		  :pattern pattern)))

    ;; Link it in.
    (db-call 'link-etags-else-error
       (or (db-call 'global-after-point)
	   (etag 'node-1 :end))
       (etag node-tag :begin))
    (db-call 'link-etags-else-error
       (etag node-tag :end)
       (etag 'node-2 :begin))

    ;; Post an EXPAND to expand it
    (post-expand `(:EXPAND ,node-tag ,pattern)))

  ;; If executing, make sure the departure lists take any new nodes
  ;; into account.
  (when (member :exec (global-activities))
    (post-agenda `(:CONTINUE-EXECUTION :AFTER-ADDED-ACTION))))


;;; ---------------------------- Change History ----------------------------
;;; (Who)   (When) 		     (What)
;;;
