;;;; File: KS-OR.lsp
;;; SCCS Version: %W%
;;; Contains: Handles AND/OR interaction tree.
;;; Author: Richard Kirby and Jeff Dalton
;;; Created: Wed Nov 27 15:56:27 1991
;;; Updated: Thu Jun 29 17:21:50 1995 by Jeff Dalton
;;; Release Version: %Y%
;;; Copyright: (c) 1992, 1993, 1994, AIAI, University of Edinburgh
;;; This material may be reproduced by or for the U.S. Government pursuant
;;; to the copyright license under the clause at DFARS 252.227-7032
;;; (June 1975) -- Rights in Technical Data and Computer Software (Foreign).

(in-package :knowledge-source)

(use-package :QA)

#| This handles the AND/OR linkage/binding tree returned from the TGM when
adding effects or conditions. Basically it selects a choice at the top-level,
posts KS-LINKs and KS-BINDs for the linkages and/or bindings required, a KS-OR
to handle the next level (if there is one), and an alternative to handle any
other choices.

/\/: No it doesn't.  It directly asks the DM to bind, restrict, link, etc.
rather than posting KS-LINKs and KS-BINDs.  KS-LINK doesn't even exist.  [JD]
|#

(defun KS-OR (ag)
  (STAGE-MANAGER
   nil					;no state vars
   (cond satisfy-p or-tree body contribs
    initial-n-branches)
   ag
   
   (setq body (cdr (ag-body ag)))
   (setq cond (first body))
   (setq contribs (second body))
   (setq satisfy-p (third body))
   (setq or-tree (cdddr body))
   (setq initial-n-branches (length or-tree))

   (assert (&= initial-n-branches (ag-branch-1 ag)))

   (loop

      (when (null or-tree)
	(return))

      ;; A list of alternatives, so take the first and post the rest back.
      (multiple-value-bind
	    (choice rest) (pick-from-next-level or-tree)
	(when (and rest (not (eql satisfy-p :NOT-AGAIN)))
	  (with-ag-copy (ag ag)
	    (setf (ag-body ag) `(:OR ,cond ,contribs ,satisfy-p ,@rest))
	    ;(setf (ag-body ag) `(:OR ,cond ,contribs :NOT-AGAIN ,@rest))
	    (setf (ag-branch-1 ag) (length rest))
	    (post-alternatives ag)))
	(setq or-tree choice))
   
      (whats-going-on "Doing the following:")
   
      (if (null satisfy-p)
	  (if (or (and (eql (con-type cond) 'achievable)
		       (every #'(lambda (x)
				  (eql (cdr x) :already-satisfied))
			      contribs))
		  ;; This OR-tree posted to resolve an interaction with an
		  ;; achievable which was already-satisfied. So what we do here
		  ;; is set up an alternative where the opposite bindings and
		  ;; restrictions are made and a request to reachieve the
		  ;; achievable any old how.
		  #|(member (con-type cond) '(only_use_for_query))|#
		  ;; This OR-tree posted to resolve an interaction with a
		  ;; reestablishable type of condition (ie only_use_for_query).
		  ;; So what we do here is to set up an alternative where we
		  ;; try to resatisfy the condition with the opposite bindings
		  ;; and restrictions.
		  )
	      (set-up-alternative-with-opposite cond (car or-tree)))
	  (unless (eql satisfy-p :NOT-AGAIN)
	    (if (member (con-type cond) '(only_use_for_query))
		(set-up-alternative-with-opposite cond (car or-tree)))))

      (unless (process-or-rec cond (car or-tree) satisfy-p)
	(return-from KS-OR nil))		;will poison, so quit now.

      ;; We now consider the next level of the or-tree.  The right thing
      ;; to do is probably to post back to the AM so that this OR can be
      ;; compared to other agenda entries that might be better choices.
      ;; But to decrease the difference from past behavior, we send the
      ;; OR back to the AM only if the number of branches is greater than
      ;; it was initially.  This ensures that if this OR was picked because
      ;; it's branch-1 was 1, it doesn't go on to less constrained parts
      ;; of the or-tree.

      (setq or-tree (cdr or-tree))
      (when (> (length or-tree) initial-n-branches)
	;; Send the OR back to the AM.
	;; /\/: Should the ID change?
	(post-or-agenda-record cond contribs satisfy-p or-tree)
	(return))
      ;; Otherwise, go 'round the loop to process the next level of the tree.
      )))

;; Keep simple at mo.
(defun pick-from-next-level (res-list)
  (values (car res-list) (cdr res-list)))


(defun process-or-rec (cond or-rec update-p)
  
  (dolist (action (or-info-rec-actions or-rec)
	   t)				;success
    (ecase (car action)
      (:BIND
       (whats-going-on "Binding ~S to ~S" (second action) (third action))
       (when (null (set-binding (second action) (third action)))
	 (post-agenda `(:POISON-STATE :COULD-NOT-BIND
			,(second action) ,(third action)))
	 (return nil)))
      (:RESTRICT
       (whats-going-on "Making ~S not equal to ~S"
		       (second action) (third action))
       (when (null (car (add-restriction-to-psv
			 (second action) :NOT (third action))))
	 (post-agenda `(:POISON-STATE :COULD-NOT-RESTRICT
			,(second action) ,(third action)))
	 (return nil)))
      (:LINK
       (whats-going-on "Linking ~S to ~S" (second action) (third action))
       (when (null (db-request :ADD-LINK (second action) (third action)))
	 (post-agenda `(:POISON-STATE :COULD-NOT-ADD-LINK :FROM
			,(second action) :TO ,(third action)))
	 (return nil)))
      (:UPDATE
       (when update-p
	 (whats-going-on "Updating the contributors for~%~S" cond)
	 (db-request :UPDATE-CONTRIBUTORS cond (second action))))
      (:REACHIEVE
       (whats-going-on "Posting to reachieve ~%~S" cond)
       ;; Remove GOST entry for the condition.
       (let ((old-contribs
	      (db-request :UPDATE-CONTRIBUTORS cond :undef)))
	 ;; Strip off the satisfaction method keyword.
	 (setq old-contribs
	       (mapcar #'car old-contribs))
	 (if (eql (cadr (second action)) 'achievable)
	     (post-achieve `(:ACHIEVE ,(cdr (second action)))
		 :trigger t
		 :stage   0
		 :info    (list nil nil nil old-contribs))
	     (post-condition `(:CONDITION ,(cdr (second action)))
		 :trigger t
		 :stage   0
		 :info    old-contribs)))))))


(defun set-up-alternative-with-opposite (cond or-rec)
  ;; As long as the set of actions in the or-tree do not contain any linkings
  ;; set up an alternative with an or-tree which gives the not of the bindings
  ;; and restrictions required and an action to reachieve the cond.
  (unless (null (or-info-rec-actions or-rec))
    (let (opposite-actions)
      (dolist (action (or-info-rec-actions or-rec))
	(case (car action)
	  (:BIND
	   (push (list :RESTRICT (second action) (third action))
		 opposite-actions))
	  (:RESTRICT
	   (push (list :BIND (second action) (third action))
		 opposite-actions))
	  (:LINK
	   (push (list :NOT-LINK (second action) (third action))
		 opposite-actions)
	   #|(MONITOR :KP
		    (:DETAILED "LINKings not handled so ignore at moment."))|#)
	  (:UPDATE)
	  (:REACHIEVE)))
      (when opposite-actions
	(unless (every #'(lambda (x) (eql (car x) :NOT-LINK))
		       opposite-actions)
	  (post-alternatives
	    (assign-or-level
	     (make-agenda-rec
		:body `(:OR ,cond nil :NOT-AGAIN
			,(list (qa::make-or-info-rec
				:actions (list
					  (list :REACHIEVE
						(cons 'GOST cond))))))
		:trigger (make-wait-for-effect-trigger cond nil)
		:branch-1 1))
	    #|opposite-actions|#))))))


(defun some-bindings-to-do-p (or-tree)
  "Returns t if there are some bindings, or restrictions to be made."
  (some #'(lambda (x) (or (eql (car x) :BIND) (eql (car x) :RESTRICT)))
	(qa::or-info-rec-actions (car or-tree))))


;;; ---------------------------- Change History ----------------------------
;;; (Who)   (When) 		     (What)
;;;
