;;;; File: kp-supportlib.lsp
;;; SCCS Version: %W%
;;; Contains: Various functions for use by KS writers.
;;; Author: Richard Kirby (rbk)
;;; Created: Fri Jun 22 09:02:34 1990
;;; Updated: Mon Jul  3 19:57:12 1995 by Jeff Dalton
;;; Release Version: %Y%
;;; Copyright: (c) 1992, AIAI, University of Edinburgh
;;; This material may be reproduced by or for the U.S. Government pursuant
;;; to the copyright license under the clause at DFARS 252.227-7032
;;; (June 1975) -- Rights in Technical Data and Computer Software (Foreign).

(in-package :knowledge-platform)

(import '(qa:or-info-rec-actions))


;;;; Global variables

(defvar *ag*)				;used by ACHIEVE and EXPAND


;;;; List of global activities

(definit :kp *global-activities* '()
  "A list of keywords that say whether we're planning, executing,
   etc.  E.g.: (:plan) while planning, (:plan :exec) if planning
   while executing.")

(defun global-activities ()
  *global-activities*)

;;; /\/: Define a macro that defines a fn that accesses a variable,
;;; or maybe a fn that accesses any "place".

(defun set-global-activities (activities)
  (setq *global-activities* activities))

(defsetf global-activities set-global-activities)


;;;; Etc.
  
;; Macro sometimes used when making a copy for posting alternatives.
(defmacro with-ag-copy ((var ag) &body body)
  `(let ((,var (copy-agenda-rec ,ag)))
     ,@body))

;; Returns a copy of a condition cond with the condition node-end replaced
;; with node-end.
(defun copy-cond-and-replace-cond-node-end (cond node-end)
  (let ((result (copy-list cond)))
    (setf (con-at-node-end result) node-end)
    result))

;; Checks if there is a requirement for linking.
(defun linking-suggestions-p (satisfy)
  (let ((link-no-bindings-level
	 (tgm-get-satisfaction-method-level :LINK-NO-BINDINGS)))
    (some #'(lambda (x)
	      (let* ((update (assoc :UPDATE (or-info-rec-actions (car x))))
		     (update-level (cdar (cadr update))))
		(>= (tgm-get-satisfaction-method-level
		     update-level)
		    link-no-bindings-level)))
	  (cadr satisfy))))

(defun check-only-use-ifs (cond bindings)
  ;; Makes a request on the DM for checking the applicability of a schema via
  ;; its only-use-if conditions, passing in the schema vars bindings. The
  ;; return result is either nil indicating a failure, or else an adjusted set
  ;; of bindings.
  (let (result)
    (setq result (db-request :CHECK-ONLY-USE-IFS cond bindings))
    (when (and (tgm-positive-result-p result)
	       (eql (length (cadr result)) 1)
	       (eql (length (caadr result)) 1))
      ;; If there is only one result, then set any of the bindings (if this is
      ;; an only_use_for_if).
      (if (eql (con-type cond) 'only_use_if)
	  (dolist (action (or-info-rec-actions (caaadr result)))
	    (if (and (eql (first action) :BIND)
		     (not (psvar-p (second action))))
		(setf (second (assoc (second action) bindings))
		      (third action))))))
    (tgm-positive-result-p result)))

(defun can-evaluate-compute-p (cond bindings)
  (let ((p (con-pattern cond)))
    ;; Check that all vars in p have values
    (db-request :fully-instantiable-p p bindings)))

(defun evaluate-compute (cond bindings)
  ;; Returns true or false to indicate success or failure.
  ;; May also set the values of some variables in bindings.
  (let ((p (con-pattern cond))
	(v (con-value cond)))
    ;; It is assumed that all vars in p have values.
    (let ((fn-result
	   (db-request :apply 'funcall
		       (db-request :fully-instantiate p bindings))))
      (setf (con-pattern cond)
	    `(compute-memo ,p ,fn-result))
      (case v
	;; True and false are special cases
	((true)  (not (null fn-result)))
	((false) (null fn-result))
	(t
	 ;; Otherwise we match the fn-result against v.
	 (let ((new-binds (db-request :obmatch fn-result v bindings)))
	   (if (atom new-binds)
	       new-binds		;presumably t or nil
	     (progn
	       (maybe-set-values bindings new-binds)
	       t))))))))

(defun maybe-set-values (old-binds new-binds)
  ;; /\/: This is not perfect, given that the binding lists may contain
  ;; unifications and PSV references, but it should work about as well
  ;; as our current handling of only_use_if and expands patterns.
  ;; The idea is that if an entry in new-binds has a value when the
  ;; corresponding entry in old-binds has value :undef, we can change
  ;; the entry in old-binds.
  (macrolet ((v-name (tuple) `(first ,tuple)) 	;/\/ shouldn't need own
	     (v-val (tuple) `(second ,tuple))) 	;/\/ shouldn't need own
    (dolist (b new-binds)
      (let ((v (v-val b)))
	(when (not (eq v :undef))
	  (let ((old-tuple (assoc (v-name b) old-binds)))
	    (when (and old-tuple
		       (eq (v-val old-tuple) :undef))
	      (setf (v-val old-tuple) v))))))))

#|
(defun filter-on-only-use-ifs (schema bindings psuedo-node-end)
  (let ((conditions (schema-conditions schema)))
    (dolist (cond conditions t)
      (if (member (con-type cond) '(only_use_if only_use_for_query))
	  ;; Need to check if the condition is satifyable, but
	  ;; we don't do any work if linking in or out required
	  ;; since this schema may not be chosen anyway.
	  ;; Should keep the result around to save having to do the work again
	  ;; if we eventually pick this schema.
	  (progn
	    (setq cond (copy-cond-and-replace-cond-node-end
			cond psuedo-node-end))
	    (unless (check-only-use-ifs cond bindings)
	      ;; Now check if there is an outstanding Agenda Entry which could
	      ;; potentially provide the effect that we are looking for to
	      ;; satisfy this condition. If there is such a thing, then record
	      ;; this schema as a possibility to be reconsidered.
	      (whats-going-on
	       "Throwing away ~S on a failed only_use_if check"
	       (schema-name schema))
	      (return nil)))))))
|#

(defun make-wait-for-effect-trigger (condition bindings)
  (declare (ignore bindings))
  ;; bindings may be used later on for instantiating the cond pattern which
  ;; would allow for a more specific trigger.
  (cons :WAIT-ON-EFFECT (con-pattern condition)))

(defun possibly-satisfying-plan-state-changes-p (condition bindings)
  (let ((effect (con-pattern condition)))
    (db-request :CHECK-FOR-EFFECT effect bindings (ag-id *ag*))))

;;; (filter-on-only-use-ifs schema bindings psuedo-node-end) --
;;;
;;; Given the schema with the set of bindings, checks all the
;;; only_use_ifs and only_use_for_querys.  For each such condition, if
;;; it can be satisfied there and then, then carry on, otherwise if
;;; can't satisfy, check if there is an outstanding agenda entry that
;;; could possibly cause an effect to be added that could satisfy the
;;; condition.  If so then if this is an only_use_if set up a trigger
;;; to wait for the effect, or if it is an only_use_for_query carry
;;; on, otherwise if there is no such agenda-entry return nil.
;;;
;;; Returns one of:
;;;   nil - throw away the schema;
;;;   t   - keep schema;
;;;   <trigger> - keep schema as an alternative but with a trigger.

(defun filter-on-only-use-ifs (schema bindings psuedo-node-end)
  (let ((conditions (schema-conditions schema))
	trigger)
    (dolist (cond conditions)
      (case (con-type cond)
	((only_use_if only_use_for_query)
	 ;; Need to check if the condition is satisfiable, but we don't do
	 ;; any work to change the plan state since this schema may not be
	 ;; chosen anyway.
	 ;; Should keep any TGM result around to save having to do the work
	 ;; again if we eventually pick this schema.
	 (setq cond (copy-cond-and-replace-cond-node-end cond psuedo-node-end))
	 (unless (check-only-use-ifs cond bindings)
	   ;; Now check if there is an outstanding Agenda Entry which could
	   ;; potentially provide the effect that we are looking for to satisfy
	   ;; this condition. If there is such a thing then add a trigger for
	   ;; the cond to wait for the Agenda Entrys to be processed.
	   (if (possibly-satisfying-plan-state-changes-p cond bindings)
	       (push (make-wait-for-effect-trigger cond bindings)
		     trigger)
	       (progn
		 (whats-going-on
		    "Throwing away ~S on a failed only_use_if check"
		    (schema-name schema))
		 (return-from filter-on-only-use-ifs nil)))))
	((compute)
	 ;; If the condition can be evaluated, then evaluate it.  This
	 ;; may bind some variables by matching against the value pattern.
	 ;; If the match fails, reject the schema.
	 (when (can-evaluate-compute-p cond bindings)
	   (unless (evaluate-compute cond bindings)
	     (return-from filter-on-only-use-ifs nil))))))
    ;; If reach here then we have checked all the conditions and either they
    ;; all passed or trigger has been set to a list of triggers to cause this
    ;; schema to wait for some agenda entries that may introduce the effects
    ;; required to possibly satisfy the conditions of the schema.
    (if trigger
	(if (> (length trigger) 1)
	    ;; Want to wait for all such agenda entries to finish.
	    (cons :AND trigger)
	    (car trigger))
	t)))

;;; (filter-schemas candidates fn) applies fn to the list of candidate
;;; schema/bindings pairs.  fn is expected to take a schema and a
;;; binding alist as its arguments and to return t if can use the
;;; schema now, a trigger if could possibly use the schema after some
;;; upcoming changes to the plan state have been made, or nil if the
;;; schema cannot be used.  This function then returns two results. A
;;; list of those schemas that could be used now, and a list of those
;;; schemas that require some changes to the plan state, with their
;;; triggers.

(defun filter-schemas (candidates fn)
  (let (now-schemas to-wait-schemas)
    (dolist (candidate candidates (values (nreverse now-schemas)
					  (nreverse to-wait-schemas)))
      (let ((fn-result (funcall fn (car candidate) (cdr candidate))))
	(if (eql fn-result t)
	    (push candidate now-schemas)
	    (unless (null fn-result)
	      (push (cons candidate fn-result) to-wait-schemas)))))))

#+:undef
(defun schema-expansion (node schema)
  (list node (schema-nodes schema) (schema-orderings schema)))

#+:undef
(defun schema-effects-conditions (node schema)
  (list node (schema-effects schema) (schema-conditions schema)))

(defun expand-from-schema (node schema)
  (let ((subnodes
	 (db-request :EXPAND
	   node (schema-nodes schema) (schema-orderings schema))))
    (loop for (tag type pattern) in subnodes
	  unless (eq type 'dummy)
	  do (post-expand `(:EXPAND ,tag ,pattern)))
    t))

(defun schema-may-add-more-than-one-node-p (schema)
  (or (length>1 (schema-nodes schema))
      (and (length=1 (schema-nodes schema))
	   (node-iterators (first (schema-nodes schema))))))


(defvar *kp-debug-p* nil)

#+undef ; see support/ipc for the definition
(defun db-request (ident &rest args)
  (let ((answer (apply #'ipc-ask :DBIN ident args)))
    (whats-going-on "db-request:~%~S~%answer:~%~S"
                    (cons ident args) answer)
    answer))

(defun post-agenda (what &rest initargs)
  (let ((agenda-entry (apply #'make-agenda-rec :body what initargs)))
    (if *kp-debug-p* (format t "~%post-agenda - posting~%~A~%" agenda-entry))
    (when (eq (car what) :POISON-STATE)
      (db-request :POISON-CONTEXT (cdr what))
      (whats-going-on "Bad plan state:")
      (db-request :PRINT-DATABASE))
    (whats-going-on "Posting agenda:~%~S" agenda-entry)
    (ipc-send :AGENDAIN :AGENDA agenda-entry)))

(defun post-alternatives (agenda-rec)
  (whats-going-on "Posting alternative:~%~S~%" agenda-rec)
  (ipc-ask :AGENDAIN :ALT-AGENDA agenda-rec))

(defun post-agent-agenda (what)
  (whats-going-on "Posting agent agenda ~S~%" what)
  (ipc-send :AGENDAIN
	    :EVENT
	    (make-agenda-rec
	      :trigger t
	      :body what)))

(defun post-expand (body &rest initargs)
  ;; Need a more abstract way to deal with expand bodies /\/
  (assert (eq (car body) :expand))
  (let ((pattern (third body)))
    (apply #'post-agenda
	   body
	   :level (action-level (car pattern))
	   :branch-1 (db-request :number-of-expand-schemas pattern)
	   initargs)))

(defun post-achieve (body &rest initargs)
  ;; Need a more abstract way to deal with achieve bodies /\/
  (assert (eq (car body) :achieve))
  (let ((cond (second body)))
    (apply #'post-agenda
	   body
	   :level (effect-level (car (con-pattern cond)))
	   initargs)))

(defun post-condition (body &rest initargs)
  ;; Need a more abstract way to deal with condition bodies /\/
  (assert (eq (car body) :condition))
  (let ((cond (second body)))
    (apply #'post-agenda
	   body
	   :level (effect-level (car (con-pattern cond)))
	   initargs)))

(defun post-bind (body)
  (assert (eq (car body) :bind))
  (let ((psv-name (second body)))
    (assert (psvar-p psv-name))
    (let* ((possibles (get-bindings psv-name))
	   (n-possibles (if (consp possibles) (length possibles) 0)))
      ;; N.B. n-possible might be zero if no values were left when
      ;; the PSV was created.  Also possibles might be :already-bound.
      ;; In both cases (consp possibles) is false and we let branch-1 = 1.
      (post-agenda body
	:level :inf
	:branch-1 (max 1 n-possibles)))))


;;; /\/: If we need fns for PSVs and binding tuples, they should be
;;; in a module that's common to the DM and the KPs.

;; Predicate for determining if a symbol is a PSV.
(defun psvar-p (var)
  (and (symbolp var)
       (> (length (symbol-name var)) 4)
       (string= var "PSV-" :end1 4)))

;; Predicate for determining if a binding tuple has a Plan State var.
(defun var-p (tuple)
  (psvar-p (cadr tuple)))

;; Predicate for determining if a binding tuple has no value.
(defun undef-p (tuple)
  (eql (cadr tuple) :undef))

;; Returns the name of the PSVAR.
(defun psvar-name (var)
  (second var))

(defun sort-schema-alternatives (from)
  ;; We have a list of schemas, all wanting to be chosen, so use the criteria
  ;; of least number of unbound variables, and then first come first served to
  ;; order the list.  The "from" list has the form ((schema . bindings) ...).
  ;; /\/: Does t ever actually appear as the bindings?
  (let ((preferred-order
	 (stable-sort from
	   #'(lambda (x y)
	       (if (eql (cdr x) t)
		   t			;/\/ s.b. (not (eql (cdr y) t))
		   (if (eql (cdr y) t)
		       nil
		       (< (count-if #'undef-p (cdr x))
			  (count-if #'undef-p (cdr y)))))))))
    (if (eq *schema-selection-mode* :ask)
	(ask-user-to-pick-best-schema preferred-order)
      preferred-order)))

(defun ask-user-to-pick-best-schema (from)
  ;; /\/: Really we ought to go via a :USER agenda entry,
  ;;      but this will do for now.
  (ensure-loaded-ks :USER)
  (ks-user-schema-choice from))

(defun pick-alternative (reason)
  (ipc-ask :AGENDAIN :PICK-ALT reason))

#|
stage-manager takes the following form:
(stage-manager (<state-vars>) (<local-vars>)
               arg
               <stage 0 forms>
	       (stage <why>)
	       <stage 1 forms>
	       (stage <why>)
	       <stage 2 forms>
	       (stage <why>)
	       ...
	       (stage <why>)
	       <stage n forms>)

and converts it to the following form:
(block .stage-out.
  (let ((<state-var 1> (nth 0 (ag-info arg)))
	(<state-var 2> (nth 1 (ag-info arg)))
	...
	(<state-var n> (nth n-1 (ag-info arg)))
	<local-vars>)
  (tagbody
     (case (ag-stage arg)
       (0)
       (1 (go 1))
       (2 (go 2))
       ...
       (n (go n)))
   <stage 0 forms>
     (if <why> (progn
		 (stage-agenda arg 1 (list <state-vars>))
		 (return-from .stage-out.)))
   1
   <stage 1 forms>
     (if <why> (progn
		 (stage-agenda arg 2 (list <state-vars>))
		 (return-from .stage-out.)))
   2
   ...
     (if <why> (progn
		 (stage-agenda arg n (list <state-vars>))
		 (return-from .stage-out.)))
   <stage n forms>)))

|#

(defun stage-agenda (agenda-rec stage state)
  (setf (ag-stage agenda-rec) stage)
  (setf (ag-info agenda-rec) state)
  (ipc-send :AGENDAIN :AGENDA agenda-rec))

(defmacro stage-manager (state-vars local-vars agenda-rec &rest stages)
  (flet ((is-stage-form-p (x) (and (listp x) (eq (car x) 'stage))))
    (let ((stage-num 1)
	  (state-var 0)
	  var-bindings-list case-list stage-body-list)
      (dolist (var state-vars)
	(push (list var `(nth ,state-var (ag-info ,agenda-rec)))
	      var-bindings-list)
	(incf state-var 1))
      (setq var-bindings-list (append local-vars var-bindings-list))
      (dolist (form stages)
	(if (is-stage-form-p form)
	    (progn
	      ;; Do the case statement.
	      (push (list stage-num (list 'go stage-num)) case-list)
	      ;; Do the staging.
	      (push (list 'if (cadr form)
			  (list 'progn
				(list 'stage-agenda
				      agenda-rec stage-num
				      `(list ,@state-vars))
				(list 'return-from '.stage-out.)))
		    stage-body-list)
	      (push stage-num stage-body-list)
	      (incf stage-num 1))
	    (push form stage-body-list)))
      `(block .stage-out.
	(let ,var-bindings-list
	  (tagbody
	     (case (ag-stage ,agenda-rec)
	       (0)
	       ,@(nreverse case-list))
	     ,@(nreverse stage-body-list)))))))

(defun some-work-to-do-p (or-trees)
  "Returns t if there are some bindings, restrictions, or linkings to be made,
or some deletors to remove."
  (some #'(lambda (or-tree)
	    (or (> (length or-tree) 1)
		(notevery #'(lambda (x) (eql (car x) :UPDATE))
			  (qa:or-info-rec-actions (car or-tree)))))
	or-trees))

;; Traverses or-tree replacing all the satisfaction methods of :UPDATE actions
;; with method. DESTRUCTIVE.
(defun set-update-contributor-satisfaction-method (or-tree method)
  (mapc #'(lambda (x)
	    (if (listp x)
		(set-update-contributor-satisfaction-method x method)
		(let ((update-action (assoc :UPDATE (or-info-rec-actions x))))
		  (if update-action
		      (mapc #'(lambda (x)
				(setf (cdr x) method))
			    (cadr update-action))))))
	or-tree))

(defun pattern-function (pattern)
  "Returns the function (the first fixed-word) of the pattern"
  (car pattern))

(defun pattern-args (pattern)
  "Returns a list of the arguments to the function of the pattern"
  (cdr pattern))

(defun post-to-reachieve-cond-else-poison (cond contributors alt-p)
  "Posts the reachievment of the cond as an alternative if alt-p, else posts
the reachievement of the cond as something else to do, as long as methods are
still available, otherwise POISON-PLAN-STATE"
  (let ((max-satisfaction-method 0))
    ;; Find out the most work that has been done to satisfy this condition.
    (mapc #'(lambda (contributor)
	      (setq max-satisfaction-method
		    (max max-satisfaction-method
			 (tgm-get-satisfaction-method-level
			  (cdr contributor)))))
	  contributors)
    (if (= (tgm-get-satisfaction-method-level
	    (tgm-condition-type-to-allowable-satisfaction-methods
	     (con-type cond)))
	   max-satisfaction-method)
	;; Already done the maximum amount of work, so poison plan state,
	;; unless this is just an alternative.
	(unless alt-p
	  (post-agenda `(:POISON-STATE :COULD-NOT-RESATISFY-THE-CONDITION
			 ,cond)))
	(if alt-p
	    (post-alternatives (make-agenda-rec
				:body `(:ACHIEVE ,cond)
				:trigger (make-wait-for-effect-trigger
					  cond nil)))
	    (if (eql (con-type cond) 'achievable)
		(post-achieve `(:ACHIEVE ,cond)
		     :trigger (make-wait-for-effect-trigger cond nil))
		(post-condition `(:CONDITION ,cond)
		     :trigger (make-wait-for-effect-trigger cond nil)))))))

(defun post-or-agenda-record (cond contributors for-satisfying-p or-tree)
  (post-agenda `(:OR ,cond ,contributors ,for-satisfying-p ,@or-tree)
     :level :inf
     :branch-1 (length or-tree)))

(defun assign-or-level (ag)
  ;; Called by KS-OR.
  (setf (ag-level ag) :inf)
  ag)

;; Goes through all the interactions in interact-list ignoring the :ALWAYS and
;; :ALWAYS-CONFLICT, and posting KS-ORs for any :INTERACTs. After that it will
;; commit to adding the effect, unless the effect matches an :ALWAYS or
;; :ALWAYS-CONFLICT.
(defun handle-adding-effect-interaction (effect interact-list)
  (let ((add-effect-p t))
    (if (consp interact-list)
	(dolist (interaction interact-list)
	  (case (car interaction)
	    (:INTERACT
	     (let ((qa-result (tgm-interact-qa-result interaction))
		   (cond (cdar (tgm-interact-cond interaction)))
		   (contribs (cadr (tgm-interact-cond interaction))))
	       (if qa-result
		   (post-or-agenda-record cond contribs nil qa-result)
		   ;; No way to resolve interaction, so post to reachieve
		   ;; the cond, unless this is a condition type that rules
		   ;; out being reachieved, or has already been satisfied by
		   ;; the maximum method allowed.
		   (if (member (con-type cond)
			       '(supervised unsupervised only_use_if))
		       (progn
			 (abort-add-cond-or-effect)
			 (return-from handle-adding-effect-interaction nil))
		       (post-to-reachieve-cond-else-poison
			cond contribs nil)))))
	    (:ALWAYS
	     (let ((bindings (tgm-always-bindings interaction)))
	       (unless bindings (setq add-effect-p nil))))
	    (:ALWAYS-CONFLICT
	     (let ((bindings (tgm-always-bindings interaction)))
	       (unless bindings
		 (MONITOR :KP-1 (:DETAILED "~W conflicts with ~W~%"
					   effect (tgm-always-effect
						   interaction)))
		 (setq add-effect-p nil))))))
	(setq add-effect-p interact-list))
    (if add-effect-p
	(really-add-effect effect)
	(abort-add-cond-or-effect))
    interact-list))

(defun find-contributors (or-rec)
  "Searches actions field of or-info-rec for an UPDATE entry, and returns the
arguments."
  (second (find :UPDATE (qa:or-info-rec-actions or-rec)
		:key #'car)))

;; Goes through all the interactions in interact-list working out all the
;; KS-ORs possible. Returns t if added the cond, or nil if couldn't.
(defun handle-adding-condition-interaction (cond interact-list)
  (let ((qa-result (tgm-satisfy-qa-result interact-list)))
    (if (consp qa-result)
	(if (some-work-to-do-p qa-result)
	    (progn
	      (post-or-agenda-record cond nil t qa-result)
	      (really-add-condition cond)
	      t)
	    (progn
	      (really-add-condition cond)
	      (db-request :UPDATE-CONTRIBUTORS
			  cond (find-contributors (caar qa-result)))
	      t))
	(if (or qa-result (eql (con-type cond) 'supervised))
	    (progn
	      (really-add-condition cond)
	      t)
	    (progn
	      (abort-add-cond-or-effect)
	      nil)))))

(defun try-adding-effect (effect)
  (let (result)
    (setq result  (car (db-request :CHECK-TOME effect)))
    (if *kp-debug-p* (format t "~&try-adding-effect ~A returns" effect))
    (if *kp-debug-p* (format t "~%~A~%" result))
    result))

(defun really-add-effect (effect)
  (if *kp-debug-p* (format t "~&really-add-effect ~A~%" effect))
  (db-request :ADD-TOME effect))

(defun try-adding-condition (cond &optional bindings contributors-to-ignore)
  (let ((cond-temp cond)
        result)
    (if contributors-to-ignore
        (setq cond-temp (append cond-temp
                                (list (cons :NOT contributors-to-ignore)))))
    (setq result (db-request :CHECK-GOST cond-temp bindings))
    (if *kp-debug-p* (format t "~&try-adding-condition ~A returns" cond))
    (if *kp-debug-p* (format t "~%~A~%" result))
    result))

#+:undef
(defun try-adding-condition (cond &optional bindings contributors-to-ignore)
  (let ((result
	 (db-request :CHECK-GOST
	   (if contributors-to-ignore
	       (add-contributors-to-ignore-to-cond cond contributors-to-ignore)
	     cond)
	   bindings)))
    (if *kp-debug-p* (format t "~&try-adding-condition ~A returns" cond))
    (if *kp-debug-p* (format t "~%~A~%" result))
    result))

(defun add-contributors-to-ignore-to-cond (cond contribs-to-ignore)
  (destructuring-bind
      (type pat val at-node-end &optional contrib-list)
      cond
    (assert (null contrib-list))
    (list type pat val at-node-end (cons :NOT contribs-to-ignore))))

(defun really-add-condition (cond)
  (if *kp-debug-p* (format t "~&really-add-condition ~A~%" cond))
  (db-request :ADD-GOST cond))

(defun abort-add-cond-or-effect nil
  (if *kp-debug-p* (format t "~&*** abort-add-cond-or-effect ***~%"))
  (db-request :ABORT-ADD-TOME-OR-GOST))

#+:undef ; /\/ does anyone call this?
(defun do-alt-work (alt)
  (dolist (binding (alt-bindings alt))
    (set-binding (car binding) (cdr binding)))
  (dolist (restrict (alt-restrictions alt))
    (add-restriction-to-psv (car restrict) :NOT (cdr restrict)))
  (dolist (link (alt-linkages alt))
    (if (null (db-request :ADD-LINK (car link) (cadr link)))
	(cerror "Will continue" "Failed to add a link when told we can!"))))

;; Just a test staging rule!
(defun if-want-real-time-response-p ()
  nil)

;; Looks through a form for any mention of PSVs, and returns those PSV's
;; mentioned.
(defun find-cond-psvs (cond &optional accumulator)
  (cond ((psvar-p cond)
	 (cons cond accumulator))
	((consp cond)
	 (find-cond-psvs (car cond) (find-cond-psvs (cdr cond) accumulator)))
	(t accumulator)))

;; If the condition mentions a Plan State var, then we return a trigger that
;; requires the PS var to be bound. Handles multiple PS vars.
(defun trigger-for-cond (cond)
  (let ((psvs-found (find-cond-psvs cond)))
    (or (if psvs-found
	    (if (= (length psvs-found) 1)
		(cons :BIND (car psvs-found))
		(cons :AND (mapcar #'(lambda (x) (cons :BIND x))
				   psvs-found))))
	t)))

;; Takes a list of conditions and returns only those that are not
;; only_use_for_query or only_use_if, minus their condition type.
(defun get-conditions-for-triggering (conds)
  (loop for c in conds
	unless (or (eq (con-type c) 'only_use_for_query)
		   (eq (con-type c) 'only-use-if))
	collect
	 (case (con-type c)
	   ((achievable) (cdr (con-minus-after-point c)))
	   (     t       (cdr c)))))

;; Adds the <type> of restriction to <psv> with <value>.
(defun add-restriction-to-psv (psv type value)
  (db-request :ADD-VAR-RESTRICTION psv type value))

(defun get-bindings (var)
  (db-request :GET-VAR-BINDINGS var))

(defun set-binding (var value)
  (db-request :SET-VAR-BINDING var value))

(defun instantiate-schema (schema)
  (db-request :INSTANTIATE-SCHEMA schema))

(defun get-effect-schemas (patt)
  (ipc-with-own-copy
    (result (db-request :GET-EFFECT-SCHEMAS patt)
	    (deep-copy result))))

(defun process-node-end (node parent-node parent-p)
  ;; The atnode number can be either :self or a number.
  ;; If it is :self, then replace with node
  ;; If it is a number and parent-p is nil then replace with node
  ;; Otherwise append the number to node and use that.
  ;; DESTRUCTIVE
  (if (or (eql (ds-node-tag node) :self)
	  (null parent-p))
      (setf (ds-node-tag node) parent-node)
      (setf (ds-node-tag node)
	    (intern (concatenate 'string (symbol-name parent-node) "-"
				 (prin1-to-string (ds-node-tag node)))))))

;;; Process-schema
;;
;; Takes an instantiated schema and processes all the fields. IE adds the
;; effects, conditions etc.  Called from KS-ACHIEVE and KS-EXPAND.
;; If parent-p is t then this schema is from an expansion, and node is the
;; parent node being expanded. If parent-p is nil, then this schema is from
;; an achieve, and node is the added node.
;; Return t if okay, otherwise nil indicates a problem.

(defun process-schema (schema node parent-p)
  (and (add-effects schema node parent-p)
       (add-conditions schema node parent-p)
       (add-time-windows schema node parent-p)
       (add-resources schema node parent-p)))
    
(defun add-effects (schema node parent-p)
  ;;; ADDING EFFECTS
  (let (interact)
    ;; For each effect, ask the DM to add it, and if okay continue, else if
    ;; there are some interactions with GOST entries then post the interactions
    ;; as KS-OR records, and if the condition types are appropriate, post the
    ;; reachievement of the conditions as alternatives. If we cannot add an
    ;; effect then POISON PLAN.
    (dolist (effect (append (schema-only-use-for-effects schema)
			    (schema-effects schema))
	     t)
      ;; Change the atnode number.
      (process-node-end (effect-at-node-end effect) node parent-p)
      (whats-going-on "Trying to add effect~%")
      (setq interact (try-adding-effect effect))
      (if (handle-adding-effect-interaction effect interact)
	  (whats-going-on "Okay to add effect.")
	  ;; Can't add this effect so POISON PLAN STATE.
	  (progn
	    (whats-going-on "Couldn't add effect.")
	    (post-agenda `(:POISON-STATE :COULD-NOT-ADD-EFFECT ,effect
			   :WHEN :EXPANDING ,node))
	    (return-from add-effects nil))))))

(defun add-conditions (schema node parent-p)
  ;;; ADDING CONDITIONS
  (let (interact)
    ;; For each condition, ask the DM to add it, and if okay continue, else if
    ;; there are some interactions with TOME entries then post the interactions
    ;; as KS-OR records, unless the condition type warrents some other action.
    ;; If we cannot add a condition then POISON PLAN.
    (dolist (cond (schema-conditions schema) t)
      ;; Change the atnode number.
      (process-node-end (con-at-node-end cond) node parent-p)
      ;; If this is a supervised condition, then the fifth item of cond
      ;; is a list of contributing nodes, which need to be changed to the
      ;; actual nodes.
      (when (eq (con-type cond) 'supervised)
	(mapc #'(lambda (n)
		  (process-node-end n node parent-p))
	      (con-contributors cond)))
      (case (con-type cond)
	(supervised
	 ;; Add GOST entry regardless.
	 (whats-going-on "Adding ~:@(~A~) condition~%"
			 (symbol-name (con-type cond)))
	 (setq interact (try-adding-condition cond))
	 (handle-adding-condition-interaction cond interact))
	(only_use_if
	 ;; Fail if cannot add GOST entry.
	 (whats-going-on "Adding ~:@(~A~) condition~%"
			 (symbol-name (con-type cond)))
	 (setq interact (try-adding-condition cond))
	 (unless (handle-adding-condition-interaction cond interact)
	   ;; This should not happen since we check only_use_ifs before
	   ;; expanding the schema.
	   (post-agenda `(:POISON-STATE :COULD-NOT-ADD-COND ,cond
			  :WHEN :EXPANDING ,node))
	   (return-from add-conditions nil)))
	(only_use_for_query
	 (whats-going-on "Adding ~:@(~A~) condition~%"
			 (symbol-name (con-type cond)))
	 (setq interact (try-adding-condition cond))
	 (unless (handle-adding-condition-interaction cond interact)
	   ;; Failed to add the only_use_for_query cond, so first check to
	   ;; see if there are any outstanding agenda-entries that could
	   ;; possibly provide an effect that could satisfy the cond, and if
	   ;; so post to add this cond with a trigger to wait for the effect,
	   ;; otherwise fail.
	   (if (possibly-satisfying-plan-state-changes-p cond nil)
	       (post-condition `(:CONDITION ,cond)
		   :trigger (make-wait-for-effect-trigger cond nil))
	       (progn
		 (post-agenda `(:POISON-STATE :COULD-NOT-ADD-COND ,cond
				:WHEN :EXPANDING ,node))
		 (return-from add-conditions nil)))))
	(achievable
	 (let ((c (con-minus-after-point cond)))
	   (post-achieve `(:ACHIEVE ,c)
			 :trigger (trigger-for-cond c))))
	(unsupervised
	 ;; This just checks that the condition holds by the time
	 ;; that most of the work in expanding and adding nodes is
	 ;; done.
	 (post-condition `(:CONDITION ,cond) :trigger (trigger-for-cond cond)))
	(compute
	 ;; Evaluate p = v as soon as all the vars in p have values.
	 ;; If they all already have values, we can do it now.
	 ;; /\/: Note that trigger-for-cond doesn't have to be given a
	 ;; condition per se.
	 (let ((trigger (trigger-for-cond (con-pattern cond)))
	       (p (con-pattern cond))
	       (v (con-value cond)))
	   (if (eq trigger t)
	       ;; Eval now
	       (unless (db-request :EVALUATE-COMPUTE p v) 
		 (post-agenda `(:POISON-STATE :COMPUTE-FAILED ,p = ,v))
		 (return-from add-conditions nil))
	     ;; Eval later, when the PSVs in p have been bound.
	     (post-agenda `(:COMPUTE ,cond)
			  :trigger trigger
			  :level 0))))
	(otherwise
	 (cerror "Will continue"
		 "Condition type ~S not handled" (con-type cond)))))))

(defun add-time-windows (schema node parent-p)
  (dolist (time-window (schema-time-windows schema) t)
    (let ((from (time-constraint-start time-window))
	  (to (time-constraint-end time-window))
	  (min (time-constraint-min time-window))
	  (max (time-constraint-max time-window)))
      (unless (eql from :ABST0)
	(process-node-end from node parent-p))
      (process-node-end to node parent-p)
      (whats-going-on "Adding time window of (~A, ~A) between ~A and ~A"
		      min max from to)
      (unless (db-request :ADD-TIME-WINDOW from to min max)
	(post-agenda `(:POISON-STATE :INVALID-TIME-WINDOWS))
	(return-from add-time-windows nil)))))

(defun add-resources (schema node parent-p)
  (cond ((or (null (schema-resources schema))
	     (db-request :ADD-RESOURCES
			 (schema-resources schema) node parent-p))
	 t)
	(t
	 (post-agenda `(:POISON-STATE :INVALID-RESOURCES))
	 nil)))


;;; ---------------------------- Change History ----------------------------
;;; (Who)   (When) 		     (What)
;;;
