;;;; File: viewer.lsp
;;; SCCS Version: %W%
;;; Contains: AutoCAD-based drawing package for the O-Plan system
;;; Authors: Brian Drabble and Jeff Dalton
;;; Created: Wed Mar  4 17:00:03 1992
;;; Updated: Mon Jul 10 20:25:22 1995 by Jeff Dalton
;;; Release Version: %Y%
;;; Copyright: (c) 1992, 1993, AIAI, University of Edinburgh
;;; This material may be reproduced by or for the U.S. Government pursuant
;;; to the copyright license under the clause at DFARS 252.227-7032
;;; (June 1975) -- Rights in Technical Data and Computer Software (Foreign).

;;; This module is given a plan or world description and produces output
;;; in various forms: text to screen, text to file, AutoCAD, etc.

(in-package :plan-world-viewer)

;;; AutoCAD world-view
      
(defvar *acad-file* nil)

(defparameter *acad-world-preds*
  '(;; Space-platform preds
    joint_at truss_at tube_at module_at radiator_at solar_panel_at
    antenna_at
    ;; PRECiS / Pacifica preds
    at
    ))

(defun acad-world-pred-p (name)
  (and (symbolp name)
       (member (symbol-name name) *acad-world-preds*
	       :key #'symbol-name
	       :test #'string-equal)))

(defun output-acad-world (p-v-pairs)
  ;;
  ;; output the simulation of a particular node 
  ;;
  (let ((*print-case* :downcase)
	(*print-length* nil)
	(*print-level* nil)
	(*package* *viewer-io-package*)
	(foundflag nil))
    (with-open-file (*acad-file* (temp-filename "tfin.lsp")
				 :direction :output)
      (format *acad-file* "(defun C:TFIN () ~%")
      (dolist (p-v p-v-pairs)
	(let ((pattern (car p-v))
	      (value (cadr p-v))
	      pred)
	  (when (acad-world-pred-p (car pattern))
	    (setq foundflag t)
	    ;; /\/ Drop-at is really only for space platform.
	    (setq pred (drop-at (car pattern)))
	    (setq pattern (cons pred (cdr pattern)))
	    (cond ((string-equal pred 'joint)
		   ;; /\/ Remove?  Really?  Remove _more than one_???
		   (setq pattern (remove (cadr pattern) pattern)))
		  ((string-equal pred 'at)
		   ;; Pacifica.  We're outputing a Lisp fn call so
		   ;; words after the 1st must be strings.  And since
		   ;; we know they're symbols now ... /\/
		   (setq pattern
			 `(,(car pattern) ,(string (cadr pattern))
			                  ,(string value)))))
	    (write pattern :stream *acad-file*)
	    (terpri *acad-file*))))
      (format *acad-file* "(command ~Credrawall~C) (princ) ~%" #\" #\")
      (format *acad-file* ") ~%"))
    ;; 
    ;; Now inform the user the file is ready for viewing
    ;;
    (if foundflag
	(report-action "Autocad file is ready for viewing.")
	(report-action "Autocad file is empty."))
    (sound-alarm)
  )
)

;;; AutoCAD plan view

;;; At this point, we have a list of node-description structures.  This
;;; can be considered a graph, but not one we can draw directly.  So we
;;; proceed in two steps:
;;;
;;;   1. Transform the list into whatever graph we want to draw.
;;;   2. Produce AutoCAD instructins to draw the transformed graph.
;;;
;;; Step 2 should be able to handle a number of different graphs, not
;;; only the ones we produce.  To keep step 2 sufficiently abstract,
;;; it works with a list of vertex structures.  We'll avoid the term
;;; "node" to avoid any confusion with plan network nodes.
;;;

(defvar *vertices* nil)			;to aid debugging, etc.

;;; The vertex struct

(defstruct (vertex (:conc-name v-))
  name		;a symbol
  begin-suc	;names of immediate left-end successors 
  end-suc	;names of immediate right-end successors
  type
  pattern
  offset	;length of longest path from a unique start vertex.
  x-pos
  y-pos)

(defvar *vertex-lookup-list*)		;for looking up vertices

(defun find-vertex (name &optional (vertices *vertex-lookup-list*))
  (find name vertices :key #'v-name))


;;; Top level of the AutoCAD drawer

(defun output-autocad-plan (nodes)
  (let ((vertices nil))
    ;;
    ;; Set max plan level (depth) and slice off the nodes and links below it.
    ;; Note that we get back copies of the nodes.  We can modify the copies
    ;; without fear...
    ;;
    (setq nodes (slice-to-level nodes))
    ;;
    ;; Remove the link between the two ends of each node.
    ;;
    (mapc #'remove-self-link nodes)
    ;;
    ;; Convert nodes to vertices
    ;;
    (let ((*node-lookup-list* nodes))	;for find-node to use
      (setq vertices (convert-nodes-to-graph nodes)))
    ;;
    ;; Output the plan graph as an AutoCAD network.
    ;;
    (output-graph vertices)
    ;; 
    ;; Now inform the user the file is ready for viewing
    ;;
    (report-action "Autocad file is ready for viewing.")
    (sound-alarm)))

(defun remove-self-link (n)
  (setf (nd-begin-suc n)
	(remove (ne-tag 'end_of (nd-node-name n))
		(nd-begin-suc n)
		:test #'equal))
  (setf (nd-end-pre n)
	(remove (ne-tag 'begin_of (nd-node-name n))
		(nd-end-pre n)
		:test #'equal))
  n)

;;; From plan to graph

;;; For vertices, we record only successors, leaving predecessors
;;; implicit.

;;; Non-parent nodes are converted directly to vertices.  A vertex is
;;; drawn as a single box but the two ends are distinguished just as
;;; they are for nodes.  For instance, a link from the end_of v1 to
;;; begin_of v2 will be drawn from the right side of the v1 box to the
;;; left side of the v2 box.

;;; Parent nodes are either eliminated completely or converted to two
;;; "dummy" vertices, one for each end of the parent.

;;; A dummy is drawn as a dot and serves only to join a number of
;;; lines together.  Like other vertices, dummies have two ends, but
;;; it doesn't make very much difference which end is used.  By
;;; convention, all the successors of a dummy are recorded as end_of
;;; successors, and all references to a dummy from predecessors are
;;; references to the begin_of end.  (One way to remember this is to
;;; think of all lines as being drawn from a predecessor to a successor,
;;; with successors further to the right.  Think of lines being drawn
;;; to the left side of a dot or from the right side.)

;;; A parent is eliminated if it's possible to use unique child
;;; nodes instead.  This happens if the begin_of the parent has one
;;; successor and the end_of the parent has one predecessor -- other
;;; than redundant links with children and the link from the begin to
;;; the end of the parent.  For instance, suppose we have something
;;; like this:

;;;   begin_of node-3 ---> node-3-1 ---> node-3-2 ---> end_of node-3

;;; Whenever we see begin_of node-3 as a successor, we can point
;;; to begin_of node-3-1 instead; and whenever end_of node-3 is a
;;; predecessor, we can use end_of node-3-2.

;;; /\/: Finish explanation ...

;;; The rename-table has ne-tags as both keys and values.

;;; The reason for our treatment of parents is that we don't want to
;;; see nodes that are expanded, only their expansions.  

;;; Note that what counts as a parent will depend on the level to
;;; which we're viewing the plan.  For instance, if we're viewing
;;; levels 1 and 2, node-3-2 will _not_ be considered a parent even
;;; if it has node-3-2-1, node-3-2-2, etc as children in the full plan.

(defun convert-nodes-to-graph (nodes)
  (let* ((rename-table (make-hash-table :test #'equal))
	 (vertices
	  (mapcan #'(lambda (node)
		      (node-to-vertices node rename-table))
		  nodes)))
    (rename-vertex-successors vertices rename-table)))

(defun node-to-vertices (node rename-table) ; -> list of vertices
  (if (parent-node-p node)
      (convert-parent-to-vertices node rename-table)
    (list
      (copy-node-to-vertex node))))

(defun copy-node-to-vertex (node)
  (make-vertex
    :name (nd-node-name node)
    :begin-suc (nd-begin-suc node)
    :end-suc (nd-end-suc node)
    :type (nd-node-type node)
    :pattern (nd-pattern node)))

(defun convert-parent-to-vertices (node rename-table) ; -> list of vertices
  ;; If the fan-out and fan-in are both 1, we can eliminate the parent
  ;; entirely.  The test for this is complicated by the possibility of
  ;; redundant links to child ends (so that we have to figure our which
  ;; child ends are minimal or maximal) and by the link between the two
  ;; ends of the parent.  Here we assume the link between the two ends
  ;; has already been removed.
  (let ((min-c (minimal-child-ends node))
	(max-c (maximal-child-ends node)))
    (if (and (= 1 (length min-c) (length max-c))
	     (null (non-child-ends node (nd-begin-suc node)))
	     (null (non-child-ends node (nd-end-pre node))))
	;; Eliminate parent in favor of min and max child ends
	(eliminate-parent node rename-table (first min-c) (first max-c))
      ;; Split parent into two dummies.
      (convert-parent-to-dummies node rename-table))))

(defun eliminate-parent (node rename-table first-child-end last-child-end)
  ;; Check ends, assuming a correct plan.
  (assert (eq (ne-tag-end first-child-end) 'begin_of))
  (assert (eq (ne-tag-end last-child-end) 'end_of))
  ;; We'll rename begin_of node to first-child-end.
  (setf (gethash (ne-tag 'begin_of (nd-node-name node))
		 rename-table)
	first-child-end)
  ;; The only successor references to end_of node should be from
  ;; begin_of node or last-child-end.  The reference from begin_of
  ;; node should already have been eliminated, but it doesn't matter
  ;; because we're eliminating node completely.  So what's left is
  ;; to remove the entry in the successors of last-child-end.  Then
  ;; we can put :poison in the rename table to let us detect any
  ;; unexpected references to end_of node.
  (let ((last-child-node (find-node (ne-tag-node last-child-end))))
    (setf (nd-end-suc last-child-node)
	  (remove (ne-tag 'end_of (nd-node-name node))
		  (nd-end-suc last-child-node)
		  :test #'equal))
    (setf (gethash (ne-tag 'end_of (nd-node-name node))
		   rename-table)
	  :poison)
    ;; Add the successors of end_of node to the successors of last-child-end.
    ;; Note that last-child-end might be the end of another parent that will
    ;; be eliminated.  If it has already been eliminated, then it's too late
    ;; to add to its successors; we'd have to go one step further back.
    ;; Fortunately, this cannot happen.  Since parents appear in the list
    ;; of nodes before their children, and we're processing nodes in that
    ;; order, and last-child is a child of the node we're processing now,
    ;; it must be that last-child has not yet been processed, much less
    ;; eliminated.  However, if it had been eliminated, references to it
    ;; would have been poisoned (as above), so we can check.
    (assert (not (eq (gethash last-child-end rename-table) :poison)))
    ;; Let's assume order doesn't matter (for now) and use UNION.
    (setf (nd-end-suc last-child-node)
	  (union (nd-end-suc last-child-node)
		 (nd-end-suc node))))
  ;; Return the empty list (ie, no vertices for the parent).
  '())

(defun convert-parent-to-dummies (node rename-table)
  ;; Make a dummy vertex for each end of the parent
  (let* ((begin-name (concat-pw-symbol 'begin_of- (nd-node-name node)))
	 (end-name (concat-pw-symbol 'end_of- (nd-node-name node)))
	 (begin-vertex
	  (make-dummy-vertex 'begin_of begin-name node))
	 (end-vertex
	  (make-dummy-vertex 'end_of end-name node)))
    ;; Rename begin_of node to begin_of begin-name
    (setf (gethash (ne-tag 'begin_of (nd-node-name node))
		   rename-table)
	  (ne-tag 'begin_of begin-name))
    ;; Rename end_of node to begin_of end-name
    (setf (gethash (ne-tag 'end_of (nd-node-name node))
		   rename-table)
	  (ne-tag 'begin_of end-name))
    ;; Return a vertex for each end
    (list begin-vertex end-vertex)))

(defun make-dummy-vertex (end name node)
  (make-vertex
    :name name
    :begin-suc nil			;ignored
    :end-suc
      (ecase end
	(begin_of (nd-begin-suc node))
	(end_of (nd-end-suc node)))
    :type 'dummy
    :pattern (nd-pattern node)))

(defun rename-vertex-successors (vertices rename-table)
  (dolist (v vertices)
    (setf (v-begin-suc v)
	  (mapcar #'(lambda (ne-tag)
		      (rename-successor ne-tag rename-table))
		  (v-begin-suc v)))
    (setf (v-end-suc v)
	  (mapcar #'(lambda (ne-tag)
		      (rename-successor ne-tag rename-table))
		  (v-end-suc v))))
  vertices)

(defun rename-successor (ne-tag rename-table)
  (let ((new-ne-tag (gethash ne-tag rename-table :not-there)))
    (cond ((eq new-ne-tag :poison)
	   (error "Reference to poisoned end ~S" ne-tag))
	  ((eq new-ne-tag :not-there)
	   ne-tag)
	  (t
	   (rename-successor new-ne-tag rename-table)))))

;;; Parent/child operations on node-descriptions

(defun parent-node-p (node)
  ;; Is this node a parent?  We say it is if the begin end has a
  ;; successor that has a name that is an extension of the node's.
  ;; E.g., node-3-2 is an extension of node-3.
  (let ((node-name (nd-node-name node)))
    (some #'(lambda (ne-tag)
	      (child-name-p (ne-tag-node ne-tag) node-name))
	  (nd-begin-suc node))))

(defun child-name-p (name parent-name)
  ;; Is parent-name an initial segment of name (e.g. node-3 and node-3-2)?
  ;; Note that for e.g. node-3, we must exclude node-3 and node-3-2-1 as
  ;; the names of children.  Node-3 will be excluded by the length check.
  (let ((n (symbol-name name))
	(pn (symbol-name parent-name)))
    (and ;; Compare lengths as a quick way to eliminate some candidates.
         (< (length pn) (length n))
	 ;; Search for parent-name in name
	 (let ((p (search pn n)))
	   (and ;; Parent-name must be an initial segment of name
	        (eql p 0)
		;; And there must be only one "-" after that initial segment.
		(not (find '#\-
			   n
			   :start (1+ (length pn)))))))))

(defun get-parent-name (name) ; -> node name or nil if no parent
  ;; E.g. for node-3-2, return node-3.
  (let* ((n (symbol-name name))
         (p (position '#\- n :from-end t)))
    (if (and p
	     ;; Must be one "-" left in the parent name
             (find '#\- n :from-end t :end (1- p)))
        (values (intern (subseq n 0 p) *viewer-io-package*))
      nil)))

(defun child-ends (node ne-tags)
  (let ((node-name (nd-node-name node)))
    (loop for ne-tag in ne-tags
	  when (child-name-p (ne-tag-node ne-tag) node-name)
	  collect ne-tag)))

(defun non-child-ends (node ne-tags)
  (let ((node-name (nd-node-name node)))
    (loop for ne-tag in ne-tags
	  when (not (child-name-p (ne-tag-node ne-tag) node-name))
	  collect ne-tag)))

(defun minimal-child-ends (node)
  (let ((candidates (child-ends node (nd-begin-suc node))))
    ;; A minimal child end is one that has no predecessors among
    ;; the other candidates.  Since end_of C would have begin_of C
    ;; among its predecessors, we know the minimal ends must be
    ;; begin ends (a shortcut).
    (remove-if-not
      #'(lambda (candidate)
	  (and (eq (ne-tag-end candidate) 'begin_of)
	       (notany #'(lambda (can-pre)
			   (member can-pre candidates :test #'equal))
		       (nd-begin-pre (find-node (ne-tag-node candidate))))))
      candidates)))

(defun maximal-child-ends (node)
  (let ((candidates (child-ends node (nd-end-pre node))))
    ;; A maximal child end is one that has no successors among
    ;; the other candidates.  Since begin_of C would have end_of C
    ;; among its successors, we know the maximal ends must be
    ;; end ends (a shortcut).
    (remove-if-not
      #'(lambda (candidate)
	  (and (eq (ne-tag-end candidate) 'end_of)
	       (notany #'(lambda (can-suc)
			   (member can-suc candidates :test #'equal))
		       (nd-end-suc (find-node (ne-tag-node candidate))))))
      candidates)))


;;; Now some utilities used by the drawer.

;;; Find distances ("offsets") from the start vertex by finding the
;;; number of steps in the longest path along successor links from the
;;; start to each vertex.  Note that vertices have symbols as names
;;; and that here we do not distinguish between the two ends of a
;;; vertex.

;;; Find-longest-path-lengths is defined elsewhere.

(defun set-vertex-offsets (vertices start-name)
  (labels ((name->vertex (name)
	     (find name vertices :key #'v-name))
	   (v-successors (vertex)
	     (mapcar #'(lambda (ne-tag)
			 (name->vertex (ne-tag-node ne-tag)))
		     (append (v-begin-suc vertex) (v-end-suc vertex)))))
    ;; Note that the table keys will be the vertex structs, not their names.
    (let ((distance-table
	   (find-longest-path-lengths-while-detecting-known-bug
	      (name->vertex start-name)
	      #'v-successors)))
      (maphash #'(lambda (vertex distance)
		   (setf (v-offset vertex) distance))
	       distance-table))))

(defun find-longest-path-lengths-while-detecting-known-bug (start successor-f)
  (handler-case (find-longest-path-lengths start successor-f)
    (error (c)
      (format *debug-io* "~&~%Something is wrong:~%~A~%~%" c)
      (ecase (menu-request
	       '("-heading" "Known bug: can't draw this plan"
		 "-line"
		 "Abandon this view=:go"
		 "Break in=:break"))
	(:go
	 (exit-view))
	(:break
	 (break "AutoCAD view")
	 (exit-view))))))
		


;;; Finally, the graph output routine.

(defparameter *dummy-types* '(dummy))

(defun output-graph (vertices)
  ;;
  ;; Find distances ("offsets") from begin_of_node-1
  ;;
  (set-vertex-offsets vertices 'node-1)
  ;;
  ;; Now layout entries on the page
  ;;
  (let ((*print-case* :downcase)
	(*print-length* nil)
	(*print-level* nil)
	(*package* *viewer-io-package*)
	(*vertex-lookup-list* vertices)		;for find-vertex
	level-list
	maxwidth
	maxlength
	)
     (with-open-file (*acad-file* (temp-filename "tfin.lsp")
				  :direction :output)
       (format *acad-file* "(defun C:TFIN () ~%")
       ;;
       ;; Group vertices by offset.  Level-list will be a list of lists,
       ;; where the i-th sublist contains all the vertices with offset i.
       ;; /\/: Why not make it a vector?
       ;;
       (setq level-list (group-by-numeric-key vertices #'v-offset))
       ;;
       ;; calculate the maximum width at any level
       ;;
       ;; /\/: Fix this when reduce is available with :key
       (setq maxwidth
	     (reduce #'max
		     (mapcar #'length level-list)
		     :initial-value 0))
       ;;
       ;; we now have the max values in both directions
       ;;
       (setq maxwidth (* maxwidth 200))
       (setq maxlength (* (length level-list) 150))
       (format *acad-file*
	       "(command \"LIMITS\" \"0,0\" \"~S,~S\") ~
                (command \"ZOOM\" \"ALL\") ~%"
	       maxlength maxwidth)
       ;;
       ;; Set x-pos and y-pos for each vertex
       ;;
       (let ((x-position 20))
	 (dolist (worklist level-list)
	   (let ((y-inc (/ maxwidth (1+ (length worklist))))
		 (l-count 0))
	     (dolist (vertex worklist)
	       (let ((y-position (* (1+ l-count) y-inc)))
		 (setf (v-x-pos vertex) x-position)
		 (setf (v-y-pos vertex) y-position))
	       (incf l-count)))
	   (incf x-position 150)))
       ;;
       ;; Output a description of each vertex
       ;;
       (dolist (v vertices)
	 (if (member (v-type v) *dummy-types*)
	     (format *acad-file*
		     "(command \"INSERT\" \"DUMMY\" \"~S,~S\" 1 1 0) ~%"
		     (v-x-pos v)
		     (v-y-pos v))
	     (format *acad-file*
		     "(command \"INSERT\" \"NODE\" \"~S,~S\" 1 1 0 ~
                               \"~S: ~S\" ~S) ~%"
		     (v-x-pos v) (v-y-pos v)
		     (v-name v) (v-type v) (v-pattern v))))
       ;;
       ;; draw links from each vertex to its successors
       ;;
       (dolist (v vertices)
	 (draw-links-to-successors v))
       ;;
       ;; send end of file characters for the AutoCAD
       ;;
       (format *acad-file* "(princ) ~%")
       (format *acad-file* ") ~%")
     )
   )
)

(defun draw-links-to-successors (v)
  (let ((x (v-begin-x-pos v))
	(y (v-y-pos v)))
    (dolist (s (v-begin-suc v))
      (draw-acad-line-to-end x y s)))
  (let ((x (v-end-x-pos v))
	(y (v-y-pos v)))
    (dolist (s (v-end-suc v))
      (draw-acad-line-to-end x y s))))

(defun draw-acad-line-to-end (start-x start-y ne-tag)
  (let ((v (find-vertex (ne-tag-node ne-tag))))
    (draw-acad-line
       start-x
       start-y
       (ecase (ne-tag-end ne-tag)
	 (begin_of (v-begin-x-pos v))
	 (end_of (v-end-x-pos v)))
       (v-y-pos v))))

(defun v-begin-x-pos (v)
  (v-x-pos v))

(defun v-end-x-pos (v)
  (+ (v-x-pos v) (if (member (v-type v) *dummy-types*) 20 100)))

(defun draw-acad-line (x1 y1 x2 y2)
  (format *acad-file*
	  "(command \"line\" \"~S,~S\" \"~S,~S\") (command \"\") ~%"
	  x1 y1 x2 y2))

;;; ---------------------------- Change History ----------------------------
;;; (Who)   (When) 		     (What)
;;;
;;;  jd    March 93	Extensive modification and cleanup for version 2.0.
;;;			Added utilities such as the tsort-based longest-
;;;			length algorithm.
;;;
;;;  jd    June 93      New autocad node-to-graph procedures.
;;;
;;;  jd    June 93      Receive planner output as lists rather than strings.
;;;
