;;;; File: viewer-toplevel.lsp
;;; SCCS Version: %W%
;;; Contains: Plan / World viewer toplevel
;;; Authors: Jeff Dalton and Brian Drabble
;;; Created: Wed Mar  4 17:00:03 1992
;;; Updated: Mon Jul 10 20:19:27 1995 by Jeff Dalton
;;; Release Version: %Y%
;;; Copyright: (c) 1992, 1993, 1994 AIAI, University of Edinburgh
;;; This material may be reproduced by or for the U.S. Government pursuant
;;; to the copyright license under the clause at DFARS 252.227-7032
;;; (June 1975) -- Rights in Technical Data and Computer Software (Foreign).

;;; Plan/World-Viewer
;;;
;;; This module is given a plan or world description and produces output
;;; in various forms: text to screen, text to file, AutoCAD, etc.
;;;
;;; The event-handler is an artifact of the single-process version of
;;; O-Plan.  The Plan / World Viewer runs as a separate p-process and
;;; receives the Planner's output as a message event.
;;;
;;; The message contents will be either
;;;
;;;    (:WORLD node-id ((pattern value)*) viewer-arg...)
;;; or
;;;    (:PLAN (node-description*) viewer-arg...)
;;;
;;; The viewer-args are used to answer questions about how the plan or
;;; world view should be handled (e.g. whether it should be written to
;;; a file and, if so, which file).  They should be a p-list of the form
;;; ( {keyword value}... ).  One keyword should be :mode to specify the
;;; display mode.  The other keywords will depend on the mode.
;;;
;;; The viewer-args are optional, but if any appear they indicate that
;;; the viewer is under program control.  All questions will be answered
;;; from the args and an error will be signalled if a requested keyword
;;; is not present.  Moreover, the viewer will not create a window or
;;; procuce any window output unless specifically told to put something
;;; on the screen.
;;;
;;; If no viewer-args appear, then all questions will be answered by
;;; asking the user (e.g. by putting up a menu) and the viewer process
;;; will automatically create a window when it starts running.
;;;

(in-package :plan-world-viewer :nicknames '(:view))

(use-package :tf-language)

(use-package :x)
(use-package :pseudo-process)		;/\/ too low-level?
(use-package :util)
(use-package :time-util)
(use-package :monitors)			;/\/ for mon-format
(use-package :initsystem)
(use-package :mcc)

(import '(oplan::node-1 oplan::unknown))

(defvar *window* nil)			;current request's screen-I/O stream

(defvar *viewer-io-package* nil)	;package to read and write in

(defvar *plan-display-mode* nil)
(defvar *world-display-mode* nil)
(defvar *max-plan-level* nil)		;max levels of expansion to show

(defvar *nodes* nil)			;to aid debugging, etc.

;;; Event handler

(defun pw-event-handler (self)
  (let ((planner-output (message-contents (next-event self)))
	(*window* 'not-a-window-yet))
    (setq *viewer-io-package* (find-package :oplan))
    (catch 'viewer-exit
      (apply (ecase (first planner-output)
	       (:world #'view-world)
	       (:plan  #'view-plan))
	     (rest planner-output)))))

(defun exit-viewer ()
  (throw 'viewer-exit nil))

(defun exit-view ()
  (throw 'view-exit nil))		;see below

;;;
;;; World view
;;;

(defun view-world (node-id p-v-pairs &rest viewer-args)
  (common-viewer-setup viewer-args)
  (display-banner "World View at end_of node-~A" node-id)
  (unwind-protect
      (output-viewer-world p-v-pairs)
    (common-viewer-finish)
    (display-divider)))

;;; Viewer world-view

(defun output-viewer-world (p-v-pairs)
  (loop
    (setq *world-display-mode* (get-world-display-mode))
    (catch 'view-exit
      (ecase *world-display-mode*
	(:screen  (output-text-world p-v-pairs))
	(:file    (output-file-world p-v-pairs))
	(:autocad (output-acad-world p-v-pairs))
	(:quit    (return))))
    (unless (interactivep)
      (return))))

(defun get-world-display-mode ()
  (or (viewer-arg :mode) (ask-world-display-mode)))

(defun ask-world-display-mode ()
  (menu-request
    '("-heading" "World display mode"
      "-line"
      "Text to screen=:screen"
      "Text to file=:file"
      "AutoCAD=:autocad"
      "-line"
      "QUIT=:quit")))

;;;
;;; Plan View
;;;

(defun view-plan (nodes &rest viewer-args)
  (common-viewer-setup viewer-args)
  (setq *nodes* nodes)		;to aid debugging
  (fix-up-nodes nodes)		;via side effects
  (display-banner "Plan View")
  (unwind-protect
      (output-viewer-plan nodes)
    (common-viewer-finish)
    (display-divider)))

;;; Viewer plan view

(defun output-viewer-plan (nodes)
  (loop
    (setq *plan-display-mode* (get-plan-display-mode))
    (catch 'view-exit
      (ecase *plan-display-mode*
	(:screen    (output-text-plan nodes))
	(:file      (output-file-plan nodes))
	(:psgraph   (output-psgraph-plan nodes))
	(:narrative (output-narrative-plan nodes))
	(:autocad   (output-autocad-plan nodes))
	(:krsl      (output-krsl-plan nodes))
	(:quit      (return))))
    (unless (interactivep)
      (return))))

(defun get-plan-display-mode ()
  (or (viewer-arg :mode) (ask-plan-display-mode)))

(defun ask-plan-display-mode ()
  (menu-request
    '("-heading" "Plan display mode"
      "-line"
      "Text to screen=:screen"
      "Text to file=:file"
      "PostScript graph=:psgraph"
      "Narrative=:narrative"
      "AutoCAD=:autocad"
      "KRSL=:krsl"
      "-line"
      "QUIT=:quit")))

;;; ---------------------------- Change History ----------------------------
;;; (Who)   (When) 		     (What)
;;;
;;;  jd    March 93	Extensive modification and cleanup for version 2.0.
;;;
;;;  jd    June 93      Receive planner output as lists rather than strings.
;;;
;;;  jd    October 93   Split AutoCAD drawer into a spearate file.
;;;
;;;  jd    April 94	PSGraph drawer that draws nodes as one box when
;;;			the only link from the begin_of the node is to the
;;;			end_of the same node and vice versa.
;;;
;;;  jd    Sept 94	Reorganized into separate files for text, psgraph
;;;                     etc; and into -toplevel and -services.
;;;
;;;  jd    October 94	Added viewer-args so that the viewer could be
;;;			subject to programmatic control.
;;;  jd    October 94   Changed "Print graph" plan-view option to
;;;			"PostScript graph".
;;;
