;;;; File: connect-mode.lsp
;;; SCCS Version: %W%
;;; Contains: Connected-mode
;;; Author: Jeff Dalton <J.Dalton@ed.ac.uk>
;;; Created: 31 March 1993
;;; Updated: Wed Nov  9 13:50:54 1994 by Jeff Dalton
;;; Release Version: %Y%
;;; Copyright: (c) 1992, 1993, AIAI, University of Edinburgh
;;; This material may be reproduced by or for the U.S. Government pursuant
;;; to the copyright license under the clause at DFARS 252.227-7032
;;; (June 1975) -- Rights in Technical Data and Computer Software (Foreign).

(in-package :oplan)

;;;; Connect[ed] mode

;;; Initsystem provides is-set-connect, which is used to record that we're
;;; in connect-mode, and is-connect-p which is used to test whether we are.

;;; Our task here is (1) to reassign some of the standard Lisp stream
;;; variables (such as *terminal-io*) so that they no longer use the
;;; stdin or stdout of the Lisp process, (2) to open a window that can
;;; be used for those standard stream variables and for interaction
;;; with the repl process, (3) to make *left-io* refer to the original
;;; stdin and stdout, and (4) to create a connector process that
;;; connects *left-io* to :leftin and :leftout.

;;; Tasks (1)-(3) are performed by set-up-connect-mode, which is called
;;; by oplan-top-level when it processes the "-connect" command-line
;;; argument.  Set-up-connect-mode manipulates *terminal-io* itself.
;;; It calls open-repl-window to open the window, and open-repl-window
;;; calls x-redirect-errors to change *debug-io*, *trace-output*, and
;;; *error-output*.  It's a bug if these changes are not completed before
;;; any output is directed to any of the standard streams involved,
;;; so these changes should happen as soon as possible.

;;; /\/: What about *standard-input*, *standard-output*, and *query-io*?
;;; In many Lisps they are synonym-streams for *terminal-io*, but what
;;; if they aren't?

;;; Task (4) is performed by connect-to-left-io.  This procedure should
;;; not be called until after there is a server for :leftin.  Otherwise
;;; it may try to send something to :leftin before there is anything there
;;; to receive it.  (This could happen when we're run as the Exec, because
;;; there will be some input ready in *left-io* almost as soon as Lisp
;;; starts running.)

;;; Connect-to-left-io also writes the "agent ready message" to *left-io*.
;;; It's important that this not happen until after we've successfully
;;; put up a window.  If we're creating windows at all, and we're not
;;; in subr mode, the repl window should already be up before connect-
;;; to-left-io is called.  It might be better to have the IM send the
;;; message, or to wait until all windows are up (remember that that
;;; happens in the 1st cycle of pprocess-main-look, not before), but
;;; it's also desirable to have the connect mode stuff (which includes
;;; the ready message) all in one place.  So all things considered,
;;; connect-to-left-io is a reasonable choice.

(defvar *left-io* nil)			;comms with the agent to the left

(defvar *repl-window* nil)		;just to remember it
(defvar *initial-terminal-io* nil)	;just to remember it

(defun set-up-connect-mode ()
  (is-set-connect t)
  (protect-terminal-io))

(defun protect-terminal-io ()
  (setq *left-io* *terminal-io*)	;/\/ portable?
  (setq *initial-terminal-io* *terminal-io*)
  (setq *repl-window* (open-repl-window))
  (setq *terminal-io* *repl-window*)
  )

(defun open-repl-window ()
  (x-open-and-register-io-win :repl-io
    (or (is-get-window-args :repl-io)
	(list "-title" (if (is-exec-p) "O-Plan Exec Lisp interaction"
			               "O-Plan Lisp interaction")
	      "-n" "Lisp Listener")))
  (x-redirect-errors :repl-io)
  (x-get-stream :repl-io))


(defun connect-to-left-io ()
  (new-connector :left-io
    :io-stream *left-io*
    :socket-to-read :leftout
    :socket-to-write :leftin
    :input-guard
      (if (is-exec-p)
	  'exec-input-guard
	  'planner-input-guard))
  (ipc-write-agent-ready *left-io*)
  )

;;; Input guards

;;; Note that guards get to transform the input as well as check it.

;;; /\/: At present, there's virtually no checking.

(defun planner-input-guard (self input)
  (declare (ignore self))
  (assert (consp input))
  (values t (make-leftin-event input)))

(defun exec-input-guard (self input)
  (declare (ignore self))
  (assert (consp input))
  (values t (make-leftin-event input)))

;; /\/: Agents should accept input in a less extreme format: i.e. not
;; require event structs.  Here we have the :leftin input guard create
;; the event struct so the user doesn't have to.
;;
(defun make-leftin-event (input)
  (list :EVENT (make-event :BODY input)))

;;; End
