;;;; File: manual-interface.lsp
;;; SCCS Version: %W%
;;; Contains: Manual interface to single-process O-Plan
;;; Author: Jeff Dalton <J.Dalton@ed.ac.uk>
;;; Created: October 1994
;;; Updated: Thu Jun  1 16:58:24 1995 by Jeff Dalton
;;; Release Version: %Y%
;;; Copyright: (c) 1994, AIAI, University of Edinburgh
;;; This material may be reproduced by or for the U.S. Government pursuant
;;; to the copyright license under the clause at DFARS 252.227-7032
;;; (June 1975) -- Rights in Technical Data and Computer Software (Foreign).

(in-package :oplan)

;;;; Manual interface

;;; Whether or not we're in connect-mode, we may want to create some
;;; :LEFTIN events "by hand".  This section contains some routines that
;;; are useful in common cases of that sort.

;;; Note that the current p-process will be registered as a client
;;; of :LEFTIN the first time send-to-leftin-if-valid is called.

;;; /\/: The input guards are for the connect-mode :left-io connector
;;; p-process but here we use them directly.  The IM ought to have guards
;;; of its own.

(defun send-to-leftin-if-valid (message)
  (ensure-client :leftin)
  (multiple-value-bind (ok-p transformed-message)
      (check-for-leftin message)
    (if ok-p
	(ipc-write :leftin transformed-message)
      (error "Invalid message ~S." message))))

(defun check-for-leftin (message) ; -> ok-p, transformed-message
  (funcall (leftin-input-guard)
	   *pprocess*
	   message))

(defun leftin-input-guard ()
  (if (is-exec-p) #'exec-input-guard #'planner-input-guard))


;;; Some messages

(defun add-action (pattern)
  (send-to-leftin-if-valid
    `(:add-to-task :action ,pattern)))


;;;; Exec

;;; Exec from file can be used to tell the Exec to start executing a
;;; plan.  It makes sure that there's a :LEFTOUT server and hence that
;;; messages to :LEFTOUT will be handled.  If there's not already a
;;; server, a server that simply prints the messages to *terminal-io*
;;; will be created.

(defun exec-from-file (filename)
  (assert (is-exec-p))
  (ensure-leftout-connection)
  (let ((file-contents (file->list filename)))
    (unless (and (consp file-contents) (null (cdr file-contents)))
      (error "More than one expression in Exec input file ~S." filename))
    (let ((message (car file-contents)))
      (send-to-leftin-if-valid message)))
  t)

(defun ensure-leftout-connection (&optional out-stream)
  (unless (socket-server (find-socket :leftout))
    (new-connector :leftout-server
      :io-stream (or out-stream *terminal-io*)
      :socket-to-read :leftout)))


;;;; Force-reset

;;; Attempt to stop the planner from doing whatever it's currently
;;; doing and then to reinitialize it.  Might be called from an error
;;; "break".

;;; If KP-1 is active, it will send a :KPREADY when it's unwound
;;; because of an unwind-protect.

;;; /\/: Assumes only one KP.

(defun force-reset ()
  (declare (special *ready-kps*))
  (dolist (p *all-pprocesses*)
    (setf (queue-contents (pprocess-event-queue p)) nil))
  (unless (or *ready-kps*
	      (ipc-pprocess-is-active-p (find-pprocess :kp-1)))
    (push '(:KPREADY . :KP-1) *ready-kps*))
  (send-to-leftin-if-valid '(:init))
  (throw :pprocess-cycle-exit nil))


;;;; Testing

;;; Sanity checks
;;;
;;; The number of errors will be sent to the TA.
;;; Error messages will appear in the Lisp interaction window.

(defun sanity-check ()
  (send-to-leftin-if-valid '(:check-plan)))


;;; A quick way to run the auto-tester.

;;; Also, "(go-faster)" can be used to make things faster.

(defun standard-test (&rest test-keys)
  (run-test-from-scratch "atest:run-standard-test-sequence" test-keys))

(defun full-test (&rest test-keys)
  (run-test-from-scratch "atest:run-full-test-sequence" test-keys))

(defun release-test (&rest test-keys)
  (run-test-from-scratch "atest:run-release-test-sequence" test-keys))

(defun run-test-from-scratch (fn-name-string test-keys)
  (go-faster)
  (set-up-test)
  (apply (read-from-string fn-name-string) test-keys)
  nil)

(defun set-up-test ()
  (load-system 'auto-tester :recursive nil))

(defun go-faster ()
  #+kcl (use-fast-links t)
  (run-lights-off)
  (output-off))

(defun output-off ()
  ;; /\/: This works only if all conponents are in the same Lisp image.
  (let ((none (dev::dev-get-level-num-from-level-name :none)))
    (dolist (p *all-pprocesses*)
      (when (components::component-p p)
        (setf (pprocess-symbol-value p 'dev::*dev-debug-level*)
	      none)))))

;;; ---------------------------- Change History ----------------------------
;;; (Who)   (When) 		     (What)
;;;
