;;;; File: program-interface.lsp
;;; SCCS Version: %W%
;;; Contains: Programmatic interface to single-process O-Plan
;;; Author: Jeff Dalton <J.Dalton@ed.ac.uk>
;;; Created: October 1994
;;; Updated: Tue Apr  4 05:10:03 1995 by Jeff Dalton
;;; Release Version: %Y%
;;; Copyright: (c) 1994, AIAI, University of Edinburgh
;;; This material may be reproduced by or for the U.S. Government pursuant
;;; to the copyright license under the clause at DFARS 252.227-7032
;;; (June 1975) -- Rights in Technical Data and Computer Software (Foreign).

(in-package :oplan)


;;;; The programmatic interface is used in subr (subroutine) mode.

(defun pi-check-state (checking-for)
  (unless (and (pprocess-p *pprocess*)
	       (eq (pprocess-name *pprocess*) :user))
    (error "~A called when *pprocess* was not :user" checking-for)))


;;;; High-level interface

;;; Planning and replanning

(defun plan-for (domain-name task-name)
  (send-to-oplan :init)
  (receive-else-error '(:init :ok))
  (send-to-oplan :domain domain-name)
  (receive-else-error '(:domain ($tf-file-name . $task-names)))
  (send-to-oplan :set-task task-name)
  (receive-plan-status))

(defun replan ()
  (send-to-oplan :poison-state :replanning)
  (receive-plan-status))

(defun receive-plan-status ()
  (receive-else-error '($status)
    #'(lambda (status)
	(ecase status
	  (:finished t)
	  (:no-more-alternatives nil)))))


;;; Plan statistics

(defun request-plan-statistics-list ()
  (send-to-oplan :check-plan :get-plan-statistics)
  (receive-else-error '(:checked :statistics $stats)
    #'(lambda (stats)
	stats)))


;;; Plan views

(defun request-plan-view-list ()
  (error "request-plan-view-list not implemented yet"))

(defun request-plan-view (&rest viewer-args)
  (apply #'send-to-oplan :user :plan-view viewer-args)
  (receive-else-error '(:finished-view)))

(defun request-psgraph (format &rest more-args
			       &key (levels :all)
			       &allow-other-keys)
  (apply #'request-plan-view
      :mode :psgraph
      :format format
      :levels levels
      more-args))


;;; World views

(defun request-world-view-list (node-id)
  ;; Node-id is something like 3-2 for end_of node-3-2.
  (check-type node-id string)
  (send-to-oplan :user :program-request :world-view node-id)
  (receive-else-error '(:world-view $p-v-pairs)
    #'(lambda (p-v-pairs)
	p-v-pairs)))

(defun request-world-view (node-id &rest viewer-args)
  (check-type node-id string)
  (apply #'send-to-oplan :user :world-view node-id viewer-args)
  (receive-else-error '(:finished-view)))


;;; Time bounds (on a node-end)

(defun request-time-bounds (end-tag) ; -> (min max)
  ;; end-tag is e.g. (node-2 :begin)
  (db-request :funcall 'get-ne-time-bounds end-tag))

(defun get-ne-time-bounds (end-tag)
  (let* ((ne (get-node-end end-tag))
	 (tp (ne-time-point ne)))
    (list (tpoint-min tp) (tpoint-max tp))))


;;; Component output

;;; The component had better not be running when these things happen.

(defun redirect-component-output (component file-name)
  (let ((c (if (symbolp component) (find-pprocess component) component)))
    (check-type c component)
    (let ((stream (open file-name :direction :output)))
      ;; We seem to have a component and a stream, so ...
      (set-component-output-stream c stream)
      ;; Try to make sure the stream is closed when we exit
      (add-exit-action
        #'(lambda () (close stream)))
      stream)))

(defun set-component-output-stream (component stream)
  (let ((c (if (symbolp component) (find-pprocess component) component)))
    (check-type c component)
    (setf (pprocess-symbol-value c '*debug-io*)
	  stream)))

(defun set-component-debug-level (component level)
  (let ((c (if (symbolp component) (find-pprocess component) component)))
    (check-type c component)
    (call-in-pprocess-env c
      #'(lambda (c)
	  (declare (ignore c))
	  (dev-set-debug-level level)))))


;;;; Lower level communications

;;; Ask-oplan and send-to-oplan

(defun ask-oplan (id &rest args)
  (pi-check-state 'ask-oplan)
  (send-to-leftin-if-valid (cons id args))
  (pprocess-main-loop)
  (if (next-event-p)
      (next-event)
    nil))

(defun send-to-oplan (id &rest args)
  (pi-check-state 'send-to-oplan)
  (send-to-leftin-if-valid (cons id args))
  (pprocess-main-loop))


;;; Receive-from-oplan

(defun receive-from-oplan (pattern &key (succeed 'default-receive-succeed)
			                (fail 'default-receive-fail))
  (pi-check-state 'receive-from-oplan)
  (let* ((message (peek-next-event))
	 (contents (if message (message-contents message) :nothing))
	 (matching (match pattern contents)))
    (if matching
	(progn (next-event)
	       (apply succeed (matching-values matching)))
      (funcall fail contents))))

(defun matching-values (matching)
  ;; /\/: Assumes that the matcher constructs the result in the
  ;; right order (or, rather, the reverse of it).
  (and (consp matching)
       (nreverse (mapcar #'cdr matching))))

(defun default-receive-succeed (&rest values)
  (declare (ignore values))
  t)

(defun default-receive-fail (contents)
  (declare (ignore contents))
  nil)


;;; Receive-else-error

(defun receive-else-error (pattern &optional (succeed
					      'default-receive-succeed))
  (receive-from-oplan pattern
    :succeed succeed
    :fail    (partial1 #'receive-error pattern)))

(defun receive-error (pattern message-contents)
  (error "Recieved ~S from O-Plan when expecting ~S."
	 message-contents pattern))


;;;; Timeouts

;;; The user pprocess will signal a timeout condition when its
;;; wakeup-time arrives.  An error is signalled if the timout isn't
;;; handled.  See subr-mode.lsp for the details.

(defun set-timeout (seconds)
  (pi-check-state 'set-timeout)
  (set-user-pprocess-timeout seconds))

(defun cancel-timeout ()
  (pi-check-state 'cancel-timeout)
  (set-user-pprocess-timeout nil))

(defmacro with-timeout (seconds &body forms)
  `(call-with-timeout ,seconds #'(lambda () ,@forms)))

(defun call-with-timeout (seconds thunk)
  (unwind-protect
       (progn (set-timeout seconds)
	      (funcall thunk))
     (cancel-timeout)))

;;; ---------------------------- Change History ----------------------------
;;; (Who)   (When) 		     (What)
;;;
