;;;; File: top-level.lsp
;;; SCCS Version: %W%
;;; Contains: O-Plan top-level 
;;; Author: Jeff Dalton <J.Dalton@ed.ac.uk>
;;; Created: March 1993
;;; Updated: Sun Nov  6 20:02:22 1994 by Jeff Dalton
;;; Release Version: %Y%
;;; Copyright: (c) 1992, 1993, AIAI, University of Edinburgh
;;; This material may be reproduced by or for the U.S. Government pursuant
;;; to the copyright license under the clause at DFARS 252.227-7032
;;; (June 1975) -- Rights in Technical Data and Computer Software (Foreign).

;;; The wonderful top-level of O-Plan!

;;; /\/: Merge with main.lsp and initsystem.lsp now that it's working.
;;; The implementation-dependent parts belong in dependent-support.

;;; To build an executable image:
;;;   0. Run a Lisp.
;;;   1. (load "oplan.lsp").
;;;   2. Maybe:
;;;       2.1. (compile-system 'oplan), or
;;;            (compile-system 'everything)
;;;       2.2. Exit Lisp, rerun, and reload "oplan.lsp".
;;;   3. (load-system 'oplan), and maybe
;;;      (preload-knowledge-sources).
;;;   4. (save-oplan <filename>).

;;; Note that we call string-upcase before interning, thus implicitly
;;; assuming that the internal case is upper.  Franz Inc users beware.

(in-package :oplan)


(export 'save-oplan)

;;; Save-oplan creates an executable image

(defun save-oplan (filename)
  (util:save-image filename #'oplan-top-level))


;;; Semi-portable top-level

(defun oplan-top-level ()
  (handler-bind ((condition #'last-resort-condition-handler))
    (top-level-init)
    (unwind-protect
        (if (get-parameter :subr)
	    (enter-subr-mode)
	  (time (run)))
      (terminate-xterms-if-necessary)
      (exit-lisp 0))))

(defun top-level-init ()
  (setq *package* (find-package :oplan))
  (set-parameter :process-model :single-process)
  ;; Set up other parameter defaults
  (is-init-parameter-defaults)
  ;; Process some values from environment variables as more defaults.
  (is-init-env-parameters)
  ;; Process argv
  (set-parameter :image-name (argv 0))
  (let ((init-file-p t)
	(i 1))
    (labels ((arg ()
	       (argv i))
	     (pop-arg ()
	       (prog1 (arg) (incf i)))
	     (arg->keyword (string)
	       (string->keyword (string-upcase (subseq string 1)))))
      (loop
        (when (null (arg))
	  (return))
	(case (intern (string-upcase (arg)))
	  (-tfc     (pop-arg) (tf-compiler:run-as-tfc i))
	  (-load    (pop-arg) (load-most-recent (pop-arg)))
	  (-break   (pop-arg) (break "~A" '-break))
	  (-eval    (pop-arg) (eval (read-from-string (pop-arg))))
	  (-noinit  (pop-arg) (setq init-file-p nil))
	  (-config  (pop-arg) (is-set-config-file (pop-arg)))
	  (-connect (pop-arg) (set-up-connect-mode))
	  (-asexec  (pop-arg) (is-set-exec-p t))
	  ((-no -not)
	   ;; Used to set parameters to nil (false)
	   ;; /\/: Perhaps shouldn't require - or : on the parameter?
	   (let ((negator (pop-arg))
		 (parameter (arg)))
	     (if (and parameter (find (char parameter 0) "-:"))
		 (set-parameter (arg->keyword (pop-arg)) nil)
	       (warn "~S was not followed by a parameter to negate"
		     negator))))
	  (t
	   ;; Other -name or :name args are given to set-parameter
	   ;; as keywords.  The parameter value will be a string.
	   (cond ((string= (arg) "")
		  (warn "Skipping null argument."))
		 ((find (char (arg) 0) "-:")
		  (let* ((arg (pop-arg))
			 (value (pop-arg)))
		    (if value
			(set-parameter (arg->keyword arg) value)
		      (warn "No value for ~A" arg))))
		 (t
		  (warn "Skipping unknown argument: ~A" (pop-arg))))))))
    ;; Load init file if wanted.
    (when init-file-p (load-init-file "oplan-init" :oplan))))

(defun exit-oplan ()
  (throw :pprocess-main-loop-exit nil))


;;;; The "last resort" condition handler

(defvar *lrch-attempt* 0)

(defun last-resort-condition-handler (c)
  (unless (or (get-parameter :interactive)
	      (and (typep c 'warning)
		   (not *break-on-warnings*)))
    (incf *lrch-attempt*)
    (case *lrch-attempt*
      (1
       ;; Print a message and exit.  If a condition is signalled before
       ;; exit-lisp causes an unwind (e.g. if it happens while printing
       ;; the error message), then this handler will run again, but with
       ;; an attempt count of 2.
       (format *error-output*
	       "~&~%The following condition was not handled:~%~% ~A~%~%"
	       c)
       (exit-lisp -1))
      (2
       ;; This time we won't try printing a message.
       (exit-lisp -1))
      (t
       ;; It looks like the ordinary Lisp mechanisms are not going
       ;; to get us out.  We should try to call the C exit, but
       ;; right now we don't know how.  /\/
       (exit-lisp -1)))))

;;; End
