;;;; File: web-support.lsp
;;; SCCS Version: %W%
;;; Contains: Support for Web / CGI demos
;;; Author: Jeff Dalton <J.Dalton@ed.ac.uk>
;;; Created: October 1994
;;; Updated: Fri Jun 23 17:36:05 1995 by Jeff Dalton
;;; Release Version: %Y%
;;; Copyright: (c) 1994, AIAI, University of Edinburgh
;;; This material may be reproduced by or for the U.S. Government pursuant
;;; to the copyright license under the clause at DFARS 252.227-7032
;;; (June 1975) -- Rights in Technical Data and Computer Software (Foreign).

(in-package :oplan)

;;; Contents:
;;;  * General framework (description)
;;;  * Web environment / setup
;;;  * Tmp directory
;;;     - Routines for creating filenames and URLs
;;;     - Special URLs
;;;     - Notes file
;;;  * Query arg (QUERY_STRING) parsing
;;;  * Query arg checking and conversion
;;;  * HTML output support
;;;     - Output routines
;;;     - Encoding for PRE blocks
;;;     - Links to standard results
;;;  * TF output support
;;;  * Output macro
;;;  * Demo utilities
;;;  * String and other (low-level) utilities


;;;; General framework
;;;
;;; This framework is chiefly for non-interactive demos in which O-Plan
;;; is run to produce HTML output.
;;;
;;; For a Web demo D, the O-Plan "source" directory will usually contain
;;; the following files:
;;;   web/D.html           Introduction to the demo plus a FORM
;;;   web/D.cgi            Shell script that runs O-Plan with appropriate args
;;;   web/D-support.lsp    Lisp routines that control the demo
;;;
;;; These files may actually be in one or more subdirectories rather
;;; than directly in "web"; it depends on how much can be the same for
;;; the different hosts on which the demo runs.  For the details, see
;;; web/README and the routines below that construct URLs and file names.
;;;
;;; The FORM in D.html will have D.cgi as its ACTION.  How this is
;;; made to work depends on the httpd configuration.  (Perhaps there's
;;; a symbolic link from the CGI bin area to the script, or it may be
;;; necessary to copy the script to the CGI bin area or to take even
;;; more elaborate steps.)
;;;
;;; The D.cgi script will run O-Plan in "subr" mode without windows
;;; and arrange for D-support.lsp to be loaded.  It will also cause a
;;; procedure named D, usually defined in D-support.lsp, to be called.
;;; That procedure is responsible for getting O-Plan to do whatever
;;; the demo requires.  The definition will look something like this:
;;;
;;;    (defparameter *D-parameters* <list of query arg descriptions>)
;;;
;;;    (defun D ()
;;;      (with-web-environment "D"
;;;        (parse-query-args)
;;;        (convert-query-args *D-parameters*)
;;;        ...))
;;;
;;; D should print HTML to the stdout of the Lisp process.  The "html-"
;;; procedures defined below are a convenient way to do this.  The HTML
;;; can contain hypertext links to any files produced by the demo
;;; (such as a PostScript graphs).  Such files are normally placed in
;;; a tmp directory as described in the next section.
;;;
;;; A typical D.cgi script might be:
;;;
;;;    #!/bin/bash
;;;
;;;    echo Content-type: text/html
;;;    echo
;;;
;;;    cd the-root-oplan-directory/source
;;;
;;;    exec ./web/oplan-cgi -load web/D-support -do "(D)"
;;;
;;; Here the script produces the initial lines of output that precede
;;; HTML text.  Doing this in the script makes it more likely that
;;; some sort of readable output will appear even when things go wrong
;;; on the Lisp side.
;;;
;;; For some demos, the script will do even more of the initial HTML
;;; output.  For instance, it may print the TITLE.
;;;
;;; oplan-cgi is a script that runs O-Plan, inserting the command-line
;;; arguments that indicate subr mode, no windows, etc.  It can also
;;; set a time limit.
;;;
;;; Note that it may be necessary to run oplan-cgi using a name such as
;;; "./web/oplan-cgi" that contains a "/", because the current directory
;;; may not be in the httpd user's PATH.
;;;


;;;; Web environment / setup

(defvar *web-demo-name* nil)		;a string that identifies the demo
(defvar *web-notes* nil)		;an open stream or else nil

(defmacro with-web-environment (demo-name &body body)
  `(let ((*web-demo-name* ,demo-name))
     (call-with-web-environment
       #'(lambda () ,@body))))

(defun call-with-web-environment (thunk)
  (check-type *web-demo-name* string)
  (setq *web-notes* nil)
  (catch :web-demo-exit
    (handler-case
        (progn
	  (establish-web-setup)
	  (go-faster)			;may decrease our ability to debug
	  (funcall thunk))
      (condition (c)
        (web-demo-report-condition c)))))

(defun exit-web-demo ()
  (throw :web-demo-exit nil))

(defun establish-web-setup ()
  ;; /\/: More checks to be filled-in.
  ;; Checks
  (assert (stringp *web-demo-name*))
  ;; Setup
  (set-parameter :oplan-tf-dir
    (concat-string (get-parameter :oplan-dir) "/source/web/demo"))
  (ensure-web-demo-dir)
  (generate-web-demo-id)
  (ensure-web-special-url)
  )

(defvar *in-html* t)

(defun start-html (&key (no-line nil))
  ;; /\/: At present, a horizontal line follows successful startup.
  (html-line "</LISTING>")
  (unless no-line
    (html-line "<HR>"))
  (setq *in-html* t))

(defun stop-html ()
  (html-line "<LISTING>")
  (setq *in-html* nil))

(defun web-demo-report-condition (c)
  ;; Tell the user
  (if *in-html*
      (html-report-condition c)
    ;; Not HTML yet...
    (progn
      (format t "~&~%Processing stopped because:~%~%~A~%~%" c)
      ;; We should now be on our way out of Lisp.  Start-html will
      ;; end the LISTING block, which doesn't seem to be necessary
      ;; at least for Mosaic.  /\/
      (start-html)))
  ;; Make a note of it too.
  (web-note-failure c))


;;;; Tmp directory for files created by Web demos
;;;
;;; After ensure-web-demo-dir and generate-web-demo-id have been called
;;; (usually by establish-web-setup, above), the following functions
;;; can be called to produce names of temporary files and URLs that
;;; refer to the same files:
;;;
;;;   (web-tmp-filename <name> <type>)
;;;   (web-tmp-url <name> <type>)
;;;
;;; The name describes the file contents; the type is a file type
;;; (extension) such as "ps" for PostScript or "html" for HTML.
;;; The type "txt" is often used for plain (non-HTML) text.
;;;
;;; File names and URLs relative to the main Web demo directory
;;; (which contains the tmp directory as a subdirectory) can be
;;; created by
;;;
;;;   (web-demo-filename <filename>)
;;;   (web-demo-url <filename>)
;;;
;;; However, for these functions to work some setup may be required in
;;; addition to the parameters manipulated here.  This may involve
;;; symbolic links and perhaps even changes to the httpd configuration.
;;; The setup used on AIAI machines is described in Notes/web-setup.
;;;
;;; *web-demo-dir-name* is name of the Web demo directory.  It is normally
;;; "source/web/demo" in the main O-Plan directory.  This directory must
;;; contain a subdirectory called "tmp".
;;;
;;; *web-demo-dir-url* is a URL that refers to the Web demo directory.
;;; Symbolic links might be needed to set this up.
;;;
;;; *web-demo-id* is used to create unique filenames.  The id is
;;; determined by generate-web-demo-id.
;;;
;;; When O-Plan is installed, the "tmp" directory is initially
;;; a symbolic link to "web-tmp" in the main O-Plan directory.
;;; "web-tmp" must be created by by hand.  It should be writable
;;; by everyone (or at least by whatever user is used by the http
;;; daemon, typically "nobody").  Note that the http configuration
;;; may restrict the use of symbolic links.
;;;

(defparameter *web-demo-dir-name* nil)
(defparameter *web-demo-dir-url* nil)

(defparameter *web-demo-id* nil)

(defun ensure-web-demo-dir ()
  (unless *web-demo-dir-name*
    (setq *web-demo-dir-name* (default-web-demo-dir-name)))
  (unless *web-demo-dir-url*
    (setq *web-demo-dir-url* (default-web-demo-dir-url))))

(defun default-web-demo-dir-name ()
  (concat-string (get-parameter :oplan-dir) "/source/web/demo"))

(defun default-web-demo-dir-url ()
  (let ((server-name (getenv-else-error "SERVER_NAME")))
    (cond ((string= server-name "www.aiai.ed.ac.uk")
	   (concat-string
	     "http://" server-name
	     #+:undef "/~jeff/oplan-web-ptr/demo"
	     "/~jeff/oplan-demo-ptr/demo")) ; ref to source/web/aiai-demo/demo
	  ((string= server-name "spottisvax")
	   "http://localhost/home/jeff/oplan/2.x/source/web/demo")
	  (t
	   (error "Can't determine the Web demo directory URL.")))))

(defun generate-web-demo-id ()
  (let ((random-state (make-random-state t)))
    (loop
      (let* ((id (int->string (random 10000 random-state)))
	     (guard-name
	      (concat-string *web-demo-dir-name* "/tmp/"
			     *web-demo-name* "-" id "-info"))
	     (guard-stream
	      (open guard-name
		    :direction :output
		    :if-exists nil)))
	(when guard-stream
	  ;; We've found an id all our own.
	  (setq *web-demo-id* id)
	  ;; And we can record some things in the guard file.
	  (with-open-stream (s guard-stream)
	    (record-web-demo-info s))
	  ;; Re-open the guard file and use it for notes
	  (use-web-demo-notes-file guard-name)
	  ;; Done.
	  (return id))))))

(defun record-web-demo-info (stream)
  (flet ((line (format-string &rest format-args)
	   (format stream "~&~?~%" format-string format-args)))
    (line "~:(~S~) demo" *web-demo-name*)
    (line "")
    ;; O-Plan version
    (print-oplan-greeting stream)
    (line "")
    ;; Environment
    (line "REMOTE_ADDR=~A" (getenv "REMOTE_ADDR"))
    (line "REMOTE_HOST=~A" (getenv "REMOTE_HOST"))
    (line "")
    (line "REQUEST_METHOD=~A" (getenv "REQUEST_METHOD"))
    (line "PATH_INFO=~A" (getenv "PATH_INFO"))
    (line "")
    ;; Query-args
    (describe-query-args-before-parsed stream)
    (line "")
    ;; Parameters
    (line "Parameter values:")
    (loop for (p . v) in (parameter-alist) do
      (line "  ~S = ~S" p v))
    (line "")))

(defun web-tmp-filename (name type)
  ;; N.B. Don't call until after establish-web-setup has been called.
  (concat-string *web-demo-dir-name* "/tmp/"
		 *web-demo-name*
		 "-" *web-demo-id*
		 "-" name
		 "." type))

(defun web-tmp-url (name type)
  ;; N.B. Don't call until after establish-web-setup has been called.
  (concat-string *web-demo-dir-url* "/tmp/"
		 *web-demo-name*
		 "-" *web-demo-id*
		 "-" name
		 "." type))

(defun web-demo-filename (filename)
  ;; N.B. Don't call until after establish-web-setup has been called.
  (concat-string *web-demo-dir-name*
		 "/" filename))

(defun web-demo-url (filename)
  ;; N.B. Don't call until after establish-web-setup has been called.
  (concat-string *web-demo-dir-url*
		 "/" filename))

(defun web-demo-trace-filename ()
  ;; After generate-web-demo-id has determined *web-demo-id*,
  ;; this function will return the name of the guard / info file
  ;; in case we want to send more output to it.
  (concat-string *web-demo-dir-name* "/tmp/"
		 *web-demo-name*
		 "-" *web-demo-id*
		 "-info"))


;;; Special URLs

;;; These refer to a site-specific directory, which is unfortunately
;;; often necessary for anything that involves CGI, since one is typically
;;; at the mercy of whoever controls the httpd configuration.  /\/

(defparameter *web-special-url* nil)

(defun ensure-web-special-url ()
  (unless *web-special-url*
    (setq *web-special-url* (default-web-special-url))))
  
(defun default-web-special-url ()
  (let ((server-name (getenv-else-error "SERVER_NAME")))
    (cond ((string= server-name "www.aiai.ed.ac.uk")
	   (concat-string
	     "http://" server-name "/~jeff/oplan-demo-ptr"))
	  ((string= server-name "spottisvax")
	   "http://localhost/home/jeff/oplan/2.x/source/web/386-demo")
	  (t
	   (error "Can't determine the Web demo special URL.")))))

(defun web-special-url (filename)
  (concat-string *web-special-url*
		 "/" filename))

(defun web-mail-comment-link (&optional (description "Mail a comment"))
  (html-anchor (web-special-url "comment-form.html")
	       description))


;;; Notes file

;;; This file is established by the standard setup code (see above).

(defun use-web-demo-notes-file (filename)
  (setq *web-notes*
	(open filename :direction :output
	               :if-exists :append))
  (add-exit-action
    #'(lambda ()
	(close *web-notes*))))

(defun web-note-p ()
  (and (streamp *web-notes*)
       #+:cltl2
       (open-stream-p *web-notes*)))

(defun web-note (format-string &rest format-args)
  (apply #'format *web-notes* format-string format-args))

(defun web-note-success ()
  (web-note "~&RESULT: ~A wins!~%~%" (getenv "REMOTE_HOST")))

(defun web-note-failure (reason)
  ;; N.B. Some "failures" are expected: e.g. when there's supposed
  ;; to be no plan and there isn't one.
  (when (web-note-p)
    (web-note "~&RESULT: ~A loses because ~A.~%~%"
	      (getenv "REMOTE_HOST")
	      reason)))


;;;; Query arg (QUERY_STRING) parsing

;;; The args are a string: name1=value1&name2=value2&...
;;; The names are converted to keyword symbols by PARSE-QUERY-ARGS
;;; and stored, with their corresponding values, in a hash table.
;;; The values can then be converted from strings to symbols, ints,
;;; etc, by CONVERT-QUERY-ARGS.

;;; After parsing, (QUERY-ARG keyword) returns the corresponding value.

;;; Note that the names and values are encoded using the conventions
;;; for URLs.  We do not do any decoding automatically, since most
;;; args from our forms will avoid most special characters.  Space
;;; turns into "+", and we leave it as "+".  However, the procedure
;;; decode-query-string can be called if necessary.  It replaces + by
;;; space and %xx (% followed by two hex digits) by the corresponding
;;; character.

;;; Parsing ignores the type and value information that is used in
;;; checking and conversion.  A perhaps non-obvious consequence of
;;; this is that an arg that's supposed to be a list of values may
;;; not end up as a list, because a list results only when the 2nd
;;; value comes along (see add-query-value).

(defvar *query-arg-table* (make-hash-table :test #'eq))

(defun query-arg (kwd)
  (gethash kwd *query-arg-table*))

(defun set-query-arg (kwd value)
  (setf (gethash kwd *query-arg-table*) value))

(defsetf query-arg set-query-arg)

(defun parse-query-args (&optional (args (query-arg-string)))
  (clrhash *query-arg-table*)
  (dolist (arg (break-string-at #\& args))
    (let* ((pair (break-string-at #\= arg))
	   (name (first pair))
	   (value (second pair)))
      (add-query-value (string->keyword (string-upcase name))
		       value))))

(defun add-query-value (kwd value)
  (multiple-value-bind (val val-p) (gethash kwd *query-arg-table*)
    ;; If there's already one value, we make a list of values.
    ;; If there's alreay a list, we add to the end.
    (setf (gethash kwd *query-arg-table*)
	  (if val-p
	      (if (consp val) (nconc val (list value)) (list val value))
	    value))))

(defun describe-query-args-before-parsed (stream)
  (flet ((line (format-string &rest format-args)
	   (format stream "~&~?~%" format-string format-args)))
    (line "Query-argument values:")
    (loop for assign
	  in (break-string-at #\& (query-arg-string))
	  do (line "  ~A" assign))))

;;; Getting the query args

;;; Call QUERY-ARG-STRING which is a memoized GET-QUERY-ARG-STRING.
;;; It handles both GET and POST METHODs.

(defvar *query-arg-string* nil)

(defun query-arg-string ()
  (or *query-arg-string*
      (setq *query-arg-string* (get-query-arg-string))))

(defun get-query-arg-string ()
  (let ((request-method (getenv "REQUEST_METHOD")))
    (cond ((string= "GET" request-method)
	   (getenv "QUERY_STRING"))
	  ((string= "POST" request-method)
	   (get-POST-query-string))
	  (t
	   (error "No query args.")))))

(defun get-POST-query-string (&optional (stream *standard-input*))
  (let* ((len (string->int (getenv "CONTENT_LENGTH")))
	 (s (make-string len)))
    (dotimes (i len)
      (let ((char (read-char-no-hang stream))) ;trust no one
	(if char
	    (setf (schar s i) char)
	  (error "Not enough POST input."))))
    s))


;;;; Query arg checking and conversion

;;; Checking and conversion are controlled by argument descriptions
;;; that match the Q-ARG struct defined below.

;;; Type descriptions are:
;;;   (:INT low [high])    An int i such that low <= i [ <= high ].
;;;   (:INT* low [high])   Like :INT but allows "inf" as the upper bound.
;;;   (:NAME n1 n2 ...)    A symbol that is one of n1, n2, ...
;;;   (:NAMES n1 n2 ...)   A list of symbols, each one of n1, n2, ...
;;;   (:STRING)            An arbitrary string.
;;;   (:TEXT)              A "textarea" string.  + and %xx are converted.
;;;   (:OPTIONAL type)     Allows the arg to be NIL if not supplied.
;;;   (:CHECKBOX)          Equivalent to (:optional (:name On)).

;;; New types can be defined by define-query-arg-type.

;;; Note that conversion and checking continue with the next argument
;;; after an error.  We try to report all invalid parameters to the user.
;;; This is usually unnecessary, since in many cases (especially SELECT)
;;; a correct HTML FORM shouldn't be able to produce invalid values.
;;; In other cases, users will tend to specify reasonable values unless
;;; trying to break things.

;;; The collected errors are signalled as a new error.  /\/: We should
;;; probably define a new condition type for this case, so that it could
;;; be handled specially at higher levels; but for now, we don't.

(defstruct (q-arg (:type list))
  name					;a keyword
  type					;a type description
  description)				;a string

(defun convert-query-args (arg-descriptions)
  (let ((failures '()))
    (dolist (argd arg-descriptions)
      (handler-case (convert-query-arg argd)
	(error (c)
	  (push (list argd c) failures))))
    (when failures
      (error "Errors in query arguments:~%~{  ~S (~A):~%     ~A~^~%~}"
	     (mapcan #'(lambda (failure)
			 (destructuring-bind (argd condition) failure
			   (list (q-arg-name argd)
				 (q-arg-description argd)
				 condition)))
		     failures)))))

(defun convert-query-arg (argd)
  (let ((name (q-arg-name argd))
	(type (q-arg-type argd)))
    (setf (query-arg name)
	  (apply (q-arg-type-converter type)
		 argd
		 (query-arg name)
		 (cdr type)))))

(defmacro define-query-arg-type (name lambda-list &body body)
  (let ((converter-name (concat-name "%" name "QUERY-ARG-CONVERTER")))
    `(progn
       (setf (get ',name 'query-arg-converter) ',converter-name)
       (defun ,converter-name ,lambda-list
	 ,@body))))

(defun q-arg-type-converter (type)
  (or (get (car type) 'query-arg-converter)
      (error "Unknown query arg type: ~S." type)))

;;; Predefined converters

;;; :int

(define-query-arg-type :int (argd argval low &optional high)
  (unless argval
    (error "No value for ~S." (q-arg-description argd)))
  (let ((i (string->int argval)))
    (if (if high (<= low i high) (<= low i))
	i
      (error "The value of ~S, ~S, is not in the range ~S..~@[~S~]."
	     (q-arg-description argd)
	     argval
	     low
	     high))))

;;; :int*

(define-query-arg-type :int* (argd argval low &optional high)
  ;; "*" can be used instead of "inf".
  (unless argval
    (error "No value for ~S." (q-arg-description argd)))
  (let ((i (string->int* argval)))
    (if (if (numberp i)
	    ;; See if an ordinary int is in range
	    (if high (<= low i high) (<= low i))
	  ;; See if infinity is in range
	  (null high))
	i
      (error "The value of ~S, ~S, is not in the range ~S..~@[~S~]."
	     (q-arg-description argd)
	     argval
	     low
	     high))))

(defun string->int* (string)
  (cond ((string-equal string "inf")
	 :inf)
	((string-equal string "*")
	 :inf)
	(t
	 (string->int string))))

(defun int*-p (x)
  (or (integerp x)
      (eq x :inf)))

;;; :name

(define-query-arg-type :name (argd argval &rest allowed)
  (unless argval
    (error "No value for ~S." (q-arg-description argd)))
  (let ((name (intern (string-upcase argval))))
    (if (member name allowed)
	name
      (error "Invalid value for ~S: ~S;~%~
              Legal values are: ~:(~{~S~^, ~}~)."
	     (q-arg-description argd)
	     argval
	     allowed))))

;;; :names

(define-query-arg-type :names (argd argval &rest allowed)
  ;; Allows there to be no names.
  (let ((q-arg-name-converter (get :name 'query-arg-converter)))
    (mapcar #'(lambda (n)
		(apply q-arg-name-converter argd n allowed))
	    (if (listp argval)
		argval
	      (list argval)))))

;;; :string

(define-query-arg-type :string (argd argval)
  (or argval (error "No value for ~S." (q-arg-description argd))))

;;; :text

(define-query-arg-type :text (argd argval)
  (if argval
      (decode-query-string argval)
    (error "No value for ~S." (q-arg-description argd))))

;;; :optional

(define-query-arg-type :optional (argd argval type)
  (if (null argval)
      nil
    (apply (q-arg-type-converter type)
	   argd
	   argval
	   (cdr type))))

;;; :checkbox

(define-query-arg-type :checkbox (argd argval)
  (if (null argval)
      nil
    (let ((name (intern (string-upcase argval))))
      (if (eq name 'on)
	  name
	(error "Checkbox ~S has value ~S instead of \"on\"."
	       (q-arg-description argd)
	       argval)))))


;;; DECODE-QUERY-STRING handles + and %xx decoding.  It returns its
;;; argument unmodified if no decoding is needed.

(defun decode-query-string (input)
  (let ((in-len (length input))
	(in-pos 0)
	(output nil)			;allocated only if needed
	(out-pos 0))			;N.B. out-pos <= in-pos
    (while (< in-pos in-len)
      (let ((char (schar input in-pos)))
	(cond ((char= char #\+)
	       ;; + becomes space
	       (when (null output)
		 (setq output (copy-seq input)
		       out-pos in-pos))
	       (setf (schar output out-pos) #\Space)
	       (incf in-pos)
	       (incf out-pos))
	      ((and (char= char #\%)
		    (< (+ in-pos 2) in-len))
	       (let ((x1 (digit-char-p (schar input (+ in-pos 1)) 16))
		     (x2 (digit-char-p (schar input (+ in-pos 2)) 16)))
		 (cond ((and x1 x2)
			;; %xx becomes the corresponding char
			(when (null output)
			  (setq output (copy-seq input)
				out-pos in-pos))
			(setf (schar output out-pos)
			      (int-char (+ (* 16 x1) x2)))
			(incf in-pos 3)
			(incf out-pos))
		       (t
			;; Not sure if % can occur without xx,
			;; but we allow it.
			(when output
			  (setf (schar output out-pos) char)
			  (incf out-pos))
			(incf in-pos)))))
	      (t
	       ;; An ordinary character: no conversion.
	       (when output
		 (setf (schar output out-pos) char)
		 (incf out-pos))
	       (incf in-pos)))))
    ;; We're done.
    (if output
	(if (= out-pos (length output))
	    output
	  (subseq output 0 out-pos))
      input)))


;;;; HTML output

;;; See also start-html and *in-html* above.

;;; HTML output produced by the routines in this section goes to the
;;; stream *html-out*.  Normally this will be the same as *terminal-io*.

(defparameter *html-out* nil)		;stream to send HTML output to

(defvar *printing-html* nil)		;true if we are /\/ but not yet

(defun ensure-html-out ()
  (declare (special *left-io*))
  (if *left-io*
      (error "Can't do HTML output in connect mode.")
    (unless *html-out*
      (setq *html-out* *terminal-io*))))

(defun html-format (format-string &rest format-args)
  (ensure-html-out)
  (apply #'format *html-out* format-string format-args))

(defun html-xp-format (format-string &rest format-args)
  (ensure-html-out)
  (apply #'mon-format *html-out* format-string format-args))

(defun html-line (format-string &rest format-args)
  (ensure-html-out)
  (format *html-out* "~&~?~%" format-string format-args))

(defmacro html-paragraph (&body forms)
  `(progn
     (html-line "<P>")
     ,@forms))

(defmacro html-block (type &body forms)
  (check-type type string)
  `(unwind-protect
        (progn (html-line "<~A>" ,type)
	       ,@forms)
     (html-line "</~A>" ,type)))

(defmacro html-item (tag &body forms)
  ;; Used in blocks, e.g. for an <LI> item in a <UL>
  (check-type tag string)
  `(progn
     (html-line "<~A>" ,tag)
     ,@forms))

(defun html-tag (tag format-string &rest format-args)
  (check-type tag string)
  (unwind-protect
       (html-format "<~A>~?" tag format-string format-args)
    (html-format "</~A>" tag)))

(defun html-tag-line (tag format-string &rest format-args)
  (check-type tag string)
  (unwind-protect
       (html-format "~&<~A>~?" tag format-string format-args)
    (html-format "</~A>~%" tag)))

(defun html-anchor (url description)
  (check-type url string)
  (check-type description string)
  (unwind-protect
       (html-line "<A HREF=~S>~%~A" url description)
    (html-line "</A>")))

(defun html-tmp-anchor (name type description)
  (html-anchor (web-tmp-url name type) description))

(defun html-report-condition (c)
  (html-paragraph
    (html-line "Processing stopped because the following condition")
    (html-line "was signalled:")
    (html-block "PRE"
      (html-format "~A"
	(html-encode-pre-string (princ-to-string c))))))

#+:undef
(defun html-report-condition (c)
  (html-paragraph
    (html-line "Processing stopped because the following condition")
    (html-line "was signalled:")
    (html-block	"LISTING"		;"BLOCKQUOTE" ?
      (html-format "~A" c))))

(defun html-describe-query-args ()
  (html-paragraph
    (html-line "Query-argument values:")
    (html-block "PRE"
      (loop for (name . value)
	    in (canonical-description-order
		(hash-table-alist *query-arg-table*))
	    do (html-line "  ~S = ~S" name value)))))

(defun html-print-oplan-greeting ()
  (html-block "PRE" (print-oplan-greeting *html-out*)))

(defmacro html-time (form)
  `(html-block "PRE" (time ,form)))

(defun html-report-plan-statistics ()
  (html-paragraph
    (html-line "Planning statistics:")
    (html-block "PRE"
      (loop for (stat . value) in (request-plan-statistics-list)
	    do (html-line " ~3T ~S ~23T = ~S" stat value)))))


;;; Length for strings containing HTML tags

(defun html-print-length (string)
  ;; Doesn't count chars in <...>.  Assumes no <...>s are nested.
  (declare (string string))
  (let ((in-tag-p nil)
	(len 0))
    (declare (fixnum len))
    (fix-dotimes (i (length string) len)
      (let ((char (schar string i)))
	(declare (character char))
	(cond ((char= char #\<)
	       (setq in-tag-p t))
	      ((char= char #\>)
	       (setq in-tag-p nil))
	      ((not in-tag-p)
	       (setq len (fix1+ len))))))))


;;; Encoding for PRE blocks

(defparameter *html-pre-char-alist*
  '(( #\< . "&lt;"  )
    ( #\> . "&gt;"  )
    ( #\& . "&amp;" )))

(defun html-encode-pre-string (s)
  ;; Converts chars into char sequences as specified by *html-pre-char-alist*.
  ;; Assumes that all transformations make the text longer.
  (let ((outlen 0))
    ;; See how long the output string needs to be.
    (dotimes (i (length s))
      (let ((e (lookup (schar s i) *html-pre-char-alist*)))
	(incf outlen (if e (length e) 1))))
    ;; If the output would be the same length as the input, no
    ;; encoding is needed.
    (when (= outlen (length s))
      (return-from html-encode-pre-string s))
    ;; Encode away...
    (let ((output (make-string outlen))
	  (outpos 0))
      (dotimes (i (length s))
	(let* ((c (schar s i))
	       (e (lookup c *html-pre-char-alist*)))
	  (cond (e (dotimes (ei (length e))
		     (setf (schar output outpos) (schar e ei))
		     (incf outpos)))
		(t (setf (schar output outpos) c)
		   (incf outpos)))))
      (assert (= outpos (length output)))
      output)))


;;; Links to standard results

(defvar *web-results-wanted* nil)	;fn or list of symbols

(defun html-standard-result-links (domain task *web-results-wanted*)

  ;; A PostScript graph

  (when (web-result-wanted-p 'PostScript+Graph)
    (request-psgraph :file-one-page
      :title (concat-string domain ":" task)
      :output-file (web-tmp-filename "graph" "ps"))
    (html-item "LI"
      (html-tmp-anchor "graph" "ps" "PostScript graph")
      (html-line "of the plan")))

  ;; Plan narrative

  (when (web-result-wanted-p 'Narrative)
    (request-plan-view
      :mode :narrative
      :levels :all
      :output-file (web-tmp-filename "narrative" "txt"))
    (html-item "LI"
      (html-tmp-anchor "narrative" "txt" "Plan narrative")))

  ;; KP trace

  (when (web-result-wanted-p 'KP+trace)
    (html-item "LI"
      (html-tmp-anchor "kp-trace" "txt" "KP trace output")))

  ;; World at end of node-1
  
  (when (web-result-wanted-p 'World-1)
    (request-world-view "1"
      :mode :file
      :output-file (web-tmp-filename "world-1" "txt"))
    (html-item "LI"
      (html-tmp-anchor "world-1" "txt"
        "World state when the plan starts")))

  ;; World at end of node-2

  (when (web-result-wanted-p 'World-2)
    (request-world-view "2"
      :mode :file
      :output-file (web-tmp-filename "world-2" "txt"))
    (html-item "LI"
      (html-tmp-anchor "world-2" "txt"
	"World state when the plan finishes"))))

(defun web-result-wanted-p (key)
  (if (functionp *web-results-wanted*)
      (funcall *web-results-wanted* key)
    (member key *web-results-wanted*)))


;;;; TF output

(defvar *tf-out* nil)

(defmacro with-tf-output-file (name &body forms)
  `(with-open-file (*tf-out* ,name :direction :output :if-exists :error)
     ,@forms))

(defun tf-format (format-string &rest format-args)
  (apply #'format *tf-out* format-string format-args))

(defun tf-line (format-string &rest format-args)
  (format *tf-out* "~&~?~%" format-string format-args))


;;;; Output macro
;;;
;;; (OUTPUT directive...) is similar to the Franz Lisp "msg" macro.
;;; It is in some ways an alternative to FORMAT.  A directive can be:
;;;   a literal string       --  the string is printed by PRINC
;;;   the symbol //          --  a newline is printer by TERPRI
;;;   some other symbol      --  a variable: the value is printed by PRINC
;;;   a cons which can be:
;;;     a form               --  the value is printed by PRINC
;;;     a special directive  --  see below
;;;   something else         --  treated as a literal and printed by PRINC
;;;
;;; Special directives are defined by DEFINE-OUTPUT-EXPANDER.  This
;;; defines a function that is called with the cdr of the directive
;;; as it's arguments.  The function returns a form that is used
;;; directly in the expansion of OUTPUT.  Predefined special directives
;;; are:
;;;   (:STREAM form)
;;;      redirects output to the value of the form
;;;   (:INCLUDE form)
;;;      places the form in-line in the expansion of OUTPUT.
;;;   (:FORMAT format-string format-arg...)
;;;      becomes a call to FORMAT.
;;;

(defvar *output* (make-synonym-stream '*terminal-io*))

(defmacro output (&rest directives)
  `(let ((*output* *output*))
     ,@(mapcar #'expand-output-directive directives)))

(eval-when (eval compile load)
  
(defun expand-output-directive (d)
  (cond ((stringp d)
	 `(princ ,d *output*))
	((eq d '//)
	 `(terpri *output*))
	((symbolp d)
	 `(princ ,d *output*))
	((consp d)
	 (if (symbolp (car d))
	     (let ((expander (get (car d) :output-expander)))
	       (if expander
		   ;; Special-case expansion
		   (apply expander (cdr d))
		 ;; A random form
		 `(princ ,d *output*)))
	   (error "Illegal output directive: ~S." d)))
	(t
	 ;; A random object
	 `(princ ',d *output*))))

);end eval-when

(defmacro define-output-expander (name parameters &body forms)
  (let ((fn-name (concat-name name "-OUTPUT-EXPANDER")))
    `(eval-when (eval compile load)
       (defun ,fn-name ,parameters
	 ,@forms)
       (setf (get ',name :output-expander) ',fn-name))))

;;; :STREAM sets *output* and hence the output destination (dynamically)
;;; within an OUTPUT form.

(define-output-expander :stream (form)
  `(setq *output* ,form))

;;; :INCLUDE gives a form complete control over what output occurs.

(define-output-expander :include (form)
  form)

;;; :FORMAT

(define-output-expander :format (format-string &rest format-args)
  `(format *output* ,format-string ,@format-args))


;;; Outputting comma-separtated items, usually TF

(defun output-comma-separated (prefix strings)
  (when strings
    (output // prefix (car strings))
    (cond ((null (cdr strings))
	   (output ";"))
	  (t
	   (output ",")
	   (output-comma-separated prefix (cdr strings))))))


;;;; Demo utilities

;;; An assortment of potentially useful items.

;;; KP-1 trace to a file.  Assumes the output level starts at 0
;;; and that the output level wanted for the trace is 1.  

(defun start-kp-trace ()
  ;; Redirects output to a trace file and arranges for the file to
  ;; be closed when O-Plan (running in subr mode) exits.
  (redirect-component-output :kp-1 (web-tmp-filename "kp-trace" "txt"))
  (set-component-debug-level :kp-1 1))

(defun stop-kp-trace ()
  ;; Use after start-kp-trace.  Turns off output but leaves the
  ;; trace stream open.
  (finish-output (pprocess-symbol-value (find-pprocess :kp-1) '*debug-io*))
  (set-component-debug-level :kp-1 0))

(defmacro with-kp-trace (&body body)
  ;; N.B. *not* equiv to start-kp-trace + stop-kp-trace, because it
  ;; handles closing the stream itself rather than posting an exit-action
  ;; to do it.
  (let ((stream (gensym "STREAM")))
    `(with-open-file (,stream (web-tmp-filename "kp-trace" "txt")
			      :direction :output
		              :if-exists :error)
       (unwind-protect
	   (progn (set-component-output-stream :kp-1 ,stream)
		  (set-component-debug-level :kp-1 1)
		  ,@body)
	 (set-component-debug-level :kp-1 0)))))


;;;; String and other (low-level) utilities

(defun getenv-else (default string)
  (or (getenv string)
      default))

(defun getenv-else-error (string)
  (or (getenv string)
      (error "No value for environment variable ~S." string)))

;;; Print-length determines how many characters are needed to print
;;; various simple objects (symbols, strings, integers) w/o escape chars.

(defun print-length (x)
  (cond ((stringp x) (length x))
	((symbolp x) (length (symbol-name x)))
	((integerp x)
	 (cond ((>= x 0) (int-print-length x))
	       (t        (1+ (int-print-length x)))))
	(t
	 (error "Can't determine the print-length of ~S." x))))

(defun int-print-length (i &optional (base 10))
  (do ((c 1 (1+ c))
       (q (abs i) (floor q base)))
      ((< q base)
       c)))

;;; *char=
;;;
;;; /\/: *char= is needed at least in AKCL.  With *char= we get
;;; if(!((V1)==(((V2))->ust.ust_self[V6]))) ...; otherwise we get
;;; if(!(char_code(code_char(V1))==char_code(code_char(((V2))->
;;; ust.ust_self[fix(V7)]))))... .  Why we have fix(V7) is not clear.
;;; [This is from (disassemble 'break-string-at).]

(defmacro *char= (c1 c2)
  `(char= (the character ,c1) (the character ,c2)))

;;; (Break-string-at char string) takes a string containing fields
;;; separated by char and returns a list of the fields, in order, as
;;; strings.

(defun break-string-at (char string)
  (declare (character char) (simple-string string))
  (let ((start 0)
	(parts '()))
    (declare (fixnum start))
    (fix-dotimes (i (length string))
      (when (*char= char (schar string i))
	(push (subseq string start i) parts)
	(setq start (fix1+ i))))
    (nreverse (cons (subseq string start) parts))))

;;; ---------------------------- Change History ----------------------------
;;; (Who)   (When) 		     (What)
;;;
