;;;; File: allegro-xwindowio.lsp
;;; SCCS Version: %W%
;;; Contains: Allegro code for using X windows for IO.
;;; Author: Jeff Dalton
;;; Created: April 1994
;;; Updated: Fri Apr 22 18:08:47 1994 by Jeff Dalton
;;; Release Version: %Y%
;;; Copyright: (c) 1994, AIAI, University of Edinburgh
;;; This material may be reproduced by or for the U.S. Government pursuant
;;; to the copyright license under the clause at DFARS 252.227-7032
;;; (June 1975) -- Rights in Technical Data and Computer Software (Foreign).

(in-package :util)

;;; *xterm-debug* is a stream for debugging output produced when
;;; trying to start xterms.  It's normally bound to a stream that
;;; discards all output, obtained by calling make-broadcast-stream
;;; with no arguments.

(defparameter *xterm-debug* (make-broadcast-stream))

(defun set-term-modes (tty)
  (declare (ignore tty))
  nil)

(defun run-xterm-for-io (&optional (args ""))
  "Opens a window for input and output. 
   Arguments:
     args - A string or list containing extra args for xterm.
   Returns:
     A bidirectional stream, and the device name of the tty device."
  
  (declare (type (or string list) args))
  
  (prog (pseudo-term pseudo-term-letter pseudo-term-number
	 io-term io-term-name)
    ;; Find a pseudo terminal to use.
    RETRY
    (block found-a-term
      (dolist (letter '("p" "q" "r"))
	(dolist (number '("0" "1" "2" "3" "4" "5" "6" "7" "8" "9" 
			  "a" "b" "c" "d" "e" "f"))
	  (setq pseudo-term
		(ignore-errors
		  (open (concatenate 'string "/dev/pty" letter number)
			:class 'excl::bidirectional-terminal-stream
			:direction :io
			:if-exists :append
			:if-does-not-exist nil)))
	  (when pseudo-term
	    (setq pseudo-term-letter letter)
	    (setq pseudo-term-number number)
	    (setq io-term-name
		  (concatenate 'string "/dev/tty" letter number))
	    (return-from found-a-term))))
      (error "Can't find a free pseudo terminal."))

    (format *xterm-debug* "~&Using /dev/pty~A~A; "
	    pseudo-term-letter pseudo-term-number)

    ;; Run xterm
    (multiple-value-bind (comms-stream error-stream pid)
	(excl:run-shell-command
	   (concat-shell-command
	      "xterm"
	      `(,(concatenate 'string "-S"
		    pseudo-term-letter
		    pseudo-term-number
		    (prin1-to-string
		     (excl::stream-output-fn pseudo-term)))
		,@(if (listp args) args (break-args args))))
	   :wait nil)
      (declare (ignore comms-stream error-stream))
      (register-xterm-pid pid)
      (format *xterm-debug* "xterm pid = ~S.~%" pid))

    #+:undef
    (close pseudo-term)			;child process should have it by now

    (setq io-term
	  (open io-term-name
		:class 'excl::bidirectional-terminal-stream
		:direction :io
		:if-exists :append))
    
    ;; Set terminal modes
    ;; Note that the redirection (> or <) depends on the version of stty.
    (system (concat-string "stty sane > " io-term-name))
    
    ;; Wait for some gumpf from xterm, so that we know that it has started ok.
    (read io-term nil :eof)
    ;; Success!
    (close pseudo-term)			;/\/ movd from above
    (set-term-modes io-term)
    (return
      (values io-term io-term-name))))


;;; Break-args from support/xfork.

;;; Process expects a list of argv arguments, but you may have them
;;; all together in a string.

;;; (Break-args string) takes a string containing space-separated
;;; command-line arguments and returns a list of the arguments, each
;;; as a separate string.  It can handle args that begin and end
;;; with " but cannot handle escape chars.

;;; /\/: This is a bit perverse in Allegro, because we have to concat
;;; the args into a single string in order to run a program.  So if we
;;; break up a string, we'll eventually put it back together again.

(defun break-args (string)
  (let ((p 0)
	(args '())
	(chars '()))
    (flet ((current () (schar string p))
	   (next () (incf p))
	   (done () (>= p (length string))))
      (loop
        ;; Scan off leading spaces
        (loop (when (done) (return))
	      (if (char= (current) #\space)
		  (next)
		(return)))
	;; Scan next arg
	(when (done)
	  (return (nreverse args)))
	(setq chars '())
	(cond ((char= (current) #\")
	       ;; Scan quoted string
	       (next)
	       (loop (cond ((or (done) (char= (current) #\"))
			    (push (coerce (nreverse chars) 'string)
				  args)
			    (next)
			    (return))
			   (t (push (current) chars)
			      (next)))))
	      (t
	       ;; Scan to next space
	       (loop (cond ((or (done) (char= (current) #\space))
			    (push (coerce (nreverse chars) 'string)
				  args)
			    (next)
			    (return))
			   (t (push (current) chars)
			      (next))))))))))


;;; Child process operations

;;; Record of xterm pids

(defvar *xterm-pids* '())

(defun register-xterm-pid (pid)
  (assert (not (member pid *xterm-pids*)))
  (setq *xterm-pids* (nconc *xterm-pids* (list pid))))

;;; The xterms exit if they get an INT signal.  HUP isn't enough on Suns.

(defun terminate-xterms-if-necessary ()
  (let ((live-pids (remove-if-not #'still-running-p *xterm-pids*)))
    (when live-pids
      (system
        (format nil "kill -INT ~{~D~^ ~}" live-pids)))))


;;; (WAITPID-NO-HANG pid) returns (pid . status) or (0 . ?) if
;;; the status is not yet available.  (-1 . ?) indicates an error
;;; of some sort, including the case where a signal was received
;;; by the calling process.
;;;
;;;   int waitpid(pid, statusp, options)
;;;   int pid;
;;;   int *statusp;
;;;   int options;
;;;
;;; From <sys/wait.h>:
;;;
;;;   #define WNOHANG         1       /* dont hang in wait */
;;;

; (in-package :util)

(export '(waitpid-no-hang still-running-p))

(defconstant WNOHANG 1)

(defun still-running-p (pid)
  nil)					;pretend we know it isn't /\/

;;; ---------------------------- Change History ----------------------------
;;; (Who)   (When) 		     (What)
;;;
