;;;; File: kcl-conditions.lsp
;;; SCCS Version: %W%
;;; Contains: "Fake" CL condition system for KCL
;;; Author: Jeff Dalton <J.Dalton@ed.ac.uk>
;;; Created: June 1994
;;; Updated: Sat Jan 28 00:29:19 1995 by Jeff Dalton
;;; Release Version: %Y%
;;; Copyright: (c) 1994, AIAI, University of Edinburgh
;;; This material may be reproduced by or for the U.S. Government pursuant
;;; to the copyright license under the clause at DFARS 252.227-7032
;;; (June 1975) -- Rights in Technical Data and Computer Software (Foreign).

;;; This is a minimal condition system for KCL and AKCL which do not have
;;; the standard CL condition system built-in.

(in-package :util)

;; Condition types
(export '(condition simple-condition serious-condition))
(export '(error simple-error))
(export '(warning simple-warning))

;; Defining form
(export '(define-condition))		;N.B.: a very limited version

;; Constructor
(export '(make-condition))

;; Operations on condition objects
(export '(condition-p condition-format-string condition-format-arguments))

;; Condition slots
; (export '(format-string format-arguments))

;; Signalling operations
(export '(signal signal-error))

;; Condition handling
(export '(handler-bind handler-case ignore-errors))

;; Restarts
; /\/: No restart support


;;; Defining condition types

;;; We support only a very limited form of define-condition.  There can
;;; be only one parent type and there is no way to specify slots or options.
;;; This is promarily because define-condition takes defclass-style slot
;;; specs which are too much trouble for us to handle when we can't call on
;;; defclass.

(defmacro define-condition (name parent-types &rest slots-and-options)
  (unless (= (length parent-types) 1)
    (error "Condition type ~S can have only one parent." name))
  (unless (null slots-and-options)
    (error "Condition type ~S cannot have slots or options." name))
  `(define-condition-struct (,name (:include ,(car parent-types)))))

;;; Define-condition-struct is internal.

(defmacro define-condition-struct (spec &rest slots)
  (let* ((name (if (consp spec) (car spec) spec))
	 (opts (if (consp spec) (cdr spec) nil))
	 (make-fn (make-make-fn name)))
    `(progn
       (setf (condition-make-fn ',name) ',make-fn)
       (defstruct (,name
		   (:print-function print-condition)
		   (:constructor ,make-fn)
		   ,@opts)
	 ,@slots))))

(defmacro condition-make-fn (cond-type) `(get ,cond-type 'condition-make-fn))

(eval-when (eval compile load)
  (defun make-make-fn (name)
    (intern (concatenate 'string "MAKE-" (symbol-name name)))))

(defun print-condition (cond stream depth)
  (declare (ignore depth))
  (if (or *print-escape* (null (condition-format-string cond)))
      (format stream "#<~S \"~?\">"
	      (type-of cond)
	      (or (condition-format-string cond) "")
	      (condition-format-arguments cond))
    (apply #'format stream
	   (condition-format-string cond)
	   (condition-format-arguments cond))))

;;; Constructor

(defun make-condition (type &rest slot-initializations)
  (apply (or (condition-make-fn type)
	     (error "~S is not a condition type." type))
	 slot-initializations))


;;; Built-in condition types

;;; I think it's only simple-{condition,warning,error} that are defined
;;; to have the format slots.  But to do that with only single inheritance,
;;; we have to introduce the slots at a common ancestor, with condition
;;; as the obvious choice.

(defstruct (condition
	     (:constructor %make-condition)
	     (:print-function print-condition))
  ;; Need to define condition by hand because the constructor can't
  ;; be called make-condition.
  (format-string nil)
  (format-arguments nil))

(setf (condition-make-fn 'condition) '%make-condition)

(define-condition-struct (simple-condition  (:include condition)))
(define-condition-struct (serious-condition (:include condition)))
(define-condition-struct (error             (:include serious-condition)))
(define-condition-struct (simple-error      (:include error)))
(define-condition-struct (warning           (:include condition)))
(define-condition-struct (simple-warning    (:include warning)))


;;; Signalling

;;; *Handlers* is a list of lists of handler bindings, one list of bindings
;;; for each dynamically nested handler-bind.  Each binding has the form
;;; (typespec . function).  Each handler-bind rebinds *handlers*, adding
;;; its list of bindings to the front.

(defvar *handlers* '())

(defun signal (datum &rest arguments)
  (let ((c (condition-from-args 'signal 'simple-error datum arguments)))
    (do ((handlers *handlers* (cdr handlers)))
	((null handlers) nil)
      ;; The car of handlers is a list of handler bindings.
      ;; The cdr of handlers is the value *handlers* had when those
      ;; bindings were established.  We have to restore that value
      ;; when trying the bindings.
      (let ((*handlers* (cdr handlers)))
	;; Try each binding in turn.
	(dolist (binding (car handlers))
	  (let ((typespec (car binding))
		(handler (cdr binding)))
	    ;; Call the handler if the typespec matches.
	    (when (typep c typespec)
	      (funcall handler c))))))))

(defun signal-error (datum &rest arguments)
  "Signal a condition and enter the debugger if its not handled."
  (let ((c (condition-from-args 'signal-error 'simple-error datum arguments)))
    (signal c)
    (break "No handler for ~S" c)))

(defun condition-from-args (caller default-condition-type datum arguments)
  (cond ((condition-p datum)
	 (if (null arguments)
	     datum
	   (error "~S called with condition ~S;~%~
                   Arguments should be NIL, not ~S."
		  caller datum arguments)))
	((and (symbolp datum) (subtypep datum 'condition))
	 (apply #'make-condition datum arguments))
	((stringp datum)
	 (make-condition default-condition-type
	   :format-string datum
	   :format-arguments arguments))
	(t
	 (error "~S called with illegal arguments (~S ~{~S~^ ~})"
		caller datum arguments))))


;;; Handler-bind

(defmacro handler-bind (bindspecs &rest forms)
  ;; A bindspec has the form (typespec handler-expr).
  (flet ((handler-bindspec->cons (bindspec)
	   (destructuring-bind (typespec handler) bindspec
	     `(cons ',typespec ,handler))))
    `(let ((*handlers*
	    (cons (list ,@(mapcar #'handler-bindspec->cons
				  bindspecs))
		  *handlers*)))
       ,@forms)))

;;; Handler-case

(defmacro handler-case (expression &rest clauses)
  (if (assoc :no-error clauses)
      (expand-no-error-handler-case expression clauses)
    `(block .handler-case-exit.
       (funcall				;call thunk returned by handler
	 (block .handler-case-unwind.
	   (handler-bind ,(mapcar #'hc-clause->handler-binding clauses)
	     (return-from .handler-case-exit.
	       ,expression)))))))

(defun hc-clause->handler-binding (clause)
  (destructuring-bind (typespec var &rest forms) clause
    (let ((handler-body
	   ;; Unwind the dynamic env by returning a thunk
	   `(return-from .handler-case-unwind. #'(lambda () ,@forms))))
      (if var
	  `(,typespec #'(lambda ,var ,handler-body))
	  `(,typespec
	    #'(lambda (.ignore.)
		(declare (ignore .ignore.))
		,handler-body))))))

(defun expand-no-error-handler-case (expression clauses)
  (let* ((no-err-clause (assoc :no-error clauses))
	 (other-clauses (remove no-err-clause clauses)))
    (unless (eq no-err-clause (car (last clauses)))
      (error "In handler-case, :no-error must be last."))
    (destructuring-bind (lambda-list &rest forms) (cdr no-err-clause)
      ;; Almost straight from CLtL II:
      `(block .error-return.
	 (multiple-value-call #'(lambda ,lambda-list . ,forms)
	   (block .normal-return.
	     (return-from .error-return.
	       (handler-case (return-from .normal-return. ,expression)
		 ,@other-clauses))))))))

;;; Ignore-errors

(defmacro ignore-errors (&rest forms)
  `(handler-case (progn ,@forms)
     (error (c) (values nil c))))


;;; Intercepting standard KCL errors

(defun new-error-handler (error-name correctable-p function-name
			  continue-format-string error-format-string
			  &rest args)
  (apply #'signal error-format-string args)
  (apply #'original-error-handler
	 error-name correctable-p function-name
	 continue-format-string error-format-string
	 args))

(unless (fboundp 'original-error-handler)
  (setf (symbol-function 'original-error-handler)
	(symbol-function 'si:universal-error-handler)))

(setf (symbol-function 'si:universal-error-handler)
      (symbol-function 'new-error-handler))

;;; ---------------------------- Change History ----------------------------
;;; (Who)   (When) 		     (What)
;;;
