;;;; File: mcc-client.lsp
;;; SCCS Version: %W%
;;; Contains: Module Communication Channel client code.
;;; Author: Richard Kirby (rbk) and Jeff Dalton
;;; Created: Fri Feb  2 15:30:05 1990
;;; Updated: Sun May 14 01:31:22 1995 by Jeff Dalton
;;; Release Version: %Y%
;;; Copyright: (c) 1992, AIAI, University of Edinburgh
;;; This material may be reproduced by or for the U.S. Government pursuant
;;; to the copyright license under the clause at DFARS 252.227-7032
;;; (June 1975) -- Rights in Technical Data and Computer Software (Foreign).

;;; This file contains routines used by MCC clients.  Server routines
;;; are part of the IM.  See source/interface-manager/mcc-server.lsp.

(in-package :MCC)

;;; /\/: The MCC and IPC packages use each other.  We deal with this
;;; by putting the parts of MCC that IPC needs into mcc-defs, by having
;;; files loaded in the order mcc-defs, ipc, mcc-client (= this file),
;;; and by having mcc-client do the (use-package :IPC), here:

(use-package :IPC)

(defun mcc-wait-for-ack (&optional (channel :mcc))
  "Waits for a :OK message on the given channel."
  ;; /\/: But at present, it doesn't.
  (dev-debug-msg (format nil "Pretended to get :OK on ~S~%" channel)
		 *whoami* :information))

(defun mcc-register-with-im ()
  "Sends a :REGISTER message to the Interface Manager, and then waits for an
   acknowledgment."
  (ipc-write :mcc (mcc-make-message :MCC :REGISTER *whoami*))
  (mcc-wait-for-ack))

(defun mcc-handler (event)
  (apply (ipc-get-handler :MCC-CLIENT (mcc-msg-ident event))
	 (mcc-msg-args event)))

;;; Messages
  
;;; Note that we define messages for :MCC-CLIENT instead of for a
;;; particular p-process name.  This works if we use IPC-GET-HANDLER.

(defmessage (:MCC-CLIENT :DEBUG-LEVEL) (level)
  (dev-set-debug-level level))

(defmessage (:MCC-CLIENT :DIE-DIE-DIE) ()
  ;; Quit the top-level loop.
  ; (ipc-write :mcc (mcc-make-message :MCC :OK))
  ; (throw '.top-level. nil)
  (ipc-terminate))

(defmessage (:MCC-CLIENT :EVAL) (form)
  ;; Evaluate the form safely.
  (let ((result (or (dev-safe-eval form) :ERROR)))
    (ipc-write :mcc
	       (mcc-make-message :MCC :RESULT form result))))

(defmessage (:MCC-CLIENT :OK) ()
  ;; Should never receive an :OK at this level.
  (dev-debug-msg
    (format nil "Unexpected :OK message received.")
    *whoami* :non-fatal-error))

(defmessage (:MCC-CLIENT :SINGLE-STEP) (value)
  ;; Calls the current component's single-step handler
  (when (component-single-step-fn *pprocess*)
    (funcall (component-single-step-fn *pprocess*) value)))

(defmessage (:MCC-CLIENT :INIT) ()
  ;; Calls the current component's init function, then sends OK.
  (when (component-init-fn *pprocess*)
    (funcall (component-init-fn *pprocess*)))
  #+:undef
  (ipc-write :MCC (mcc-make-message :MCC :OK)))

(defmessage (:MCC-CLIENT :INTERRUPT) ()
  ;; Calls the current component's interrupt handler
  (when (component-interrupt-fn *pprocess*)
    (funcall (component-interrupt-fn *pprocess*))))


;;; Sending to :LEFTOUT

;;; It's not necessary to call the -to-planner and -to-ta versions,
;;; but they perform some useful consistency checks.

(defun mcc-send-to-leftout (&rest message)
  (ipc-write :MCC (apply #'mcc-make-message :MCC :PRINT :LEFTOUT message)))

(defun mcc-leftout-to-planner (&rest message)
  ;; Called by the Exec.
  (assert (get-parameter :asexec))
  (ipc-write :MCC (apply #'mcc-make-message :MCC :PRINT :LEFTOUT message)))

(defun mcc-leftout-to-ta (&rest message)
  ;; Called by the Planner.
  (assert (not (get-parameter :asexec)))
  (ipc-write :MCC (apply #'mcc-make-message :MCC :PRINT :LEFTOUT message)))


;;; Sending to :RIGHTOUT

;;; It's not necessary to call the -to-exec and -to-world versions,
;;; but they perform some useful consistency checks.

;;; Note that "rightout" is sometimes :worldin.

(defun mcc-send-to-rightout (&rest message)
  (ipc-write :MCC
    (apply #'mcc-make-message :MCC :PRINT
	   (if (get-parameter :asexec)
	       :WORLDIN
	       :RIGHTOUT)
	   message)))

(defun mcc-rightout-to-exec (&rest message)
  ;; Called by the Planner.
  (assert (not (get-parameter :asexec)))
  (ipc-write :MCC (apply #'mcc-make-message :MCC :PRINT :RIGHTOUT message)))

(defun mcc-rightout-to-world (&rest message)
  ;; Called by the Exec.
  (assert (get-parameter :asexec))
  (ipc-write :MCC (apply #'mcc-make-message :MCC :PRINT :WORLDIN message)))

;;; ---------------------------- Change History ----------------------------
;;; (Who)   (When) 		     (What)
;;;
