;;;; File: system-definitions.lsp
;;; SCCS Version: %W%
;;; Contains: O-Plan system definitions
;;; Author: Jeff Dalton
;;; Created: February 1993
;;; Updated: Mon Jul 10 19:21:39 1995 by Jeff Dalton
;;; Release Version: %Y%
;;; Copyright: (c) 1993, AIAI, University of Edinburgh
;;; This material may be reproduced by or for the U.S. Government pursuant
;;; to the copyright license under the clause at DFARS 252.227-7032
;;; (June 1975) -- Rights in Technical Data and Computer Software (Foreign).

;;; O-Plan system definitions

;;; The usual procedure is to (compile-system 'everthing) and then
;;; build an image by starting a fresh Lisp, doing (load-system 'oplan),
;;; and saving a Lisp image.  The "Makefile" in the main source
;;; directory provides a convenient way to do this and other things.

;;; Note (if you're doing things by hand) that you have to load the
;;; "oplan.lsp" file before loading this one.

;;; To print a description of the system structure, evaluate
;;;   (print-system-tree 'everything)

;;; For a description of the defsystem used here, see support/defsys.lsp.

;;; /\/: We need a better, cleaner division into systems.  Even though
;;; they're all called "systems" most of the systems defined in this file
;;; are modules of other sorts; and some of the larger ones should be
;;; broken up along functional lines.  Package use also needs to be cleaned
;;; up.  At present, we have a lot of :do-after-load calls to use-package
;;; that make things visible in the OPLAN package; and some of the code
;;; that's in the OPLAN package should be given a package of its own.
;;; (We're in this mess for historical reasons.  In version 1.2, only
;;; the TF compiler was written as a system; everything else was compiled
;;; and loaded in an ad hoc way, and only the DM was actually compiled.)

;;; /\/: We may want to use a different defsystem, to explicitly
;;; distinguish between systems and modules, to put each system
;;; definition in its own file, etc.

;;; /\/: AKCL complains if packages that were used at compile-time
;;; aren't also used when the .o file is loaded even if no symbols
;;; from those packages were used in the .lsp file.  We should set
;;; up all the packages at the start?

(in-package :oplan)

(defsystem (everything
	    :required-systems
	      (oplan
	       knowledge-sources
	       known-patches
	       auto-tester
	       test-support
	       #+:undef
	        profiler
	       web-demos
	       worlds
	       simple-defsys))
  ;; Compile this system to compile everything.
  ;; Normally, only the oplan system is loaded.
  )

(defsystem (simple-defsys :directory "support")
  ;; Here just so it will be compiled when the everything system is.
  (defsys))

(defsystem (known-patches :directory "patch"
			  :required-systems (oplan))
  ;; The null-loader steps the files from being loaded when we're
  ;; compiling the system.  It also stops them from being loaded when
  ;; we load the system, but that's ok because they're loaded automatically
  ;; at run-time.
  #|
  ("alt-choice"
    :loader null-loader)
  ("deleters"
    :loader null-loader)
  |#
  )

(defun null-loader (pathname) nil)

(defsystem (auto-tester :directory "auto-tester"
			:required-systems (support
					   task-assigner)
			:do-after-load
			   (use-package :atest :oplan)) ;for the user
  ("../support/fsa")
  (auto-tester
    :requires ("../support/fsa"))
  (standard-tests
    :requires (auto-tester)
    :compiler nil)
  (extra-tests
    :requires (auto-tester
	       standard-tests)
    :compiler nil)
  (release-tests
    :requires (auto-tester
	       standard-tests)		;/\/: for default-replans
    :compiler nil)
  )

(defsystem (test-support :directory "support"
			 :required-systems (support))
  ;; The test systems that :require this are defined separately.
  (t-frame))

(defsystem (profiler :directory "profile")
  (:profile))

(defsystem (web-demos :directory "web/demo"
		      :required-systems (top-level))
  (island-rescue-support)
  (demo-tf-support)
  (simple-tpn-support)
  (london-underground-support)
  (get-url-support))


;;;; O-Plan itself

(defsystem (oplan
	    :directory "single-process"
	    :required-systems
	      (;; /\/ Initialization is (implicitly) required by all
	       ;; component systems.  Perhaps it should be in support.
	       initialization
	       ;; /\/ Put the DM first so that packages get set up
	       ;; in a way that works before other components are
	       ;; compiled.
	       database-manager
               task-assigner
	       interface-manager
	       agenda-manager
	       diary
	       micro-exec
	       world
	       ;database-manager
	       pw-viewers		; load at run-time? (used to)
	       ks-platform
	       ;knowledge-sources	; loaded at run-time
	       top-level
	       ))
  ;; The :required-systems include all the systems that must be
  ;; loaded.  Any files directly in the oplan system might not
  ;; really depend on many of them.  A separate top-level system
  ;; is therefore used.
  )

(defsystem (top-level
	    :directory "single-process"
	    :required-systems (support))
  (main)
  (top-level)
  (connect-mode)
  (interface-package
    :defines (:package)
    :do-after-load (use-package :oplan-interface :oplan))
  (subr-mode
    :requires (interface-package))
  (manual-interface
    :defines (:functions)
    :requires (interface-package))
  (program-interface
    :defines (:functions :macros)
    :requires (interface-package))
  (web-support				;should have its own system?  /\/
    :requires (interface-package)))

(defsystem (initialization
	    :directory "single-process"
	    :required-systems (support parser-kit))
  ;; /\/: Maybe move to support?
  ;; /\/: The components ought to depend on this, but we rely on
  ;;      the oplan system loading initialization first.
  (initsystem
    :do-after-load (use-package :is :oplan)))


(defsystem (task-assigner :directory "task-assigner"
			  :required-systems (support))
  ;; /\/: Sneaks dates from defsys.
  (ta-toplevel)
  )

(defsystem (interface-manager :directory "interface-manager"
			      :required-systems (support
						 structure-defs))
  (im-toplevel
    :requires (mcc-server))		;for mcc package exports
  (mcc-server)
  (inter-agent
    :requires (mcc-server))		;for mcc/ipc package setup
  (control-panel)
  (plan-viewer
    :requires (mcc-server))
  )

(defsystem (agenda-manager :directory "agenda-manager"
			   :required-systems (support
					      structure-defs))
  (am-toplevel)
  (am-services
    :requires (am-toplevel))		;for defvars /\/
  )

(defsystem (diary :directory "diary"
		  :required-systems (support
				     sim-clock))
  ;; /\/: Used only in the Exec agent so far.
  (di-toplevel)
  )

(defsystem (micro-exec :directory "single-process"
		       :required-systems (support
					  sim-clock))
  (micro-exec)
  )


(defsystem (world :directory "world"
		  :required-systems (support
				     sim-clock))
  ;; Used only in the Exec
  (world-package)
  (world-parameters
    :requires (world-package))
  (world-toplevel
    :requires (world-package))
  (separate-world
    :requires (world-package))
  (world-services
    :requires (world-package
	       world-toplevel		;for variable definitions
	       world-parameters))
  (world-loader
    :requires (world-package
	       world-toplevel))
  (world-compiler
    :requires (world-package))
  )

(defsystem (worlds
	     :directory "worlds"
	     :required-systems
	       (micro-exec-world))
  )

(defsystem (micro-exec-world
	     :directory "worlds/micro-exec"
	     :required-systems (world))
  (micro-exec-world))


(defsystem (pw-viewers :directory "pw-viewers"
		       :required-systems (support
					  initialization
					  tf-language
					  task-assigner)) ; shouldn't /\/
  ;; A new viewer should :require viewer-services and will
  ;; indirectly depend on viewer-toplevel as well.
  (viewer-toplevel
    :defines (:package :variables :functions))
  (viewer-services
    :defines (:structures :functions)
    :requires (viewer-toplevel))
  (end-graph
    :defines (:structures :functions)
    :requires (viewer-services))
  (text-viewer
    :requires (viewer-services))
  (psgraph-viewer
    :requires (viewer-services
	       end-graph))
  (autocad-viewer
    :requires (viewer-services))
  (krsl-viewer
    :requires (viewer-services))
  (plan-narrative
    :requires (viewer-services
	       end-graph))
  )

(defsystem (ks-platform :directory "ks-platform"
			:required-systems (support
					   structure-defs
					   shared-datastructures))
  ;; /\/: EXPAND/ACHIEVE requires the "tuple" struct from obactor-psvs
  ;; /\/: ACHIEVABLE comes from the tf-language package.
  (kp-package)
  (kp-toplevel
    :requires (kp-package))
  (kp-supportlib
    :requires (kp-package kp-toplevel))
  (ks-package
    :requires (kp-package))
)

(defsystem (knowledge-sources :directory "knowledge-sources"
			      :required-systems (ks-platform))
  ("KS-ACHIEVE"                :defines (:functions))
  ("KS-ADD-TO-TASK"	       :defines (:functions))
  ("KS-BIND"                   :defines (:functions))
; ("KS-BREAKUP"                :defines (:functions))
  ("KS-CHECK-PLAN"             :defines (:functions))
  ("KS-COMPUTE"                :defines (:functions))
  ("KS-CONDITION"              :defines (:functions))
  ("KS-CONTINUE-EXECUTION"     :defines (:functions))
; ("KS-DISPATCH-1"             :defines (:functions))
; ("KS-DISPATCH-2"             :defines (:functions))
  ("KS-DOMAIN"                 :defines (:functions))
  ("KS-EXECUTE"                :defines (:functions))
  ("KS-EXECUTION-FAILURE"      :defines (:functions))
  ("KS-EXECUTION-SUCCESS"      :defines (:functions))
  ("KS-EXPAND"                 :defines (:functions))
  ("KS-EXTRACT"                :defines (:functions))
; ("KS-FACT"                   :defines (:functions))
  ("KS-FIX"                    :defines (:functions))
  ("KS-INIT"                   :defines (:functions))
  ("KS-KILL"                   :defines (:functions))
; ("KS-NEWDOMAIN"              :defines (:functions))
  ("KS-NO-MORE-ALTERNATIVES"   :defines (:functions))
  ("KS-NO-MORE-AGENDA-ITEMS"   :defines (:functions))
  ("KS-OR"                     :defines (:functions))
  ("KS-PLANNER-FINISHED"       :defines (:functions))
  ("KS-POISON-STATE"           :defines (:functions))
  ("KS-SET-TASK"               :defines (:functions))
  ("KS-TESTSTAGING"            :defines (:functions))
  ("KS-UNEXPECTED-WORLD-EVENT" :defines (:functions))
  ("KS-USER"                   :defines (:functions))
  )

;;; Database manager

(defsystem (database-manager
	     :directory "database-manager"
	     :required-systems (database
				tf-compiler))
  ("dm-toplevel")
  ("dm-services")
  ("schemacode"
    :requires ("dm-services"))
  (sanity-checker)
  )

(defsystem (database
	     :directory "database-manager"
	     :required-systems
	       (support
		obase
		tf-language		;/\/ must be loaded early
		structure-defs
		shared-datastructures))
  ;; /\/ Note the hidden dependency: tf-language has to be before
  ;; structure-defs and shared-datastructures, because they add
  ;; symbols to the OPLAN package that have the same names as TF
  ;; exports.
  
  ;; Main DM files.
  (gopher)
  (psvs
    :do-after-load (use-package :psv :oplan))
  (obactor-psvs
    :requires (psvs))
  (:qa
    :defines (:package :functions)
    :requires (obactor-psvs
	       gopher))
  (:qa-all
    :requires (:qa))
  (tgm					;also requires tgm-datastructures
    :requires (gopher
	       :qa
	       psvs))
  (sc-rum)

  ;; /\/: Here we have some files in the OPLAN package that expect lots
  ;; of packages to be available in the use-list.  That's why lots of
  ;; systems and files have a :do-after-load that does a use-package.
  (atm)
  (triggers
    :requires (psvs))
  )

(defsystem (shared-datastructures
	     :directory "database-manager"
	     :required-systems (support
				structure-defs)	;nodes needs schema-defs /\/
	     :do-after-load
	       ;; /\/: For code in the OPLAN package.
	       (use-package '(:ads :tgm) :oplan))
  ;; Used by DM and KP.
  ("nodes")
  (tpn					;shared? /\/
    :requires ("nodes"))
  (tgm-datastructures)
  )

(defsystem (structure-defs :directory "common"
			   :required-systems (support
					      tf-language))
  ;; /\/ These files used to be in source/patch because they were
  ;; needed by more than one component.
  (schema-defs
    :do-after-load (use-package :schema-defs :oplan))
  (eventdef)
; (schemadef				;requires MON
;   :requires (schema-defs))
  (agendadef)
  (res-struct				;does (in-package :QA) /\/
    :do-after-load (use-package :qa :oplan))
  )


;;; OBase

(defsystem (obase :directory "database-manager"
		  :required-systems
		    (support		;only for contexts? /\/
		     tf-language)	;for the VALUE symbol /\/
		  :do-after-load
		    ;; /\/ For code in the OPLAN package
		    (use-package ':ob :oplan))
  (obase)
  )


;;; TF language

(defsystem (tf-language
	    :directory "database-manager/tf-compiler"
	    :do-after-load
	      ;; So that we can read TF files in the OPLAN package
	      (use-package :tf :oplan)
	   )
  (tf-pack
   :defines (:package))
  )

;;; TF Compiler

(defsystem (tf-compiler
	    :directory "database-manager/tf-compiler"
	    :required-systems (support
			       obase		;see com-pack /\/
			       structure-defs
			       tf-language
			       parser-kit)
	    :do-after-load
	      (use-package :tf-compiler :oplan))
  (com-pack				;required by all later files
    :defines (:package))
  (scan
    :requires (com-pack))
  (com-defs
    :requires (com-pack))
  (o-face				;/\/: What does it really require?
    :defines (:functions)
    :requires (com-pack scan com-defs))
  (parser
    :defines (:functions)		;/\/: but *user-defaults*
    :requires (com-pack scan com-defs))
  (complr
    :defines (:functions)
    :requires (com-pack com-defs))
  (graphs
    :defines (:functions)
    :requires (com-pack))
  (levels
    :defines (:functions)
    :requires (com-pack com-defs))
  (analysis
    :defines (:functions)
    :requires (com-pack com-defs))
  (top-level
    :defines (:functions)
    :requires (com-pack com-defs))
  (embedded-tfc
    :defines (:functions)
    :requires (com-pack com-defs))
  )


;;; Parser Kit

(defsystem (parser-kit
	    :directory "support/parser-kit")
  (parser-kit-pack)
  (descent
    :requires (parser-kit-pack))
  )


;;; Simulated time

(defsystem (sim-clock
	    :directory "support"
	    :required-systems (support))
  ;; Requires support for pseudo-processes, time-util, and defun-inline
  (sim-clock-package
    :defines (:package))
  (sim-clock
    :requires (sim-clock-package)))
	    


;;; Support

;;; /\/: May want to separate out the support needed only by -toplevel
;;; files and the like (ipc-support?), so that we don't e.g. recompile
;;; the whole DM every time we change something about p-processes.

;;; The support system contains definitions that can be used throughout
;;; O-Plan.  Higher-level systems typically have support as one of their
;;; :required-systems.  Modifications to support should be relatively
;;; rare, but when they happen the requiring systems will all have to be
;;; recompiled.  (If this becomes a problem, we could separate out the
;;; parts of support that affect compilation or else divide it into smaller
;;; systems; but it would require a fair amount of work.)

;;; The current subsystems of support are base-support, common-support,
;;; dependent-support, and external-support.

;;; Base-support contains portable code needed by both common-support
;;; and dependent-support.  This is the lowest level of "common code"
;;; in O-Plan, and everything outside external-support can assume it
;;; exists.

;;; Most of the support code is in common-support.  This code should work
;;; in at least to all the Common Lisps that will be used with O-Plan.
;;; The files in common-support may contain implementation-dependent code,
;;; but only internally.  (We allow this so that we can use self-contained
;;; packages that handle their own portability and that shouldn't be broken
;;; up just to follow the conventions we use in the body of O-Plan.)

;;; Some especially independent packages, typically code produced outside
;;; the O-Plan project, such as the XP pretty-printer, can be placed in
;;; external-support.  This code does not depend on base-support and is
;;; almost never recompiled.

;;; Most of O-Plan's own code that depends on particular Common Lisp
;;; implementations is confined to dependent-support.  There should be
;;; a version of dependent-support for each Common Lisp that O-Plan
;;; can use.  The conventions for adding new code of this sort are
;;; slightly complex, so we'll describe them here rather than force
;;; you to figure them out.

;;;  * Implementation-dependent code is placed in the UTIL package.
;;;    The package definition itself is portable.  That is, all
;;;    implementations define the same names with the same interface.

;;;  * The UTIL package is defined by the file "util-package" in the system
;;;    base-support (not in dependent-support).  All exports should be
;;;    in that file, but the actual, implementation-specific definitions
;;;    should be in files that are in dependent-support.  Those files
;;;    should all start with (in-package :util).

;;;  * Whenever possible, only implementation-dependent _functions_ should
;;;    be defined.  For instance, if a macro is needed, the implementation-
;;;    dependent aspects should be placed in functions called in the macro's
;;;    expansion, and definitions of those functions should be placed
;;;    in each version of dependent-support.

;;; This arrangement minimizes the amount of implementation-specific code
;;; and allows common-support to depend only on base-support.

(defsystem (support :directory "support"
		    :required-systems (base-support
				       dependent-support
				       external-support
				       common-support)
		    :do-after-load
		      ;; /\/: For code in the OPLAN package
		      (progn
			(use-package '(:dev :mcc :ipc :x) :oplan)
			(use-package :monitors :oplan)
			(use-package :ctxt :oplan)
			(use-package :time-util :oplan)
			(use-package :pseudo-process :oplan)
			(use-package :components :oplan)))
  ;; No files
  ;; /\/: Ok, one file. w/ conditionalized code within.
  ("save-image")
  )


;;; KCL support

#+kcl
(defsystem (dependent-support :directory "support"
			      :required-systems (base-support))
  (kcl-extensions
    :defines (:macros))
  (kcl-conditions
    :defines (:functions :macros :structures)
    :requires (kcl-extensions))
  (kcl-util
    :defines (:functions)
    :requires (xfork))
  (xfork
    :defines (:functions)
    :compiler compile-with-C
    :loader load-with-C)
  (kcl-xwindowio
    :defines (:functions)
    :requires (xfork)
    :compiler compile-with-C
    :loader load-with-C)
  )

;;; KCL Compile/load routine for files that include definitions in C

#+kcl
(defun compile/load-file-with-C-fns (pathname library)
  ;; /\/ Wouldn't know :output-file for compile, because it's not
  ;; an arg to load and hence not given to load-with-C.  Maybe the
  ;; special compilers and loaders should be passed the file struct.
  (let ((source (merge-pathnames (make-pathname :type "lsp") pathname))
	(object (merge-pathnames (make-pathname :type "o") pathname)))
    (when (or (not (probe-file object))
	      (> (file-write-date source) (file-write-date object)))
      (safe-compile-file-with-C-fns pathname))
    (si:faslink object library)))

#+kcl
(defun safe-compile-file-with-C-fns (filename)
  (let ((*readtable* (copy-readtable nil))) ;for % char macro
    (compile-file filename)))

#+kcl
(defun compile-with-C (pathname &key output-file)
  ;; /\/ Ignores output file
  (safe-compile-file-with-C-fns pathname))

#+kcl
(defun load-with-C (pathname)
  (compile/load-file-with-C-fns pathname "-lc"))


;;; Lucid support

#+lucid
(defsystem (dependent-support :directory "support"
			      :required-systems (base-support))
  (lucid-conditions)
  (lucid-util
    :defines (:functions))
  (lucid-xwindowio
    :defines (:functions))
  )


;;; Allegro support

#+:allegro
(defsystem (dependent-support :directory "support"
			      :required-systems (base-support))
  (allegro-util
    :defines (:functions))
  (allegro-xwindowio
    :defines (:functions))
  )


;;; Common Support

(defsystem (common-support :directory "support"
			   :required-systems (base-support
					      external-support))
  ;; For util package and macros, see system base-support.
  (util-functions
    :defines (:functions))
  (inf-arithmetic
    :defines (:functions		;but inline functions
	      :macros))
  (simple-matcher
    :defines (:functions))		;/\/ some exports too
  (time
    :defines (:package :functions))
  (pseudo-process)
  (developerlib)
  (context)				;was in DB
  (xwindowio
    :requires (developerlib))
  ("whats-going-on"			;requires UTIL package
    :requires (developerlib))
  ("pop-gensym"				;/\/ is a fn name in UTIL pkg
    :requires (context))
  (monitors				;requires xp
    :requires (developerlib))
  ;(xp)					;now in external-support
  ;(psgraph)				;now in external-support
  (run-lights
    :requires (ipc xwindowio))
  (components
    :requires (developerlib		;for var names
	       context			;for var names
	       pseudo-process))		;:includes struct
  
  ;; The IPC and MCC modules refer to each other.  This has been
  ;; resolved by putting the MCC definitions (package, structure,
  ;; and special declarations) in a separate file.
  (ipc
    :defines (:package :macros :functions)
    :requires (util-functions		;eg concat-name for defmessage
	       pseudo-process
	       developerlib
	       "whats-going-on"
	       mcc-defs))
  (ipc-inter-agent
    :requires (ipc))
  (mcc-defs
    :requires (developerlib
	       monitors
	       pseudo-process		;to use the package
	       components))		;to use the package
  (mcc-client
    :defines (:functions)
    :requires (developerlib
	       mcc-defs
	       ipc
	       components))
  #+:undef
  (mcc-server				;/\/ really part of the IM
    :requires (developerlib
	       mcc-defs
	       pseudo-process
	       ipc
	       monitors))
  )


;;; External support

(defsystem (external-support :directory "support")
  #+kcl
  (xp)
  #+kcl
  (loop)
  (psgraph
    :defines (:functions))		;so far as we're concerned
  )


;;; Base-support contains portable code needed at compile-time by both
;;; common-support and dependent-support (and then, indirectly, by pretty
;;; much everything else).  The main effect of this system is to define
;;; the UTIL package, which has some of its function definitions filled in
;;; by dependent-support and others filled in by parts of common-support.

;;; Since the base-support part of UTIL is portable, it would normally
;;; be in common-support; but then dependent- support would require
;;; common-support (instead of base-support), a much larger system.
;;; In short, the base-support system exists to eliminate some
;;; unnecessary dependencies.  It adds some unnecessary dependencies
;;; too, because many files in common-support don't need the UTIL
;;; package.  But it allows us to use UTIL freely, without having to
;;; add "util" as a required file each time, and the definitions in
;;; UTIL should seldom require modification.

;;; /\/: Parts of developerlib should probably go here as well.

(defsystem (base-support :directory "support")
  (util-package
    :defines (:package)
    :do-after-load (use-package :util :oplan))
  (util-macros
    :defines (:macros)
    :requires (util-package))
  (arithmetic-dcls
    :defines (:macros :types)
    :requires (util-package)))

;;; End
