;;;; File: ta-toplevel.lsp
;;; SCCS Version: %W%
;;; Contains: Task Assigner.
;;; Author: Richard Kirby (rbk) and Jeff Dalton
;;; Created: Tue Jul 10 13:20:48 1990
;;; Updated: Mon Jul 10 17:25:02 1995 by Jeff Dalton
;;; Copyright: (c) 1992, AIAI, University of Edinburgh
;;; This material may be reproduced by or for the U.S. Government pursuant
;;; to the copyright license under the clause at DFARS 252.227-7032
;;; (June 1975) -- Rights in Technical Data and Computer Software (Foreign).

(in-package :task-assigner)

(use-package '(:util :dev :pseudo-process :ipc :mcc :xwindowio :initsystem))

(import '(oplan::make-event))

(import 'ta-startup :oplan)		;/\/ for call-component-startup

;;; At the top of our window, we have something like this:
;;;
;;;   Status: plan option 1 - planning ...
;;;   Domain: pacifica
;;;   Task: Operation_Blue_Lagoon
;;;   Authority: plan(all=inf), execute=(all=no)
;;;

(defvar *ta-iostream* nil)

(defvar *status* "planner uninitialised") 	;but see ta-initialize
(defvar *domain* "none")
(defvar *task*   "none")
(defvar *plan-option* nil)
(defvar *plan-phase-authority* :all)		;node number or all
(defvar *plan-level-authority* :inf)		;level number
(defvar *execute-phase-authority* :all)		;node number or all
(defvar *execute-authority* :no)		;yes or no

(defvar *available-tasks* '())			;list of task names

(defparameter *status-lines*
  ;; See print-from-descriptions for an explanation.
  '( ("Status: " (when *plan-option* "plan option " *plan-option* " - ")
                 *status*)
     ("Domain: " *domain*)
     ("Task: " *task*)
     ("Authority: "
       "plan(" *plan-phase-authority* "=" *plan-level-authority* "), "
       "execute(" *execute-phase-authority* "=" *execute-authority* ")" )
   ))

(defparameter *choice-options*
  '( (1 "Initialise Planner")
     (2 "Input TF")
     (3 "Set Task")
     (4 "Add to Task")
     (5 "Plan View")
     (6 "World View")
     (7 "Replan")
     (8 "Execute Plan")
     (9 "Quit") ))

(defvar *choices* '(1 9))

;;; Present-options
;;;
;;; Active-choices holds a list of numbers of choices that are valid.
;;;

(defun present-options (active-choices)
  (clear-screen *ta-iostream*)
  (print-from-descriptions *ta-iostream* *status-lines*)
  (format *ta-iostream* "~%")
  (dolist (option *choice-options*)
    (let ((number (car option))
	  (name (cadr option)))
      (format *ta-iostream* "~A ~D) ~A~%"
	      (if (member number active-choices) "*" " ")
	      number
	      name)))
  (format *ta-iostream* "~%Please choose a number:- ")
  (force-output *ta-iostream*))


;;; The micro-language used for status lines provides for lists of
;;; line descriptions.  Each line description is a list of <d-element>s.
;;; A <d-element> can be:
;;;
;;;   A string -- printed as is, w/o escape chars or quotes
;;;   A symbol -- treated as a variable; the symbol-value is
;;;               printed w/o escape chars or quotes.
;;;   A list   -- a control sequence.
;;;
;;; The only valid control sequence is:
;;;
;;;   (when <guard-variable> <d-element>*)
;;;      If the value of the guard-variable is true, the d-elements
;;;      are processed in order; otherwise the d-elements are ignored.
;;;

(defun print-from-descriptions (stream descriptions)
  (let ((*print-case* :downcase))
    (dolist (descr descriptions)
      (print-from-descr stream descr)
      (terpri stream))))

(defun print-from-descr (stream descr)
  (dolist (d descr)
    (etypecase d
      (string (princ d stream))
      (symbol (princ (symbol-value d) stream))
      (list
       (ecase (car d)
	 ((when)
	  (let ((guard-var (cadr d))
		(sub-descr (cddr d)))
	    (when (symbol-value guard-var)
	      (print-from-descr stream sub-descr)))))))))


;;;; Menu requests.

;;; All menus have an escape entry (usually "QUIT") in case the user
;;; wants to get rid of the menu without making a selection.  When we
;;; ask the user a question by displaying the question in the TA window
;;; and calling read-line, the user can escape by typing return (thus
;;; producing a null string).

;;; Ask-user is useful when implementing the latter convention.

(defun ask-user (question &rest format-args) ; -> string or nil
  (clear-screen *ta-iostream*)
  (apply #'ask-ta-user question format-args))

(defun ask-ta-user (question &rest format-args)
  (format *ta-iostream* "~&~%~?" question format-args)
  (let ((answer (read-line *ta-iostream*)))
    (if (string= answer "")
	nil
      answer)))

(defun tell-ta-user (format-string &rest format-args)
  (format *ta-iostream* "~&~?~%" format-string format-args))


;;; Get TF file.

;;; The TF directory starts as $OPLANTFDIR but can be changed.
;;; The Planner will use $OPLANTFDIR as a pathname default, so
;;; we can send it a simple name (as a pathname) if that's still
;;; the current directory.  Otherwise, we send a full pathname.

;;; Note that the :cd option takes the directory name relative
;;; the current directory of the O-Plan process (ie, the directory
;;; it was started in from the shell) while :enter takes the file
;;; name relative to the current TF directory.

;;; We call DIRECTORY each time we're about to put up the menu,
;;; because a new TF file might have been added since the last
;;; time we looked.

;;; /\/: However, if we've just changed TF dir, we'll do two in a
;;; row when we could get away with one.  See change-tf-directory.

(defvar *current-tf-dir* nil)

(defun get-tf-file-from-user () ; -> pathname or nil
  (when (null *current-tf-dir*)
    (setq *current-tf-dir* (concat-string (get-parameter :oplan-tf-dir) "/")))
  (let ((tf-paths nil)
	(tf-names nil)
	(name nil))
    (while (null name)
      (setq tf-paths (find-all-tf-files *current-tf-dir*))
      (setq tf-names (sort (mapcar #'pathname-name tf-paths) #'string<))
      (unless tf-names
	(display-message-and-wait *ta-iostream*
	    "No domain files found in:~%~A~%" *current-tf-dir*))
      (let ((domain
	     (big-menu-request
	       `("-heading" ,(if tf-names "Domains to choose from:" "Options:")
		 ,@tf-names
		 "-line"
		 "Change directory=:cd"
		 "Enter filename=:enter"
		 "-line"
		 "QUIT=:quit")
	       :read-function #'read-line)))
	(cond ((equal domain ":quit")
	       (return-from get-tf-file-from-user nil))
	      ((equal domain ":enter")
	       (setq name (enter-tf-file-name)))
	      ((equal domain ":cd")
	       (change-tf-directory))
	      ;; If we get this far, one of the tf-names was chosen 
	      ;; rather than some option such as :cd or :quit.
	      ((equal (pathname *current-tf-dir*)
		      (pathname
		       (concat-string (get-parameter :oplan-tf-dir) "/")))
	       ;; We're in the default TF dir, so use only the name.
	       (setq name (make-pathname :name domain)))
	      (t
	       ;; We're not in the default TF dir, so use a full pathname.
	       (setq name
		     (find domain tf-paths :key #'pathname-name
			                   :test #'string=))
	       (assert (not (null name)))))))
    name))

(let ((saved-dir nil)
      (saved-write-date nil)
      (saved-result nil))
  (defun find-all-tf-files (dir)
    (let ((write-date (file-write-date dir)))
      (cond ((and (equal dir saved-dir)
		  (equal write-date saved-write-date))
	     saved-result)
	    (t
	     (setq saved-dir dir
		   saved-write-date write-date
		   saved-result (find-all-files dir "tf")))))))

(defun enter-tf-file-name () ; -> pathname or nil
  (let ((tf (ask-user "Enter name of TF file: ")))
    (when (null tf)
      (return-from enter-tf-file-name nil))
    (setq tf (merge-pathnames
	       (merge-pathnames tf (make-pathname :type "tf"))
	       *current-tf-dir*))
    (cond ((probe-file tf) tf)
	  (t (display-message-and-wait *ta-iostream*
	        "There is no file ~S." tf)
	     nil))))

(defvar *old-tf-dirs* '())

(defun change-tf-directory () ; works via side effects
  (let ((dir (get-tf-dir-from-user)))
    (when (null dir)
      (return-from change-tf-directory))
    (unless (eql #\/ (char dir (1- (length dir))))
      (setq dir (concat-string dir "/")))
    (cond ((not (probe-file dir))
	   (display-message-and-wait *ta-iostream*
	     "There is no ~S." dir))
	  ((null (find-all-tf-files dir))
	   (display-message-and-wait *ta-iostream*
	     "There are no TF files in ~S." dir))
	  (t
	   ;; Looks OK.
	   (setq *old-tf-dirs* (remove dir *old-tf-dirs* :test #'equal))
	   (push *current-tf-dir* *old-tf-dirs*)
	   (setq *current-tf-dir* dir)))))

(defun get-tf-dir-from-user () ; -> string or nil
  (if (null *old-tf-dirs*)
      (ask-user "Enter directory name: ")
    (let ((dir (menu-request
		 `("-heading" "Directories to choose from:"
		   ,@*old-tf-dirs*
		   "-line"
		   "Enter directory name=:enter"
		   "-line"
		   "QUIT=:quit")
		 :read-function #'read-line)))
      (cond ((equal dir ":quit")
	     nil)
	    ((equal dir ":enter")
	     (ask-user "Enter directory name: "))
	    (t
	     dir)))))


;;; Get task

;;; We return the name of a task schema as a string, or nil if the
;;; user quits.

(defun get-task-from-user () ; -> string or nil
  (assert (not (null *available-tasks*)))
  (let ((task (string-downcase
	       (menu-request
		 `("-heading" "Task to choose from:"
		   ,@*available-tasks*
		   "-line"
		   "QUIT")
		 :read-function #'read-line))))
    (if (string-equal task "quit")
	nil
      task)))

(defun add-to-available-tasks (more-tasks)
  ;; /\/: Should use replacing-merge from the TF compiler.
  (setq *available-tasks*
	(append (set-difference *available-tasks* more-tasks :test #'equal)
		more-tasks)))


;;; For "Add to Task"

(defun get-task-addition-from-user ()
  (ecase (task-addition-menu-choice)
    (:action
     (loop
       (let ((pattern (ask-user-for-action-pattern)))
	 (when (null pattern)
	   (return nil))
	 (tell-ta-user "Action: ~(~S~)" pattern)
	 (when (ask-if *ta-iostream* "y" "Ok?")
	   (return
	     `(:action ,pattern))))))
    (:quit
     nil)))

(defun task-addition-menu-choice ()
  (menu-request
    '("-heading" "Add to Task"
      "Add an Action=:action"
      "-line"
      "Quit=:quit")))

(defun ask-user-for-action-pattern ()
  ;; Try to make sure we get something that looks like a pattern
  (clear-screen *ta-iostream*)
  (loop
    (let ((line  (ask-ta-user "Action pattern: ")))
      (when (null line)
	(return nil))
      (handler-case (read-from-string line)
        (error (c)
	  (tell-ta-user "Error: ~A" c)
	  (tell-ta-user "Try again or enter a blank line to quit.~%"))
	(:no-error (pattern)
	  (if (and (consp pattern)
		   (symbolp (car pattern)))
	      (return pattern)
	    (tell-ta-user
	      "The pattern must be a list that begins with a symbol.")))))))


;;; Get node name for world view

;;; The user should type a node number, such as "3-9" for node-3-9.

(defun get-node-id-from-user () ; -> string or nil
  (ask-user "  View at end_of node-"))


;;; Quit?

(defconstant quit-choices '("-heading" "Quit" "Yes" "-line" "No"))

(defun really-quit-p ()
  (equal (menu-request quit-choices :read-function #'read-line)
	 "Yes"))


;;; Other utilities

(defun display-message-and-wait (stream msg &rest args)
  (clear-screen stream)
  (apply #'format stream msg args)
  (format stream "~%~%~%~%~%Hit <return> to continue")
  (read-line stream))

(defun not-implemented-yet (stream)
  (display-message-and-wait stream "~%Not implemented yet~%"))

(defun send-to-planner (&rest args)
  (ipc-send :LEFTIN :EVENT (make-event :BODY args)))


;;; Main program

(defun ta-startup ()
  
  ;; This is the initial function for the TA process.  After startup,
  ;; it is not called again.  Instead, ta-event-handler is called
  ;; whenever a message arrives.
  
  (unless (is-init *whoami*)
    (dev-error *whoami* "is-init failed."))
  (ta-initialise)

  ;; Make breaks, etc. go to the debug window.
  (x-redirect-errors :tainout)

  ;; Add :status :run to the ipc-set-event-handler if no listen
  ;; stream is registered.
  (ipc-set-event-handler 'ta-event-handler)
  (ipc-register-listen-stream *ta-iostream*)
  )

(defun ta-initialise ()
  (setq *ta-iostream* (x-get-stream :tainout))
  (setq *status*
	(format nil "Version ~A ~A uninitialised"
		user::*oplan-version*
		user::*oplan-release-date*))
  (setq *choices* '(1 9))
  (present-options *choices*))

(defun ta-event-handler (self)
  (when (next-event-p self)
    (handle-ta-event self (next-event self)))
  (ta-monitor))

(defun handle-ta-event (self event)
  (assert (message-p event))
  (assert (eq (message-route event) :leftout))
  (ipc-handle-message self event))


;;;; Messages

(defmessage (:TA :INIT) (status)
  (cond ((eq status :OK)
	 (setq *choices* '(1 2 9))
	 (setq *status* "planner-initialized")
	 (setq *plan-option* 1)
	 (setq *domain* "none")
	 (setq *task* "none")
	 (setq *available-tasks* '())
	 (present-options *choices*))
	(t
	 (dev-debug-msg "Couldn't initialise" *whoami* :fatal-error))))

(defmessage (:TA :DOMAIN) (result)
  ;; Result is (<pathname-name of the TF file> . <list of task names>)
  ;; or nil.
  (if (null result)
      (progn
	(setq *status* "domain tf error")
	(dev-debug :fatal-error "Couldn't load domain."))
    (let ((domain (car result))
	  (tasks (cdr result)))
      (setq *status* "domain loaded")
      (when tasks
	(setq *choices* '(1 2 3 9))
	(add-to-available-tasks tasks))
      (if (equal *domain* "none")
	  (setq *domain* domain)
	  (setq *domain* (concat-string *domain* "+" domain)))))
  (present-options *choices*))

(defmessage (:TA :FINISHED) ()
  (setq *status* "planner finished")
  (setq *choices*
	(if (execution-is-allowed-p)
	    '(1 2 4 5 6 7 8 9)
	    '(1 2 4 5 6 7   9)))
  (present-options *choices*))

(defun execution-is-allowed-p ()
  (if (parameter-set-p :allow-exec)
      (get-parameter :allow-exec)
    (is-get-config-entry :micro-exec)))

(defmessage (:TA :NO-MORE-ALTERNATIVES) ()
  (setq *status* "planner out of alternatives")
  (setq *choices* '(1 9))
  (present-options *choices*))

(defmessage (:TA :CHECKED) (&rest result)
  ;; For when someone manually calls :CHECK-PLAN
  (display-message-and-wait *ta-iostream*
    "Plan check result: ~{~S~^ ~}." result)
  (present-options *choices*))


;;;; User commands

(defun ta-monitor ()
  (when (listen *ta-iostream*)
    (let ((input (read *ta-iostream*)))
      (cond ((member input *choices*)
	     (process-user-command input)
	     (present-options *choices*))
	    (t (format *ta-iostream*
		       "~%Please choose a number from ~A :- " *choices*))))))

(defun process-user-command (input)
  (case input
    (1
     ;; Initialise the planner to a null state.
     (send-to-planner :INIT)
     (setq *available-tasks* '())
     (setq *choices* '(9)))
    ;; Specify the domain that the planner will work in.
    (2
     (let ((domain (get-tf-file-from-user)))
       (if domain
	   (send-to-planner :DOMAIN domain))))
    ;; Specify the task to plan.
    (3
     (let ((task (get-task-from-user)))
       (when task
	 (send-to-planner :SET-TASK task)
	 (setq *task* (subseq task 5))		;minus the initial "task_"
	 (setq *status* "planning ...")
	 (setq *choices* '(5 6 9)))))
    ;; Add to the task.
    (4
     (let ((addition (get-task-addition-from-user)))
       (when addition
	 (apply #'send-to-planner :ADD-TO-TASK addition)
	 ;; /\/: Should probably change status, etc.
	 )))
    ;; View the current state of the plan.
    (5
     (send-to-planner :USER :PLAN-VIEW))
    ;; View the world at a particular point of the plan.
    (6
     (let ((node-id (get-node-id-from-user)))
       (when node-id
	 (send-to-planner :USER :WORLD-VIEW node-id))))
    ;; Replan.
    (7
     (setq *status* "replanning ...")
     (setq *choices* '(5 6 9))
     (setq *execute-authority* :no)
     (send-to-planner :POISON-STATE :REPLANNING))
    ;; Send the plan for execution.
    (8
     (setq *status* "executing ...")
     (setq *execute-authority* :yes)
     (send-to-planner :EXECUTE))
    ;; Quit the planner.
    (9
     (when (really-quit-p)
       (send-to-planner :KILL)
       (setq *status* "exiting ...")
       (setq *choices* '(9))
       ;; Stop running after this cycle.
       (terminate-pprocess)))
    (otherwise
     (error "Totally unknown option: ~S." input))))

;;; ---------------------------- Change History ----------------------------
;;; (Who)   (When) 		     (What)
;;;
